Require Import ssreflect ssrfun ssrbool eqtype ssrnat div seq.
Require Import ssralg fintype perm.
Require Import dvdring matrix  bigop zmodp mxalgebra.
Require Import matrix_facts.

Import GRing.Theory.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.



Section FieldRank.
Variable K : fieldType.

Local Open Scope ring_scope.

(** Description of the Rank function in SSReflect *)

(* takes a seq l and return
  a : R
  E F : Matrix
such that
l * E = a :: 0 ... 0
E * F = F * E = 1
*)
Fixpoint simpl_line {n:nat} :=
 match n return 'rV[K]_n.+1 -> K * 'M[K]_n.+1  with
  | 0 => fun l => (l 0 0, 1%:M)
  | p.+1 => fun (l :'rV[K]_( 1 + (1 + p))) =>
        let hd : K := (lsubmx l) 0 0 in
        let tl := rsubmx l in
          if hd == 0 then
            let: (a,E) := @simpl_line p tl in
            let E1 := (lift0_mx E : 'M[K]_(1 + (1 + _))) in
(* tperm_mx 0 1 exchange the first two column but as a det of -1
   so we change it to  0 1 0 ....
                      -1 0 0 ....
                       0 0  I
   instead to have a det of 1
*)
            let p1 := (block_mx ((-1)%:M) 0 0 1%:M : 'M[K]_(1 + (1 + _))) in
            let P1 := (tperm_mx 0 1 : 'M[K]_(1 + (1 + _ ))) in
              (- a, E1 *m P1 *m p1)
          else
            (hd, block_mx 1%:M ((- hd^-1) *: tl) 0 1%:M)
end.



Fixpoint Rank (n m:nat) : 'M[K]_(n,m) ->  nat :=
 match n,m return 'M[K]_(n,m) -> nat with
  | _.+1, _.+1 => fun A : 'M[K]_(1+_, 1+_) =>
    let: (a, E) := simpl_line (usubmx A) in
     let B : 'M[K]_(1 + _, 1 + _) := A *m E in
      if (a == 0) then Rank (dsubmx B)
        else (1 + Rank (drsubmx B))%N
  | _,_ => fun _ => 0%N
end.

Fixpoint Det (n: nat) : 'M[K]_n -> K :=
 match n return 'M[K]_n -> K  with
  | _.+1 => fun A : 'M[K]_(1+_) =>
    let: (a, E) := simpl_line (usubmx A) in
      let B : 'M[K]_(1+_) := A *m E in
    if a == 0 then
      0
    else (a * Det (drsubmx B))
  | _ => fun _ => 1
end.


(** Description of the Rank and Det functions in an executable way *)

(* definitions and basic operations on matrices *)
(* matrix structure in block
   a    l
   c    M
with a one element, l a line, c a column and M a Matrix
*)

Inductive Matrix : Type :=
 | eM
 | cM of K & seq K & seq K & Matrix.


(* a tool function that mix zip with map *)
Definition zipWith (A B C:Type) (f: A -> B -> C)
  (l : seq A) (l' :seq B) : seq C :=
 map (fun xy: A*B => let (x,y) := xy in f x y) (zip l l').


(* seq of zeroes, of size n *)
Definition zeroes (n:nat) : seq K := nseq n 0.

(* identity Matrix of size n x n *)
Fixpoint ident (n:nat) : Matrix :=
 if n is (S p) then
  let z := zeroes p in
    cM 1 z z (ident p)
 else eM.

(* concatenate a column to a matrix *)
Fixpoint glue (c: seq K) (M : Matrix) :=
 match c,M with
  | [::],_ => M
  | hd :: tl ,eM => cM hd [::] tl eM
  | hd :: tl , cM a l c' M' =>
    cM hd (a :: l) tl (glue c' M')
end.

(* projector to get the lower part of a Matrix
              a l
   ex_dsubmx  c M  = c M
*)
Definition ex_dsubmx (M : Matrix) :=
 match M with
  | eM => eM
  | cM a l c M => glue c M
end.

(* projector to get the lower right part of a Matrix
                   a l
        ex_drsubmx c M = M
*)
Definition ex_drsubmx (M : Matrix) :=
  match M with
  | eM => eM
  | cM _ _ _ M => M
end.


(* scalar product of 2 seqs  *)
Definition prods (l c: seq K) := foldr +%R 0 (zipWith *%R l c).


(* addition on seq
   if l1 and l2 are not of the same size, the result is of
   size l1 (truncating l2 or adding trailing 0s to l2 if necessary)
 *)
Fixpoint adds (l1 l2 : seq K) : seq K :=
 match l1,l2 with
  | [::], _ => [::]
  | _, [::] => l1
  | hd :: tl, hd' :: tl' => (hd + hd') :: (adds tl tl')
end.

(* multiplication of a seq by a scalar *)
Definition multEV (a : K) (l: seq K) : seq K :=
 map (fun x => a * x) l.



(* multiply a seq by a Matrix  (line * Matrix) *)
Fixpoint multLM  (s :seq K) (A:Matrix) : seq K :=
    match s,A with
      | [::], _ => [::]
      | _, eM   => [::]
      | hd :: tl, cM a l c mat =>
        let k := multLM tl mat in
          (hd * a + prods tl c) ::
          (adds (map (fun x => hd * x) l) k)
end.


(* same with Matrix multiplied by seq (Matrix * column) *)
Fixpoint multMC (A: Matrix) (s :seq K) : seq K :=
    match A,s with
      | eM, _   => [::]
      | _,[::] => [::]
      | cM a l c mat, hd :: tl  =>
        let k := multMC mat tl in
          (a * hd + prods l tl) ::
             (adds (map (fun x => x * hd) c) k)
end.

(* addition of Matrix *)
Fixpoint add (m1 m2 : Matrix) : Matrix :=
 match m1,m2 with
  | eM, _ => m2
  | _, eM => m1
  | cM a1 l1 c1 m1, cM a2 l2 c2 m2 =>
    cM (a1+a2) (adds l1 l2) (adds c1 c2) (add m1 m2)
end.


(* multiplication of two seq which ends up in a Matrix (column * line) *)
Fixpoint mults (c l : seq K) : Matrix :=
    match c,l with
     | [::], _ => eM
     | _, [::] => eM
     | u :: us, v :: vs =>
       cM (u * v) (map (fun x => u * x) vs)
          (map (fun x => x * v) us)
          (mults us vs)
end.

(* Multiplication of Matrix *)
Fixpoint mult (M1 M2 : Matrix) : Matrix :=
 match M1,M2 with
  | eM, _ => eM
  | _, eM => eM
  | cM a1 l1 c1 m1, cM a2 l2 c2 m2 =>
    cM (a1 * a2 + prods l1 c2)
       (adds (multEV a1 l2) (multLM l1 m2))
       (adds (multEV a2 c1) (multMC m1 c2))
       (add (mults c1 l2) (mult m1 m2))
end.

(* executable version of the simple line function *)
Fixpoint ex_simpl_line (n:nat) (l:seq K) : K * Matrix :=
 match n,l with
  | 0, [::] => (0, eM)
  | 0, hd :: tl => (hd, cM 1 [::] [::] eM)
  | S _, [::] => (0, eM)
  | S p, hd :: tl =>
      if hd == 0 then
            let: (a,E) := ex_simpl_line p tl in
            let E1 := cM 1 (zeroes (size tl)) (zeroes (size tl)) E in
            let p := cM 0 (zeroes (size tl).-1)
                (zeroes (size tl).-1)
                (ident (size tl).-1) in
            let P1 := cM 0 (1 :: zeroes (size tl).-1)
                (-1 :: zeroes (size tl).-1) p in
               (- a, mult E1 P1)
          else
   (hd, cM 1 (map (fun x => - x / hd) tl) (zeroes (size tl))
     (ident (size tl)))
end.

(* Executable version of the rank, we need to know
   the number of lines in the Matrix
*)
(* n == number of lines *)
Fixpoint ex_Rank (n: nat) (M: Matrix) : nat :=
 match n,M with
  | _,eM => 0%N
  | 0,_ =>  0%N
  | S p, cM hd l c M' =>
    let: (a, E) := ex_simpl_line (size l) (hd :: l) in
    let B := mult M E in
    if a == 0 then
      ex_Rank p (ex_dsubmx B)
    else (1 + ex_Rank p (ex_drsubmx B))%N
end.

Fixpoint ex_Det (n: nat) (M: Matrix) : K :=
 match n,M with
  | _,eM => 1
  | 0,_ =>  1
  | S p, cM hd l c M' =>
    let: (a, E) := ex_simpl_line p (hd :: l) in
    let B := mult M E in
    if a == 0 then
      0
    else (a * ex_Det p (ex_drsubmx B))
end.



(** Correctness lemmas *)
Lemma zipWith0 : forall (A B C: Type) (f : A -> B -> C)
  (l : seq A), zipWith f l [::] = [::].
Proof. by move=> A B C f []. Qed.

(* the matrices returned by simpl_line are inverse of each other *)
Lemma slinv: forall (n:nat) (l:'rV[K]_(1 + n)),
   let: (_,E) := simpl_line l in  E \in unitmx.
Proof.
elim => [ | n hi] l /=.
- by rewrite unitmx1.
case hl0 : ((lsubmx l) 0 0 == 0).
- move: (hi (rsubmx l)).
  case: simpl_line => [a E] hE.
  rewrite !unitmx_mul (lift0_unitmx hE) /tperm_mx unitmx_perm /=.
  set B := block_mx ((-1) %:M) 0 0 1%:M.
  have h : B *m B = 1%:M.
  + by rewrite mulmx_block !mulmx0 !mul0mx mulmx1 !addr0 add0r -scalar_mxM
          mulNr mulrN opprK mulr1 -scalar_mx_block.
  by case (mulmx1_unit h).
set B := block_mx _ _ _ _.
set D := block_mx 1%:M ( ((lsubmx l) 0 0)^-1 *: rsubmx l) 0 1%:M
    : 'M_(1 + (1 + n)).
have h : D *m B = 1%:M.
- rewrite mulmx_block !mulmx0 !mul0mx !mul1mx mulmx1 !addr0 add0r.
  by rewrite scale_add_distr addNr  scale0mx -scalar_mx_block.
by case (mulmx1_unit h).
Qed.

Lemma sldet: forall (n:nat) (l:'rV[K]_(1 + n)),
  let : (_,E) := simpl_line l in \det E = 1.
Proof.
elim => [ | n hi] l /=.
- by rewrite det1.
case hl0 : ((lsubmx l) 0 0 == 0).
- move: (hi (rsubmx l)).
  case: simpl_line => [a E] hE.
  rewrite !det_mulmx !det_ublock hE !det1 det_perm det_scalar odd_tperm /=.
  by rewrite !mulr1 !mul1r expr1 mulrN mulNr opprK mul1r.
by rewrite det_ublock !det1 mulr1.
Qed.

(* the remaining element a is 0 iff the line was all 0s *)
Lemma sl0 : forall (n :nat) (l:'rV[K]_(1 + n)),
  let: (a,_) := simpl_line l in (a == 0) = (l == 0).
Proof.
elim => [ | n hi] l /=.
- rewrite {2}[l]mx11_scalar.
  apply/eqP/eqP.
  + move => ->.
    by apply/matrixP => i j; rewrite !mxE /= mul0rn.
  move/matrixP => h.
  move: (h 0 0).
  by rewrite !mxE /= mulr1n.
case hl : ((lsubmx l) 0 0 == 0).
- move: (hi (rsubmx l)).
  case: simpl_line => [a E] h.
  rewrite oppr_eq0 -[l]hsubmxK.
  apply/idP/eqP.
  + rewrite h => hc; rewrite (eqP hc).
    rewrite -[0:'rV_(1 + n.+1)]row_mx0.
    f_equal.
    rewrite [lsubmx l]mx11_scalar (eqP hl).
    by rewrite -scalemx1 scale0mx.
  rewrite -[0 :'rV_(1 + n.+1)]row_mx0.
  case/eq_row_mx => h1 h2.
  by rewrite h h2.
apply/eqP/eqP.
- by move => h; move: hl; rewrite h eq_refl /=.
move => ->.
by rewrite !mxE.
Qed.

Lemma aux_tperm : forall n m (c1 c2:'cV[K]_n) (M: 'M[K]_(n,m)),
   (row_mx c1 (row_mx c2 M)) *m tperm_mx 0 1 =
   (row_mx c2 (row_mx c1 M)).
Proof.
move => n m c1 c2 M.
rewrite mul_row_perm mulmx1 tpermV.
apply/matrixP => i.
rewrite [m.+2]/(1 + (1 + m))%N => j; rewrite !mxE.
case: tpermP.
- move => hj0.
  have -> : j = 0 by apply/ord_inj; rewrite hj0.
  case: splitP => x; first by rewrite [x]ord1.
  move/eqP; rewrite eqSS !mxE; move/eqP => hx0.
  have -> : x = 0 by apply/ord_inj.
  case: splitP => y //; rewrite [y]ord1; clear y => _.
  by case: splitP => y //; rewrite [y]ord1.
- move => hj1.
  have -> : j = 1 by apply/ord_inj; rewrite hj1.
  case: splitP => x //; rewrite [x]ord1; clear x => _.
  case: splitP => x; first by rewrite [x]ord1.
  move/eqP; rewrite eqSS !mxE; move/eqP => hx0.
  have -> : x = 0 by apply/ord_inj.
  by case: splitP => y //; rewrite [y]ord1.
case: splitP => x.
- rewrite [x]ord1 => hx.
  by have -> : j = 0 by apply/ord_inj.
rewrite !mxE => hx.
have -> : j = lift 0 x by apply/ord_inj.
case: splitP => y //.
rewrite [y]ord1 => hy _.
have -> : x = 0 by apply/ord_inj.
case.
by apply/ord_inj.
Qed.

Lemma slmul: forall (n:nat) (l: 'rV[K]_(1 + n)),
 let: (a , E) := simpl_line l in
   l *m E = row_mx (a%:M) 0.
Proof.
elim => [ | n hi] l /=.
- rewrite mulmx1.
  apply/rowP => i; rewrite [i]ord1 !mxE /=.
  case: splitP => // j.
  by rewrite [j]ord1 -mx11_scalar.
case hl: ((lsubmx l) 0 0 == 0).
- move: (hi (rsubmx l)).
  case: simpl_line => [a E] => hmul.
  rewrite !mulmxA -[l]hsubmxK mul_row_col !mul_mx_row hmul mulmx1 !mulmx0
    add_row_mx addr0 add0r.
  rewrite aux_tperm mul_row_col [lsubmx l]mx11_scalar (eqP hl) scalar0.
  rewrite row_mx0 mul0mx addr0 mul_scalar_mx scale_row_mx -mul_scalar_mx
      -scalar_mxM mulrN mulr1.
  f_equal.
  by apply/matrixP => i j; rewrite !mxE mulr0.
rewrite -{1}[l]hsubmxK (@mul_row_block _ 1 1 _ 1 _) !mulmx1
  mulmx0 addr0 -scalemxAr scalemxAl.
f_equal.
- apply/rowP => i; rewrite !mxE.
  by rewrite [i]ord1 lshift0 mulr1n.
apply/rowP => i; rewrite !mxE lshift0 rshift1.
rewrite !big_ord_recl big_ord0 addr0 !mxE !mulNr lshift0 rshift1 mulVf
   ?mul1r ?addNr => //.
move : hl; rewrite !mxE lshift0 => hl.
by apply/negbT.
Qed.

(* properties of the simple_line function *)
CoInductive sl_spec n (l:'rV[K]_(1 + n)): K * 'M[K]_(1 + n) -> Type :=
  slSpec a E of
  l *m E = row_mx (a%:M) 0
& E \in unitmx
& \det E = 1
& (a == 0) = (l == 0) : sl_spec l (a,E).

Lemma slP : forall (n:nat) (l: 'rV[K]_(1 + n)), sl_spec l (simpl_line l).
move => n l.
move: (slinv l).
move: (slmul l).
move: (sl0 l).
move: (sldet l).
case: simpl_line => [a E] F h1 h2 h3.
by split.
Qed.


Lemma rank0M: forall (n m:nat) (M: 'M[K]_(n,m)),
  \rank (col_mx (0: 'rV[K]_m) M) = \rank M.
Proof. by move => n m M; rewrite eqmxM0M. Qed.

(* rank ( a    0 .... 0  )  = 1 + rank M if a != 0
          c1
          ...     M
          cn
*)
Lemma rankaM : forall (n m:nat) (M: 'M[K]_(n,m)) (a:K) (c:'cV[K]_n),
 a != 0 -> \rank (block_mx (a%:M) 0 c M) = (1 + \rank M)%N.
Proof.
move => n m M a c ha.
have: ((block_mx a%:M 0 c M)^T :=: (col_mx a%:M c)^T + (col_mx 0 M)^T)%MS.
- rewrite tr_block_mx !tr_col_mx /block_mx.
  apply eqmx_sym.
  by apply addsmxE.
have hC: (col_mx a%:M c) != 0.
- apply/negP; move/eqP => h.
  move/negP : ha; apply.
  move/colP : h => h.
  move: (h 0); rewrite !mxE.
  case: splitP => i hi //.
  by rewrite [i]ord1 !mxE /= mulr1n => ->.
rewrite -mxrank_tr => ->.
rewrite mxrank_disjoint_sum.
- rewrite rank_rV mxrank_tr rank0M.
  by rewrite -trmx_neq0 hC.
rewrite !tr_col_mx !tr_scalar_mx trmx0.
apply/eqP.
apply/rowV0P => v.
rewrite sub_capmx; case/andP.
case/sub_rVP => b hb.
case/submxP => B hB.
have : v 0 0 = 0 /\ v 0 0 = b * a.
- split.
  + rewrite hB !mxE big1 // => i _.
    rewrite !mxE.
    case: splitP => j //.
    by rewrite !mxE /= mulr0.
  rewrite hb !mxE.
  case: splitP => j //.
  by rewrite [j]ord1 !mxE /= mulr1n.
rewrite {3}hb; case => ->.
rewrite -[0](@mul0r _ a). move/(mulIf ha) <-.
by rewrite scale0mx.
Qed.


Lemma rankRank : forall (n m:nat) (M: 'M[K]_(n,m)),
  Rank M = \rank M.
Proof.
elim => [ | n hi] [ | m] M //=.
case: slP => a E h1 h2 h3 h4.
case ha: (a == 0).
- rewrite -(@mxrankMfree K _ _ _ _ E) ?row_free_unit // hi dsub_mulmx.
  rewrite -{2}(@vsubmxK K 1 n _ M) (@mul_col_mx _ 1 n _ _) h1 (eqP ha).
  by rewrite -scalemx1 scale0mx row_mx0 rank0M.
rewrite -(@mxrankMfree K _ _ _ _ E) ?row_free_unit // hi.
rewrite -{2}(@vsubmxK K 1 n _ M) (@mul_col_mx _ 1 n _ _) h1.
set dME : 'M[K]_(n, 1 + m) := dsubmx (M :'M[K]_(1 + n, _)) *m E.
rewrite -[dME]hsubmxK /dME -dsub_mulmx rankaM //.
by apply/negP; rewrite ha.
Qed.

Lemma detDet : forall (n:nat) (M: 'M[K]_n),
 Det M = \det M.
Proof.
elim => [ | n hi] /=.
- by move => M; rewrite [M]flatmx0 -scalar0 det_scalar expr0.
change (n.+1) with (1 + n)%N => M.
case: slP => a E h1 h2 h3.
case ha: (a == 0).
- move/esym => h4.
  rewrite -[M]submxK.
  have -> : ursubmx M = 0 by
    rewrite /ursubmx (eqP h4) -[0]row_mx0 row_mxKr.
  have -> : ulsubmx M = 0 by
    rewrite /ulsubmx (eqP h4) -[0]row_mx0 row_mxKl.
  by rewrite det_lblock det0 mul0r.
move => h4.
rewrite -[\det M]mulr1 -h3 -det_mulmx -{2}[M]vsubmxK mul_col_mx h1.
by rewrite -[dsubmx M *m E]hsubmxK det_lblock det_scalar expr1
     hi -dsub_mulmx.
Qed.

(** executable part **)

(* Translation from seq to ssr row vector *)
Definition seq2rV(n:nat) (l: seq K) : 'rV_(n) :=
  \row_(j < n) nth 0 l j.

(* rewriting lemma to translate non empty seq into row_mx *)
Lemma seq2rVS : forall (n:nat) (a: K) (l: seq K),
  seq2rV (n.+1) (a :: l) = row_mx a%:M (seq2rV n l).
Proof.
rewrite /seq2rV => n a l.
apply/rowP => j. rewrite !mxE.
case: splitP => j0.
- rewrite [j0]ord1 !mxE /=  mulr1n => ->.
  by rewrite nth0.
by move => ->; rewrite !mxE /=.
Qed.

(* Translation from seq to ssr column vector *)
Definition seq2cV (m:nat) (c: seq K) : 'cV_(m) :=
  \col_(i < m) nth 0 c i.

(* rewrite non empty seq into col_mx *)
Lemma seq2cVS : forall (n:nat) (a: K) (l: seq K),
  seq2cV (n.+1) (a :: l) = col_mx a%:M (seq2cV n l).
Proof.
rewrite /seq2cV => n a l.
apply/colP => j. rewrite !mxE.
case: splitP => j0.
- rewrite [j0]ord1 !mxE /=  mulr1n => ->.
  by rewrite nth0.
by move => ->; rewrite !mxE /=.
Qed.

(* Translation from ssr row vector into seq *)
Definition rV2seq (n:nat) (l : 'rV_(n)) : seq K :=
  map ( fun i => l 0 i) (enum 'I_n).

Lemma size_rV2seq: forall n (l: 'rV_(n)), size (rV2seq l) = n.
Proof.
rewrite /rV2seq => n l.
by rewrite size_map size_enum_ord.
Qed.

(* Translation from ssr col vector into seq *)
Definition cV2seq (m: nat) (c : 'cV_(m)) : seq K :=
  map (fun i => c i 0) (enum 'I_m).

Lemma size_cV2seq: forall n (l: 'cV_(n)), size (cV2seq l) = n.
Proof.
rewrite /cV2seq => n l.
by rewrite size_map size_enum_ord.
Qed.

(* Translation from Matrix into ssr matrices *)
Fixpoint Matrix2mat (n m:nat) (mat : Matrix) {struct n} : 'M[K]_(n,m) :=
  match n,m with
   | O, _ => 0
   | _, O => 0
   | S p, S q => match mat with
      | eM => 0
      | cM a l c mat' => block_mx a%:M         (seq2rV q l)
                                  (seq2cV p c) (Matrix2mat p q mat')
  end
end.

(* Translation from ssr matrices to Matrix *)
Fixpoint mat2Matrix (n:nat) : forall (m:nat), 'M[K]_(n,m) -> Matrix :=
 match n as x return forall (m:nat), 'M[K]_(x,m) -> Matrix with
  | O => fun _ _ => eM
  | S p => fun m =>
     match m as y return 'M[K]_(1 + p,y) -> Matrix with
       | O => fun _ => eM
       | S q => fun (mat: 'M[K]_(1 + p,1 + q)) =>
         cM (mat 0 0) (rV2seq (ursubmx mat)) (cV2seq (dlsubmx mat))
         (mat2Matrix (drsubmx mat))
       end
  end.


(* the translation back and forth of row vectors is correct *)
Lemma r2r_ok : forall (n:nat) (l : 'rV[K]_(n)), seq2rV n (rV2seq l) = l.
Proof.
move => n l; apply/matrixP => i j.
by rewrite mxE [i]ord1 (nth_map j) ?nth_ord_enum // size_enum_ord ltn_ord.
Qed.

(* the translation back and forth of column vectors is correct *)
Lemma c2c_ok : forall (n:nat) (c : 'cV[K]_(n)), seq2cV n (cV2seq c) = c.
Proof.
move => n c; apply/matrixP => i j.
by rewrite mxE [j]ord1 (nth_map i) ?nth_ord_enum // size_enum_ord ltn_ord.
Qed.

(* the translation back and forth of matrices is correct *)
Lemma m2m_ok: forall (n m:nat) (mat : 'M[K]_(n,m)),
  Matrix2mat n m (mat2Matrix mat) = mat.
Proof.
elim => [ | n hi] [ | m] /= mat; try by apply/matrixP=> i j; case: i; case: j.
change (n.+1) with (1 + n)%N in mat; change (m.+1) with (1+m)%N in mat.
rewrite -{5}[mat]submxK.
f_equal.
- apply/matrixP => i j /= ; rewrite /ulsubmx mxE [i]ord1 [j]ord1 /= mulr1n.
  by rewrite mxE lshift0 mxE lshift0.
- by apply r2r_ok.
- by apply c2c_ok.
by apply: hi.
Qed.


(* the translation of zeroes is ssr 0 matrix *)
Lemma seq2rV_zeroes : forall n, seq2rV n (zeroes n) = 0.
Proof.
rewrite /seq2rV /zeroes => n.
apply/matrixP => i j; rewrite [i]ord1 !mxE /=.
rewrite nth_nseq.
by case: (j < n)%N.
Qed.

Lemma map1 : forall (l :seq K), map ( *%R^~ 1) l = l.
Proof.
elim => [ | hd tl hi] //=.
by rewrite mulr1 hi.
Qed.

Lemma map0 : forall (l: seq K), map [eta *%R 0] l = zeroes (size l).
Proof.
elim => [ | hd tl hi] //=.
by rewrite mul0r hi.
Qed.

Lemma rV2seq_zeroes : forall n, rV2seq ( 0 : 'rV_n) = zeroes n.
Proof.
rewrite /rV2seq => [[ | n]].
- apply/size0nil.
  by rewrite size_map size_enum_ord.
apply/(@eq_from_nth _ 0).
- by rewrite size_map size_enum_ord size_nseq.
move => i.
rewrite size_map size_enum_ord => hi.
by rewrite (@nth_map (ordinal (S n)) 0 K 0) ?size_enum_ord
      ?nth_nseq ?hi ?mxE.
Qed.

Lemma seq2cV_zeroes : forall n, seq2cV n (zeroes n) = 0.
Proof.
rewrite /seq2cV /zeroes => n.
apply/matrixP => i j; rewrite [j]ord1 !mxE /=.
rewrite nth_nseq.
by case: (i < n)%N.
Qed.

Lemma cV2seq_zeroes : forall n, cV2seq ( 0 : 'cV_n) = zeroes n.
Proof.
rewrite /cV2seq => [[ | n]].
- apply/size0nil.
  by rewrite size_map size_enum_ord.
apply/(@eq_from_nth _ 0).
- by rewrite size_map size_enum_ord size_nseq.
move => i.
rewrite size_map size_enum_ord => hi.
by rewrite (@nth_map (ordinal (S n)) 0 K 0) ?size_enum_ord
      ?nth_nseq ?hi ?mxE.
Qed.

Lemma ident2Matrix : forall n, mat2Matrix (1%:M : 'M[K]_n) = ident n.
Proof.
elim => [ | n hi] //=.
rewrite [1%:M : 'M[K]_(1 + n)]scalar_mx_block block_mxKur block_mxKdl
  block_mxKdr !mxE rV2seq_zeroes cV2seq_zeroes hi.
case: splitP => i // => _.
rewrite [i]ord1 !mxE.
case: splitP => j // => _.
by rewrite [j]ord1 !mxE.
Qed.


(* the translation of ident is ssr 1 matrix *)
Lemma Matrix2mat_ident : forall n, Matrix2mat n n (ident n) = 1%:M.
Proof.
elim => [ | n hi] //=.
- apply/matrixP => i j; rewrite !mxE.
  by case: i.
by rewrite hi seq2rV_zeroes seq2cV_zeroes -scalar_mx_block.
Qed.

Lemma mat2Matrix0n: forall m (M :'M[K]_(0,m)), mat2Matrix M = eM.
Proof.
by move => m M.
Qed.

Lemma mat2Matrix0m: forall n (M: 'M[K]_(n,0)), mat2Matrix M = eM.
Proof.
by case.
Qed.

Lemma mat2Matrix_block : forall (n p:nat) a (l: 'rV[K]_n) (c: 'cV[K]_p)
  (M: 'M[K]_(p,n)),
  mat2Matrix (block_mx a%:M l c M) =
  cM a (rV2seq l) (cV2seq c) (mat2Matrix M).
Proof.
move => n p a l c M /=.
rewrite block_mxKur block_mxKdl block_mxKdr !mxE.
case :splitP => i // _.
rewrite [i]ord1 !mxE.
case: splitP => j // _.
by rewrite [j]ord1 !mxE mulr1n.
Qed.

Lemma glueCe : forall c, match c with
 | [::] => glue c eM = eM
 | hd :: tl => glue c eM = cM hd [::] tl eM
end.
Proof.
by case.
Qed.

Lemma cV2seqS : forall n a (c: 'cV[K]_n),
  cV2seq (col_mx a%:M c) = a :: cV2seq c.
Proof.
rewrite /cV2seq => [[ | n]] a c.
- rewrite [c]flatmx0 /=.
  apply/(@eq_from_nth _ 0).
  + by rewrite size_map size_enum_ord /= size_map size_enum_ord.
  move => i.
  rewrite size_map size_enum_ord => hi.
  rewrite (@nth_map _ 0) ?size_enum_ord // !mxE.
  case: splitP => j.
  + rewrite [j]ord1 (@nth_ord_enum (1 + 0) 0 (Ordinal hi)) /= => -> /=.
    by rewrite !mxE mulr1n.
  by case: j.
apply/(@eq_from_nth _ 0).
- by rewrite size_map size_enum_ord /= size_map size_enum_ord.
move => i.
rewrite size_map size_enum_ord => hi.
rewrite (@nth_map _ 0) ?size_enum_ord // !mxE.
case: splitP => j.
- rewrite [j]ord1 (@nth_ord_enum (1 + n.+1) 0 (Ordinal hi)) /= => -> /=.
  by rewrite !mxE mulr1n.
rewrite (@nth_ord_enum _ _ (Ordinal hi)) /= => -> /=.
rewrite (@nth_map (ordinal n.+1) 0) ?size_enum_ord //.
by rewrite nth_ord_enum.
Qed.

Lemma rV2seqS : forall n a (r: 'rV[K]_n),
  rV2seq (row_mx a%:M r) = a :: rV2seq r.
Proof.
rewrite /rV2seq => [[ | n]] a r.
- rewrite [r]thinmx0 /=.
  apply/(@eq_from_nth _ 0).
  + by rewrite size_map size_enum_ord /= size_map size_enum_ord.
  move => i.
  rewrite size_map size_enum_ord => hi.
  rewrite (@nth_map _ 0) ?size_enum_ord // !mxE.
  case: splitP => j.
  + rewrite [j]ord1 (@nth_ord_enum (1 + 0) 0 (Ordinal hi)) /= => -> /=.
    by rewrite !mxE mulr1n.
  by case: j.
apply/(@eq_from_nth _ 0).
- by rewrite size_map size_enum_ord /= size_map size_enum_ord.
move => i.
rewrite size_map size_enum_ord => hi.
rewrite (@nth_map _ 0) ?size_enum_ord // !mxE.
case: splitP => j.
- rewrite [j]ord1 (@nth_ord_enum (1 + n.+1) 0 (Ordinal hi)) /= => -> /=.
  by rewrite !mxE mulr1n.
rewrite (@nth_ord_enum _ _ (Ordinal hi)) /= => -> /=.
rewrite (@nth_map (ordinal n.+1) 0) ?size_enum_ord //.
by rewrite nth_ord_enum.
Qed.


(* Correctness of all the function we use, using the same methodology
   every time:
   - reflect the ssr data into the executable ones
   - compute
   - check the result is correct
*)
Lemma glueE : forall n m (M : 'M[K]_(n,m)) (c: 'cV[K]_n),
  mat2Matrix (row_mx c M) = glue (cV2seq c) (mat2Matrix M).
Proof.
elim => [ | n hi] [ | m] M.
- move => c.
  by rewrite [M]flatmx0 [c]flatmx0 cV2seq_zeroes row_mx0 !mat2Matrix0n.
- move => c.
  by rewrite [M]flatmx0 [c]flatmx0 cV2seq_zeroes row_mx0 !mat2Matrix0n.
- change n.+1 with (1 + n)%N => c.
  rewrite [M]thinmx0 !mat2Matrix0m -{2}[c]vsubmxK [usubmx c]mx11_scalar.
  rewrite cV2seqS /=.
  f_equal.
  + rewrite !mxE lshift0.
    by case: splitP => i //; rewrite [i]ord1.
  + apply size0nil.
    by rewrite size_rV2seq.
  + f_equal.
    apply/colP => i; rewrite !mxE /= rshift1 lshift0.
    by case: splitP => j //; rewrite [j]ord1.
  by rewrite mat2Matrix0m.
move: M.
change m.+1 with (1 + m)%N.
change n.+1 with (1 + n)%N => M c.
rewrite -[c]vsubmxK -[M]vsubmxK.
rewrite [usubmx c]mx11_scalar.
rewrite cV2seqS /=.
f_equal.
- rewrite !mxE.
  case: splitP => i //; rewrite [i]ord1 !mxE.
  by case: splitP => j//; rewrite [j]ord1 !mxE mulr1n.
- rewrite -block_mxEh block_mxKur /ursubmx col_mxKu !mxE.
  case: splitP => i //; rewrite [i]ord1 -[usubmx M]hsubmxK row_mxKr
      [lsubmx (usubmx M)]mx11_scalar rV2seqS !mxE => _.
  by case: splitP => j //; rewrite [j]ord1 !lshift0 !mxE mulr1n.
- by rewrite -block_mxEh block_mxKdl.
by rewrite -block_mxEh block_mxKdr -{1}[dsubmx M]hsubmxK hi /dlsubmx
  /drsubmx col_mxKd.
Qed.

Lemma prodsE_ : forall (n:nat) (l : 'rV[K]_n) (c: 'cV[K]_n),
  (prods (rV2seq l) (cV2seq c))%:M =  l *m c.
Proof.
elim => [ | n hi].
- by move => l c; rewrite [c]flatmx0 [l]thinmx0 rV2seq_zeroes
    cV2seq_zeroes mulmx0 /= /prods /= scalar0.
change n.+1 with (1 + n)%N => l c.
rewrite -[l]hsubmxK -[c]vsubmxK [lsubmx l]mx11_scalar [usubmx c]mx11_scalar
  rV2seqS cV2seqS /prods /= mul_row_col -hi.
apply/matrixP => i j; rewrite [i]ord1 [j]ord1 !mxE !lshift0
  !big_ord_recl !big_ord0 mulr1n addr0.
congr (_ + _).
by rewrite !mxE !mulr1n.
Qed.

Lemma prodsE  : forall (n:nat) (l : 'rV[K]_n) (c: 'cV[K]_n),
  prods (rV2seq l) (cV2seq c) =  (l *m c) 0 0.
Proof.
move => n l c.
by rewrite -prodsE_ !mxE mulr1n.
Qed.

Lemma addsrVE : forall (n:nat),
 {morph (@rV2seq n) : l1 l2 / l1 + l2 >-> (adds l1 l2) }.
Proof.
elim => [ | n hi].
- by move => l1 l2; rewrite [l1]thinmx0 [l2]thinmx0 !rV2seq_zeroes addr0 /=
    rV2seq_zeroes.
change n.+1 with (1 + n)%N => l1 l2.
by rewrite -[l1]hsubmxK -[l2]hsubmxK [lsubmx l1]mx11_scalar
    [lsubmx l2]mx11_scalar !rV2seqS /= add_row_mx -scalar_add  rV2seqS
    hi.
Qed.


Lemma addscVE: forall (n:nat),
 {morph (@cV2seq n) : c1 c2 / c1 + c2 >-> (adds c1 c2)}.
Proof.
elim => [ | n hi].
- by move => c1 c2; rewrite [c1]flatmx0 [c2]flatmx0 !cV2seq_zeroes addr0 /=
    cV2seq_zeroes.
change n.+1 with (1 + n)%N => c1 c2.
by rewrite -[c1]vsubmxK -[c2]vsubmxK [usubmx c1]mx11_scalar
    [usubmx c2]mx11_scalar !cV2seqS /= add_col_mx -scalar_add  cV2seqS
    hi.
Qed.

Lemma multErVE : forall (n:nat) (a : K),
 {morph (@rV2seq n) : l / a *: l >-> multEV a l}.
Proof.
elim => [ | n hi] a.
- by move => l; rewrite [l]thinmx0 scaler0 rV2seq_zeroes.
change (n.+1) with (1 + n)%N => l.
by rewrite -[l]hsubmxK [lsubmx l]mx11_scalar scale_row_mx
    [a *: ((lsubmx l) 0 0)%:M]mx11_scalar !rV2seqS /= hi !mxE mulr1n.
Qed.

Lemma multEcVE : forall (n:nat) (a:K),
 {morph (@cV2seq n) : c / a *: c >-> multEV a c}.
Proof.
elim => [ | n hi] a.
- by move => c; rewrite [c]flatmx0 scaler0 cV2seq_zeroes.
change (n.+1) with (1 + n)%N => c.
by rewrite -[c]vsubmxK [usubmx c]mx11_scalar scale_col_mx
    [a *: ((usubmx c) 0 0)%:M]mx11_scalar !cV2seqS /= hi !mxE mulr1n.
Qed.

Lemma adds_nil: forall l, adds l [::] = l.
Proof. by case. Qed.

Lemma adds0 : forall l n, adds l (zeroes n) = l.
Proof.
elim => [ | hd tl hi] [ | n] //=.
by rewrite addr0 hi.
Qed.

Lemma multLME: forall (n m:nat) (l: 'rV[K]_(1 + n)) (A: 'M[K]_((1+n),m)),
  rV2seq (l *m A) = multLM (rV2seq l) (mat2Matrix A).
Proof.
elim => [ | n hi] [ | m].
- move => l A; rewrite [A]thinmx0 mulmx0 rV2seq_zeroes mat2Matrix0m /=.
  by case (rV2seq l).
- change m.+1 with (1 + m)%N => l A.
  rewrite -[l]hsubmxK -[A]submxK [lsubmx l]mx11_scalar
    [ulsubmx A]mx11_scalar rV2seqS mul_row_col mat2Matrix_block /=.
  rewrite mul_scalar_mx scale_row_mx mul_mx_row add_row_mx.
  rewrite [(lsubmx l) 0 0 *: _ + _ *m _]mx11_scalar rV2seqS.
  congr (_ :: _).
  + by rewrite prodsE !mxE !lshift0 mulr1n.
  by rewrite addsrVE multErVE [rsubmx l]thinmx0 mul0mx
      !rV2seq_zeroes /= adds_nil adds0.
- move => l A; rewrite [A]thinmx0 mulmx0 mat2Matrix0m rV2seq_zeroes.
  by case (rV2seq l).
change m.+1 with (1 + m)%N => l A.
rewrite -[l]hsubmxK -[A]submxK [lsubmx l]mx11_scalar
    [ulsubmx A]mx11_scalar rV2seqS mul_row_col mat2Matrix_block.
rewrite mul_scalar_mx scale_row_mx mul_mx_row add_row_mx.
rewrite [(lsubmx l) 0 0 *: _ + _ *m _]mx11_scalar rV2seqS.
rewrite addsrVE hi /= multErVE.
by rewrite prodsE !mxE !lshift0 mulr1n.
Qed.

Lemma multMCnil: forall M, multMC M [::] =  [::].
Proof. by case. Qed.

Lemma map_mulC : forall a (l : seq K),
 map ( *%R^~ a) l = map [eta *%R a] l.
Proof.
move => a; elim => [ | hd tl hi] //=.
by rewrite mulrC hi.
Qed.

Lemma multMCE: forall (n m:nat) (A: 'M[K]_(n,1+m)) (c: 'cV[K]_(1 + m)),
  cV2seq (A *m c) = multMC (mat2Matrix A) (cV2seq c).
Proof.
elim => [ | n hi] [ | m].
- by move => A c; rewrite [A]flatmx0 mul0mx mat2Matrix0n /= cV2seq_zeroes.
- by move => A c; rewrite [A]flatmx0 mul0mx mat2Matrix0n cV2seq_zeroes.
- change n.+1 with (1 + n)%N => A c.
  by rewrite -[c]vsubmxK -[A]submxK [usubmx c]mx11_scalar
    [ulsubmx A]mx11_scalar cV2seqS mat2Matrix_block /= block_mxEh
    mul_row_col [dsubmx c]flatmx0 [ursubmx A]thinmx0 mulmx0 addr0
    cV2seq_zeroes rV2seq_zeroes mul_col_mx multMCnil adds_nil -scalar_mxM
    cV2seqS /prods /= addr0 mul_mx_scalar multEcVE map_mulC.
change n.+1 with (1 + n)%N => A c.
rewrite -[c]vsubmxK -[A]submxK [usubmx c]mx11_scalar [ulsubmx A]mx11_scalar
  cV2seqS mat2Matrix_block block_mxEh mul_row_col.
rewrite !mul_col_mx mul_mx_scalar scale_scalar_mx add_col_mx /=.
rewrite [((usubmx c) 0 0 * _ 0 0)%:M + _ *m _]mx11_scalar cV2seqS.
congr (_ :: _).
- by rewrite prodsE !mxE !lshift0 mulr1n mulrC.
by rewrite addscVE mul_mx_scalar multEcVE map_mulC hi.
Qed.

Lemma add_ursubmx : forall (n m p q:nat) (M N: 'M[K]_(n+m,p+q)),
  ursubmx (M + N) = ursubmx M + ursubmx N.
Proof.
move => n m p q M N.
by apply/matrixP => i j; rewrite !mxE.
Qed.

Lemma add_drsubmx : forall (n m p q:nat) (M N: 'M[K]_(n+m,p+q)),
  drsubmx (M + N) = drsubmx M + drsubmx N.
Proof.
move => n m p q M N.
by apply/matrixP => i j; rewrite !mxE.
Qed.

Lemma add_dlsubmx : forall (n m p q:nat) (M N: 'M[K]_(n+m,p+q)),
  dlsubmx (M + N) = dlsubmx M + dlsubmx N.
Proof.
move => n m p q M N.
by apply/matrixP => i j; rewrite !mxE.
Qed.

Lemma addE : forall (n m : nat),
  {morph (@mat2Matrix n m) : A B / A + B >-> add A B}.
Proof.
elim => [ | n hi] [ | m].
- by move => A B; rewrite [A]thinmx0 [B]thinmx0 addr0.
- by move => A B; rewrite [A]flatmx0 [B]flatmx0 addr0.
- by move => A B; rewrite [A]thinmx0 [B]thinmx0 addr0.
rewrite -[n.+1]/(1 + n)%N -[m.+1]/(1 + m)%N => A B.
symmetry.
rewrite -{1}[A]submxK -{1}[B]submxK [ulsubmx A]mx11_scalar
  [ulsubmx B]mx11_scalar !mat2Matrix_block /= -addsrVE -addscVE -hi
  add_ursubmx add_drsubmx add_dlsubmx.
by rewrite !mxE !lshift0.
Qed.

Lemma multsE: forall (n m:nat) (c: 'cV[K]_n) (l: 'rV[K]_m),
  mat2Matrix (c *m l) = mults (cV2seq c) (rV2seq l).
Proof.
elim => [ | n hi] [ | m].
- move => c l; rewrite [c]flatmx0 [l]thinmx0 mulmx0.
  by rewrite cV2seq_zeroes rV2seq_zeroes mat2Matrix0n.
- move => c l; rewrite [c]flatmx0 mul0mx.
  by rewrite cV2seq_zeroes mat2Matrix0n.
- move => c l; rewrite [l]thinmx0 mulmx0.
  rewrite rV2seq_zeroes mat2Matrix0m.
  by case (cV2seq c).
change n.+1 with (1 + n)%N.
change m.+1 with (1 + m)%N => c l.
by rewrite -[c]vsubmxK -[l]hsubmxK [usubmx c]mx11_scalar
  [lsubmx l]mx11_scalar mul_col_row -scalar_mxM cV2seqS rV2seqS
  mat2Matrix_block /= hi mul_mx_scalar mul_scalar_mx multErVE
  multEcVE map_mulC.
Qed.


Lemma multLM_eM: forall l, multLM l eM = [::].
Proof. by case. Qed.

Lemma add_eM: forall M, add M eM = M.
Proof. by case. Qed.

Lemma mult_unfold : forall a1 l1 c1 m1 a2 l2 c2 m2,
  mult (cM a1 l1 c1 m1) (cM a2 l2 c2 m2) =
          cM (a1 * a2 + prods l1 c2) (adds (multEV a1 l2) (multLM l1 m2))
            (adds (multEV a2 c1) (multMC m1 c2))
            (add (mults c1 l2) (mult m1 m2)).
Proof. by []. Qed.

Lemma ex_simpl_line_unfold : forall p hd tl,
 ex_simpl_line (1 + p) (hd :: tl) =
     if hd == 0
          then
           let '(a, E) := ex_simpl_line p tl in
                let E1 := cM 1 (zeroes (size tl)) (zeroes (size tl)) E in
                let p0 :=
                  cM 0 (zeroes (size tl).-1) (zeroes (size tl).-1)
                    (ident (size tl).-1) in
                let P1 :=
                  cM 0 (1 :: zeroes (size tl).-1) (-1 :: zeroes (size tl).-1)
                    p0 in
                (- a, mult E1 P1)
          else
           (hd,
           cM 1 (map (fun x : K => - x / hd) tl) (zeroes (size tl))
             (ident (size tl))).
Proof. by []. Qed.

Lemma mult_eM : forall M, mult M eM = eM.
Proof. by case. Qed.

(* Our main tool: matrix multiplication is correct *)
Lemma multE : forall (n m p: nat) (A : 'M[K]_(n,1 + m))
  (B : 'M[K]_(1+m,1+p)),
  mat2Matrix (A *m B) =  mult (mat2Matrix A) (mat2Matrix B).
Proof.
elim => [ | n hi].
- by move => m p A B; rewrite [A]flatmx0 !mat2Matrix0n.
case => [ | m].
- move => p.
  change n.+1 with (1 + n)%N => A B.
  rewrite -[A]submxK -[B]submxK [ulsubmx A]mx11_scalar [drsubmx B]flatmx0
    [ulsubmx B]mx11_scalar [drsubmx A]thinmx0 !mat2Matrix_block mulmx_block.
  rewrite [_ *m _ + _ *m _]mx11_scalar mat2Matrix_block /= !mulmx0 mul0mx
    !addr0 mult_eM add_eM prodsE multLM_eM adds_nil.
  f_equal.
  + by rewrite !mxE !big_ord_recl !big_ord0 !mxE !lshift0 !mulr1n !addr0.
  + by rewrite mul_scalar_mx multErVE.
  + by rewrite mat2Matrix0m /= adds_nil mul_mx_scalar multEcVE.
  by rewrite multsE.
case => [ | p].
- change (1 + m.+1)%N with (1 + (1 + m))%N.
  change n.+1 with (1 + n)%N => A B.
  rewrite -[A]submxK -[B]submxK [ulsubmx A]mx11_scalar [drsubmx B]thinmx0
    [ulsubmx B]mx11_scalar !mat2Matrix_block mulmx_block !mulmx0 !addr0.
  rewrite [_ *m _ + _ *m _]mx11_scalar mat2Matrix_block /= prodsE
   mult_eM multLM_eM adds_nil add_eM.
  f_equal.
  + by rewrite !mxE !big_ord_recl !big_ord0 !mxE !lshift0 !mulr1n !addr0.
  + by rewrite mul_scalar_mx multErVE.
  + by rewrite -multEcVE -multMCE -addscVE mul_mx_scalar.
  by rewrite multsE.
change n.+1 with (1 + n)%N.
change (1 + m.+1)%N with (1 + (1 + m))%N.
change (1 + p.+1)%N with (1 + (1 + p))%N => A B.
rewrite -[A]submxK -[B]submxK mulmx_block [ulsubmx A]mx11_scalar
    [ulsubmx B]mx11_scalar [_ *m _ + _]mx11_scalar !mat2Matrix_block
    mult_unfold prodsE -multErVE -multEcVE -multLME -multMCE
    -multsE addsrVE addscVE !mul_scalar_mx !mul_mx_scalar addE hi.
by rewrite !mxE !big_ord_recl !mxE !lshift0 !rshift1 !mulr1n.
Qed.

Lemma ex_simpl_line_ok1 : forall n (l : 'rV[K]_(1 + n)),
  (ex_simpl_line n (rV2seq l)).1 = (simpl_line l).1.
Proof.
elim => [ | n hi].
- move => l; rewrite -[l]hsubmxK [lsubmx l]mx11_scalar [rsubmx l]thinmx0
  rV2seqS rV2seq_zeroes /= !mxE lshift0.
  by case: splitP => i//; rewrite [i]ord1 !mxE mulr1n.
change (1 + n.+1)%N with (1 + (1 + n))%N => l.
rewrite -[l]hsubmxK [lsubmx l]mx11_scalar rV2seqS ex_simpl_line_unfold.
case: ifP => h.
- rewrite (eqP h).
  move: (hi (rsubmx l)).
  case: ex_simpl_line => a U /= ->.
  rewrite !mxE lshift0; case : splitP => i //.
  rewrite [i]ord1 !mxE !mulr1n eqxx row_mxKr.
  by case: simpl_line => b V /=.
by rewrite /simpl_line row_mxKl ![((lsubmx l) 0 0)%:M 0 0]mxE mulr1n h /=.
Qed.


Lemma aux : forall n l,
   foldr +%R 0
     (map (fun xy : K * K => let (x, y) := xy in x * y)
        (zip l (zeroes n))) = 0.
Proof.
elim => [ | n hi] [ | hd tl] //=.
by rewrite mulr0 add0r hi.
Qed.


Lemma aux' : forall n l,
   foldr +%R 0
     (map (fun xy : K * K => let (x, y) := xy in x * y)
        (zip (zeroes n) l)) = 0.
Proof.
elim => [ | n hi] [ | hd tl] //=.
by rewrite mul0r add0r hi.
Qed.


Lemma prods0 : forall n l, prods l (zeroes n) = 0.
Proof.
rewrite /prods => n l; by rewrite aux.
Qed.

Lemma prod0s : forall n l, prods (zeroes n) l = 0.
Proof.
rewrite /prods => n l; by rewrite aux'.
Qed.

Lemma rV2seq0 : forall n (l: 'rV[K]_n),
 (forallb i, l 0 i == 0) = (rV2seq l == zeroes n).
Proof.
elim => [ | n hi].
- move => l; rewrite [l]thinmx0 rV2seq_zeroes eqxx.
  apply/forallP => x.
  by rewrite !mxE.
change n.+1 with (1 + n)%N => l.
rewrite -[l]hsubmxK [lsubmx l]mx11_scalar rV2seqS.
apply/forallP/idP.
- move => h.
  move: (h 0); rewrite !mxE.
  case: splitP => i //; rewrite [i]ord1 !lshift0 !mxE mulr1n => _ h'.
  rewrite (eqP h') /= ; apply/eqP; congr (_ :: _).
  apply/eqP; rewrite -hi.
  apply/forallP => x.
  move : (h (lift 0 x)); rewrite !mxE rshift1.
  case: splitP => j.
  + by rewrite [j]ord1.
  move/eqP; rewrite eqSS; move/eqP => h''.
  have -> : x = j by apply/ord_inj.
  by rewrite !mxE rshift1.
move/eqP => /= h.
injection h => htl hhd x; clear h.
rewrite !mxE.
case: splitP => i.
- move: hhd.
  by rewrite [i]ord1 !mxE lshift0 mulr1n => ->.
move => hx.
move: (hi (rsubmx l)); rewrite htl eqxx.
move/forallP.
by apply.
Qed.

Lemma cV2seq0 : forall n (c: 'cV[K]_n),
 (forallb i, c i 0 == 0) = (cV2seq c == zeroes n).
Proof.
elim => [ | n hi].
- move => c; rewrite [c]flatmx0 cV2seq_zeroes eqxx.
  apply/forallP => x.
  by rewrite !mxE.
change n.+1 with (1 + n)%N => c.
rewrite -[c]vsubmxK [usubmx c]mx11_scalar cV2seqS.
apply/forallP/idP.
- move => h.
  move: (h 0); rewrite !mxE.
  case: splitP => i //; rewrite [i]ord1 !lshift0 !mxE mulr1n => _ h'.
  rewrite (eqP h') /= ; apply/eqP; congr (_ :: _).
  apply/eqP; rewrite -hi.
  apply/forallP => x.
  move : (h (lift 0 x)); rewrite !mxE rshift1.
  case: splitP => j.
  + by rewrite [j]ord1.
  move/eqP; rewrite eqSS; move/eqP => h''.
  have -> : x = j by apply/ord_inj.
  by rewrite !mxE rshift1.
move/eqP => /= h.
injection h => htl hhd x; clear h.
rewrite !mxE.
case: splitP => i.
- move: hhd.
  by rewrite [i]ord1 !mxE lshift0 mulr1n => ->.
move => hx.
move: (hi (dsubmx c)); rewrite htl eqxx.
move/forallP.
by apply.
Qed.

Lemma tperm01E : forall n,
  let p := cM 0 (zeroes n)(zeroes n) (ident n) in
    let P1 := cM 0 (1 :: zeroes n)(1 :: zeroes n) p in
      @mat2Matrix n.+2 n.+2 (tperm_mx 0 1) = P1.
Proof.
move => n.
rewrite -[tperm_mx 0 1](@submxK _ 1 n.+1 1 n.+1) /=.
f_equal.
- rewrite !mxE.
  case: splitP => i //; rewrite [i]ord1 !mxE.
  by case: splitP => j //; rewrite [j]ord1 !mxE !lshift0 tpermL=> _ _.
- rewrite block_mxKur -[ursubmx _](@hsubmxK _ 1 1 n) [lsubmx _]mx11_scalar
   rV2seqS !mxE !lshift0 rshift1 tpermL.
  congr (_ :: _).
  apply/eqP.
  rewrite -rV2seq0.
  apply/forallP => i.
  by rewrite !mxE lshift0 !rshift1 tpermL.
- rewrite block_mxKdl -[dlsubmx _](@vsubmxK _ 1 n 1) [usubmx _]mx11_scalar
   cV2seqS !mxE !lshift0 rshift1.
  congr (_ :: _).
  + have -> : (@lift (S (S n)) 0 0) = 1 by apply/ord_inj.
    by rewrite tpermR.
  apply/eqP.
  rewrite -cV2seq0.
  apply/forallP => i.
  by rewrite !mxE lshift0 !rshift1 tpermD.
f_equal.
- rewrite block_mxKdr !mxE !rshift1.
  have -> : (@lift (S (S n)) 0 0) = 1 by apply/ord_inj.
  by rewrite tpermR.
- rewrite block_mxKdr.
  apply/eqP;  rewrite -rV2seq0.
  apply/forallP => x.
  rewrite !mxE !rshift1 !lshift0.
  have -> : (@lift (S (S n)) 0 0) = 1 by apply/ord_inj.
  by rewrite tpermR.
- rewrite block_mxKdr.
  apply/eqP;  rewrite -cV2seq0.
  apply/forallP => x.
  rewrite !mxE !rshift1 !lshift0.
  have -> : (@lift (S (S n)) 0 0) = 1 by apply/ord_inj.
  by rewrite tpermD.
rewrite block_mxKdr.
rewrite -ident2Matrix.
f_equal.
by apply/matrixP => i j; rewrite !mxE !rshift1 tpermD.
Qed.

Lemma liftmx2Matrix : forall n (M: 'M[K]_n),
  mat2Matrix (lift0_mx M) = cM 1 (zeroes n) (zeroes n)  (mat2Matrix M).
Proof.
move => n M; rewrite /lift0_mx [1]mx11_scalar mat2Matrix_block.
by rewrite rV2seq_zeroes cV2seq_zeroes !mxE.
Qed.

Lemma multLM01 : forall n, multLM (zeroes n) (ident n) = zeroes n.
Proof.
elim => [ | n hi] //=.
by rewrite mul0r add0r prods0 map0 hi adds0 size_nseq.
Qed.

Lemma multMC10 : forall n, multMC (ident n) (zeroes n) = zeroes n.
Proof.
elim => [ | n hi] //=.
by rewrite mulr0 add0r prods0 map_mulC map0 hi adds0 size_nseq.
Qed.

Lemma multEV0 : forall a n, multEV a (zeroes n) = zeroes n.
Proof.
move => a; elim => [ | n hi] //=.
by rewrite mulr0 hi.
Qed.

Lemma tool : forall n,
     (add (mults (zeroes n) (zeroes n)) (ident n)) = ident n.
Proof.
elim => [ | n hi] //=.
by rewrite mul0r add0r map0 map_mulC map0 !adds0 size_nseq hi.
Qed.

Lemma mult1 : forall n, mult (ident n) (ident n) = (ident n).
Proof.
elim => [ | n hi]  //=.
by rewrite prods0 multEV0 multLM01 multMC10 hi tool mulr1 addr0 adds0.
Qed.


Lemma subproof : forall n (V : 'M_(1 + n)),
   mult (cM 1 (zeroes (1 + n)) (zeroes (1 + n)) (mat2Matrix V))
     (cM 0 (1 :: zeroes n) (-1 :: zeroes n)
        (cM 0 (zeroes n) (zeroes n) (ident n))) =
   mat2Matrix (lift0_mx V *m tperm_mx 0 1 *m
      (@block_mx _ 1 (1 + n) 1 (1 + n) (-1)%:M 0 0 1%:M)).
Proof.
move => n V.
rewrite -!mulmxA !multE liftmx2Matrix tperm01E mat2Matrix_block
    rV2seq_zeroes cV2seq_zeroes [zeroes _]/= !ident2Matrix.
f_equal => /=.
rewrite !mul0r /prods /= !mulr0 !add0r aux !addr0 !mulr1 !multMC10
    !multLM01 !multEV0.
f_equal.
- by rewrite -map_mulC map1 !adds0.
- by rewrite map_mulC map0 !adds0.
by rewrite -map_mulC map1 map_mulC map0 !adds0 size_nseq mult1 !tool.
Qed.

Lemma ex_simpl_line_ok2 : forall n (l : 'rV[K]_(1 + n)),
 (ex_simpl_line n (rV2seq l)).2 = mat2Matrix ((simpl_line l).2).
Proof.
elim => [ | n hi].
- move => l.
  by rewrite -[l]hsubmxK [rsubmx l]thinmx0 [lsubmx l]mx11_scalar rV2seqS /=
   !mxE [ursubmx _]thinmx0 [dlsubmx _]flatmx0 rV2seq_zeroes cV2seq_zeroes.
change (1 + n.+1)%N with (1 + (1 + n))%N.
move => l.
rewrite -[l]hsubmxK [lsubmx l]mx11_scalar rV2seqS ex_simpl_line_unfold.
cbv zeta.
rewrite [simpl_line _]/= row_mxKl.
case: ifP => h.
- move: (hi (rsubmx l)).
  rewrite (eqP h) !mxE mulr1n eqxx row_mxKr /snd.
  case: ex_simpl_line => a U.
  case: simpl_line => b V.
  rewrite size_rV2seq [(1 + n).-1]/= => ->.
  by apply subproof.
rewrite row_mxKr ![((lsubmx l) 0 0)%:M 0 0]mxE mulr1n h /snd
  mat2Matrix_block multErVE cV2seq_zeroes size_rV2seq ident2Matrix.
f_equal.
apply eq_map => x.
by rewrite mulrC mulrN mulNr.
Qed.

(* the simpl_line function is correct *)
Lemma ex_simpl_lineE : forall n (l : 'rV[K]_(1 + n)),
  let (a1,E1) := ex_simpl_line n (rV2seq l) in
  let (a2,E2) := simpl_line l in
    a1 = a2 /\ E1 = mat2Matrix E2.
Proof.
move => n l.
move: (ex_simpl_line_ok1 l).
move: (ex_simpl_line_ok2 l).
case: ex_simpl_line => a1 E1.
by case: simpl_line =>  a2 E2 /=.
Qed.

Lemma ex_Rank_unfold : forall p hd l c M,
  ex_Rank (S p) (cM hd l c M) =
    let: (a, E) := ex_simpl_line (size l) (hd :: l) in
    let B := mult (cM hd l c M) E in
    if a == 0 then
      ex_Rank p (ex_dsubmx B)
    else (1 + ex_Rank p (ex_drsubmx B))%N.
Proof. by case. Qed.


Lemma ex_dsubmxE : forall n m (M: 'M[K]_(1+n,m)),
  mat2Matrix (dsubmx M) = ex_dsubmx (mat2Matrix M).
Proof.
move => n [ | m].
- by move => M; rewrite [M]thinmx0 [dsubmx 0]thinmx0 !mat2Matrix0m.
change m.+1 with (1 + m)%N => M.
by rewrite -[M]submxK col_mxKd [ulsubmx M]mx11_scalar mat2Matrix_block
 /= glueE.
Qed.

Lemma ex_drsubmxE : forall n m (M: 'M[K]_(1+n,1+m)),
  mat2Matrix (drsubmx M) = ex_drsubmx (mat2Matrix M).
Proof.
move => n m M.
by rewrite -[M]submxK block_mxKdr [ulsubmx M]mx11_scalar mat2Matrix_block.
Qed.


Lemma  Rank_unfold : forall n m b (l: 'rV[K]_n) (c: 'cV[K]_m)
  (M: 'M[K]_(m,n)), Rank (block_mx b%:M l c M) =
   let '(a, E) := simpl_line (row_mx b%:M l) in
     let B := (block_mx b%:M l c M) *m E in
       if a == 0
         then Rank (dsubmx B)
         else (1 + Rank (drsubmx B))%N.
Proof. move => n m b l c M /=. by rewrite col_mxKu. Qed.

Lemma subproof2 : forall n m (M : 'M_(1 + (1 + n), 1 + (1 + m))),
  (ulsubmx M) 0 0 != 0 ->
  cM 1  (map (fun x : K => - x / (ulsubmx M) 0 0) (rV2seq (ursubmx M)))
         (zeroes (1 + m)) (ident (1 + m)) =
  mat2Matrix (block_mx 1%:M (- ((ulsubmx M) 0 0)^-1 *: ursubmx M) 0 1%:M).
Proof.
move => n m M h; rewrite mat2Matrix_block ident2Matrix cV2seq_zeroes
 multErVE.
f_equal.
apply eq_map => x.
by rewrite mulNr -div1r mulrC mulNr.
Qed.

Lemma multEV1: forall l, multEV 1 l = l.
Proof.
elim => [ | hd tl hi] //=.
by rewrite hi mul1r.
Qed.

Lemma ex_Rank_ok : forall n m (M: 'M[K]_(1 + n,1 + m)),
  ex_Rank (1 + n) (mat2Matrix M) = Rank M.
Proof.
elim => [ | n hi] [ | m].
- move => M /=.
  rewrite [ursubmx M]thinmx0 size_rV2seq /= !mxE lshift0.
  case: ifP => h.
  + by case : glue.
  by case: add.
- change (1 + m.+1)%N with (1 + (1 + m))%N => M.
  rewrite -[M]submxK [ulsubmx M]mx11_scalar [drsubmx M]flatmx0
    [dlsubmx M]flatmx0 mat2Matrix_block cV2seq_zeroes mat2Matrix0n
    ex_Rank_unfold Rank_unfold size_rV2seq ex_simpl_line_unfold
    [simpl_line _]/= row_mxKl ![((ulsubmx M) 0 0)%:M 0 0]mxE mulr1n
    row_mxKr.
   move: (ex_simpl_lineE (ursubmx M)).
   case: ex_simpl_line => a U.
   case: simpl_line => b V [].
   rewrite /fst /snd => -> ->.
   case: ifP => h.
   + by rewrite (eqP h) oppr_eq0 size_rV2seq  [(1 + m).-1]/= ![zeroes _]/=.
   by rewrite h.
- change (1 + n.+1)%N with (1 + (1 + n))%N => M.
  rewrite -[M]submxK [ulsubmx M]mx11_scalar [drsubmx M]thinmx0
    [ursubmx M]thinmx0 mat2Matrix_block rV2seq_zeroes mat2Matrix0m
    ex_Rank_unfold Rank_unfold [simpl_line _]/= size_nseq
    [ex_simpl_line _ _]/= mulmx1.
  cbv zeta.
  rewrite ![(row_mx ((ulsubmx M) 0 0)%:M 0) 0 0]mxE.
  case: ifP => h.
  + rewrite (eqP h).
    case: splitP => i //; rewrite [i]ord1 !mxE mulr1n eqxx; clear i => _.
    rewrite -hi ex_dsubmxE mat2Matrix_block rV2seq_zeroes mat2Matrix0m.
    f_equal; f_equal => /=.
    rewrite /prods /= mul0r add0r multEV1 adds_nil add_eM.
    by case (cV2seq (dlsubmx M)).
  case : splitP => i //; rewrite [i]ord1 -mx11_scalar h; clear i => _.
  rewrite block_mxKdr /= add_eM.
  by case : (cV2seq (dlsubmx M)).
change (1 + n.+1)%N with (1 + (1 + n))%N.
change (1 + m.+1)%N with (1 + (1 + m))%N => M.
rewrite -[M]submxK [ulsubmx M]mx11_scalar mat2Matrix_block ex_Rank_unfold
  Rank_unfold size_rV2seq ex_simpl_line_unfold [simpl_line _]/= row_mxKl
  row_mxKr.
move: (ex_simpl_lineE (ursubmx M)).
case: ex_simpl_line => a U.
case: simpl_line => b V [].
rewrite /fst /snd => -> ->.
case: ifP => h.
- rewrite (eqP h) ![0%:M 0 0]mxE  mulr1n eqxx oppr_eq0 size_rV2seq
  [(1 + m).-1]/= [zeroes _]/=.
  case: ifP => h'.
  + by rewrite -mat2Matrix_block subproof -multE -ex_dsubmxE hi.
  by rewrite -mat2Matrix_block subproof -multE -ex_drsubmxE hi.
by rewrite ![((ulsubmx M) 0 0)%:M 0 0]mxE mulr1n !h -mat2Matrix_block
 size_rV2seq subproof2 ?h // -multE -ex_drsubmxE hi.
Qed.

(* Rank is ok *)
Lemma ex_RankE : forall n m (M: 'M[K]_(n,m)),
  ex_Rank n (mat2Matrix M) = Rank M.
Proof.
case => [ | n].
- by move => m M; rewrite [M]flatmx0.
case => [ | m].
- by move => M; rewrite [M]thinmx0.
change n.+1 with (1 + n)%N.
change m.+1 with (1 + m)%N => M.
by apply ex_Rank_ok.
Qed.

(* Mok M n m means that M is a Matrix of rectangular n x m shape *)
Fixpoint Mok (M: Matrix) (n m:nat) : bool :=
 match M with
  | eM => (n == 0)%N || (m == 0)%N
  | cM _ l c mat => match n,m with
                    | O,_ => false
                    | _,O => false
                    | S p, S q =>
     [&& size l == q , size c == p & Mok mat p q]
 end
end.

Lemma mat2Matrix_ok : forall (n m:nat) (A: 'M[K]_(n,m)),
 Mok (mat2Matrix A) n m .
Proof.
elim => [ | n hi] [ | m] //= A.
by rewrite size_rV2seq size_cV2seq hi !eq_refl.
Qed.

Lemma Mok0n: forall A m, Mok A 0 m -> A = eM.
Proof. by case. Qed.

Lemma Mokn0: forall A n, Mok A n 0 -> A = eM.
Proof. by case => [ | a l c mat ] [ | n]. Qed.

Lemma r2r_ok2 : forall n l, size l = n -> rV2seq (seq2rV n l) = l.
Proof.
elim => [ | n hi] //=.
- move => l; move/size0nil => ->.
  by rewrite seq2rV_zeroes rV2seq_zeroes.
case => [ | hd tl] //=.
move/eqP ; rewrite eqSS seq2rVS rV2seqS.
move/eqP => h; by rewrite hi.
Qed.


Lemma c2c_ok2 : forall n l, size l = n -> cV2seq (seq2cV n l) = l.
Proof.
elim => [ | n hi] //=.
- move => l; move/size0nil => ->.
  by rewrite seq2cV_zeroes cV2seq_zeroes.
case => [ | hd tl] //=.
move/eqP ; rewrite eqSS seq2cVS cV2seqS.
move/eqP => h; by rewrite hi.
Qed.

(* our translations functions are invertible
   and SSR matrices are exactly the set of
   rectangular (well formed) Matrix
*)
Lemma m2m_ok2 : forall n m M, Mok M n m ->
  mat2Matrix (Matrix2mat n m M) = M.
Proof.
elim => [ | n hi] /=.
- move => m M.
  by move/Mok0n <-.
case => [ | m] /=.
- move => M.
  by move/Mokn0 <-.
case =>  [ | a l c M] //=.
case/and3P.
move/eqP => h1.
move/eqP => h2 h3.
rewrite block_mxKur block_mxKdl block_mxKdr hi // c2c_ok2 // r2r_ok2 //.
rewrite !mxE.
case: splitP => i //; rewrite [i]ord1 !mxE => _; clear i.
by case: splitP => i //; rewrite [i]ord1 !mxE mulr1n.
Qed.

(* Rank is ok with either translation *)
Lemma ex_RankE2 : forall n m (M: Matrix), Mok M n m ->
  ex_Rank n M = Rank (Matrix2mat n m M).
Proof.
move => n m M hok.
by rewrite -{1}[M](@m2m_ok2 n m M hok) ex_RankE.
Qed.

Lemma ex_Det_unfold: forall (p: nat) hd l c M',
  ex_Det (S p) (cM hd l c M') =
    let: (a, E) := ex_simpl_line p (hd :: l) in
    let B := mult (cM hd l c M') E in
    if a == 0 then
      0
    else (a * ex_Det p (ex_drsubmx B)).
Proof. by  []. Qed.


Lemma Det_unfold : forall n b (l: 'rV[K]_n) (c: 'cV[K]_n) (M: 'M[K]_n),
 Det (block_mx b%:M l c M) =
      let '(a, E) := simpl_line (usubmx (block_mx b%:M l c M)) in
        let B := (block_mx b%:M l c M) *m E in
          if a == 0 then 0 else a * Det (drsubmx B).
Proof. by []. Qed.

Lemma ex_Det_ok: forall n (M: 'M[K]_(1 + n)),
  ex_Det (1 + n) (mat2Matrix M) = Det M.
Proof.
elim => [ | n hi].
- move => M /=.
  rewrite add_eM !mxE lshift0.
  case: ifP => h //.
  by case (cV2seq (dlsubmx M)); rewrite mulr1.
change (1 + n.+1)%N with (1 + (1 + n))%N => M.
rewrite -[M]submxK [ulsubmx M]mx11_scalar mat2Matrix_block ex_Det_unfold
  Det_unfold col_mxKu ex_simpl_line_unfold [simpl_line _]/= row_mxKl
  row_mxKr.
move: (ex_simpl_lineE (ursubmx M)).
case: ex_simpl_line => a U.
case: simpl_line => b V [].
rewrite /fst /snd => -> ->.
case: ifP => h.
- rewrite (eqP h) ![0%:M 0 0]mxE  mulr1n eqxx oppr_eq0 size_rV2seq
    [(1 + n).-1]/= [zeroes _]/=.
  case: ifP => h' //.
  by rewrite -mat2Matrix_block subproof -multE -ex_drsubmxE hi.
by rewrite ![((ulsubmx M) 0 0)%:M 0 0]mxE mulr1n !h -mat2Matrix_block
 size_rV2seq subproof2 ?h // -multE -ex_drsubmxE hi.
Qed.

(* Correctness of the determinant function *)
Lemma ex_DetE: forall n (M: 'M[K]_n),
  ex_Det n (mat2Matrix M) = Det M.
Proof.
case => [ | n].
- move => M.
  by rewrite [M]flatmx0 mat2Matrix0n.
change (n.+1) with (1 + n)%N => M.
by apply ex_Det_ok.
Qed.

Lemma ex_DetE2: forall n M, Mok M n n ->
  ex_Det n M = Det (Matrix2mat n n M).
Proof.
move => n M hok.
by rewrite -{1}[M](@m2m_ok2 n n M hok) ex_DetE.
Qed.


End FieldRank.
