Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq choice fintype.
Require Import div finfun bigop prime binomial ssralg finset fingroup finalg.
Require Import perm zmodp matrix.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section seqmx.

Local Open Scope ring_scope.

Variable mT : ringType.

Section SeqmxDef.

Definition seqmatrix := seq (seq mT).
Definition seqrow := seq mT.

Variables m n : nat.

Definition mkseqmx (f:nat->nat->mT) : seqmatrix :=
  let row i := mkseq (fun j => f i j) n in
  mkseq (fun i => row i) m.

Definition ord_enum_eq n : seq 'I_n := pmap (insub_eq _) (iota 0 n).

Lemma ord_enum_eqE p : ord_enum_eq p = ord_enum p.
Proof. by apply:eq_pmap ; exact:insub_eqE. Qed.

Definition mkseqmx_ord (f:'I_m->'I_n->mT) : seqmatrix :=
  let enum_n := ord_enum_eq n in
  map (fun i => map (f i) enum_n) (ord_enum_eq m).

Definition rowseqmx (M : seqmatrix) i := nth [::] M i.

Definition mx_of_seqmx (M:seqmatrix) :=
  \matrix_(i < m, j < n) nth 0%R (rowseqmx M i) j.

Definition fun_of_seqmx (M:seqmatrix) := fun i j => nth 0%R (rowseqmx M i) j.

Coercion fun_of_seqmx : seqmatrix >-> Funclass.

Lemma fun_of_seqmxE (M:seqmatrix) i j : M i j = nth 0%R (nth [::] M i) j.
Proof. by []. Qed.

(*
Lemma seqmxP (M N : seqmatrix) :
  [/\ size M = size N, forall i, i < size M -> size (rowseqmx M i) = size (rowseqmx N i) &
  forall i j, M i j = N i j] <-> M = N.
Proof.
split=> [[H H' HMN]|H].
  apply:(@eq_from_nth _ [::])=> // i Hi.
  apply:(@eq_from_nth _ 0)=> [|j Hj] ; first exact:(H' _ Hi) ; last exact:HMN.
by split => [|i|i j]; rewrite H.
Qed.
*)

Lemma mkseqmxE f i j : i < m -> j < n -> mkseqmx f i j = f i j.
Proof. by move=> Hi Hj ; rewrite /fun_of_seqmx /rowseqmx !nth_mkseq. Qed.

Definition seqmx_of_mx (M:'M_(m,n)) : seqmatrix :=
  map (fun i => map (fun j => M i j) (enum 'I_n)) (enum 'I_m).

Lemma size_seqmx : forall (M:'M[mT]_(m,n)),
  size (seqmx_of_mx M) = m.
Proof. by move=> M ; rewrite size_map size_enum_ord. Qed.

Lemma seqmxE : forall (M:'M[mT]_(m,n)) (i:'I_m) (j:'I_n),
  (seqmx_of_mx M) i j = M i j.
Proof.
move=> M i j ; rewrite /fun_of_seqmx /rowseqmx.
rewrite (nth_map i); last by rewrite size_enum_ord ltn_ord.
rewrite (nth_map j); last by rewrite size_enum_ord ltn_ord.
by rewrite !nth_ord_enum.
Qed.

End SeqmxDef.

Section Degenerate.

Variables m n : nat.

Lemma seqmx0n (M:'M[mT]_(0,n)) : seqmx_of_mx M = [::].
Proof. by rewrite /seqmx_of_mx (size0nil (size_enum_ord 0)). Qed.

Lemma seqmxm0 (M:'M[mT]_(m,0)) : seqmx_of_mx M = map (fun _ => [::]) (enum 'I_m).
Proof. by rewrite /seqmx_of_mx (size0nil (size_enum_ord 0)). Qed.

End Degenerate.

Section FixedDim.

Variables m n : nat.

(* Should we use a reflect predicate here ? *)
(* Is this lemma useful ?                   *)
Lemma seqmx_eqP: forall (M N:'M[mT]_(m,n)),
  seqmx_of_mx M == seqmx_of_mx N -> M = N.
Proof. by move=> M N H ; apply/matrixP=> i j ; rewrite -!seqmxE (eqP H). Qed.

Lemma size_row_seqmx : forall (M:'M[mT]_(m,n)) i,
  i < m -> size (rowseqmx (seqmx_of_mx M) i) = n.
Proof.
case:m=> [M i|p M i Hi] ; first by rewrite ltn0.
rewrite /rowseqmx (nth_map 0) ; first by rewrite size_map size_enum_ord.
by rewrite size_enum_ord.
Qed.

Lemma size_row_mkseqmx f i : 
  i < m -> size (rowseqmx (mkseqmx m n f) i) = n.
Proof. by move=> Hi ; rewrite /rowseqmx nth_mkseq // size_mkseq. Qed.

Lemma seqmxP (M : seqmatrix) (N : 'M_(m,n)) :
  [/\ m = size M, forall i, i < m -> size (rowseqmx M i) = n &
  forall (i:'I_m) (j:'I_n), M i j = N i j] <-> M = seqmx_of_mx N.
Proof.
split=> [[Hm Hn HMN]|->].
  apply:(@eq_from_nth _ [::])=> [|i Hi] ; first by rewrite size_seqmx.
  rewrite -Hm in Hi ; apply:(@eq_from_nth _ 0)=> [|j] ; rewrite (Hn _ Hi).
    by rewrite size_row_seqmx.
  by move=> Hj; rewrite -!fun_of_seqmxE (HMN (Ordinal Hi) (Ordinal Hj)) -seqmxE.
split=> [|i Hi|i j] ; first by rewrite size_seqmx.
  by rewrite size_row_seqmx.
by rewrite seqmxE.
Qed.

End FixedDim.

Section StrictMatrix.

Variables m' n' : nat.
Local Notation m := m'.+1.
Local Notation n := n'.+1.

Lemma seqmx_of_funE : forall (f : 'I_m -> 'I_n -> mT),
  seqmx_of_mx (\matrix_(i < m, j < n) f i j) = mkseqmx_ord f.
Proof.
move=>f.
rewrite /mkseqmx_ord /seqmx_of_mx. (* /mkseq. -(val_enum_ord m) -[x in _ = x]map_comp. *)
rewrite !ord_enum_eqE [x in map _ x]enumT unlock.
apply:eq_map=> i.
rewrite enumT unlock.
apply:eq_map=> j.
by rewrite mxE.
Qed.

End StrictMatrix.

Section map2.

Variables T1 : Type.
Variables x1 : T1.
Variables T2 : Type.
Variable x2 : T2.
Variable T3 : Type.
Variable x3 : T3.
Variable (f : T1 -> T2 -> T3).

Fixpoint map2 (s1 : seq T1) (s2 : seq T2) :=
  if s1 is x1 :: s1' then
    if s2 is x2 :: s2' then f x1 x2 :: map2 s1' s2' else [::]
  else [::].

Lemma size_map2 (s1 : seq T1) (s2 : seq T2) :
  size (map2 s1 s2) = minn (size s1) (size s2).
Proof.
move:s2; elim:s1=>[s2|t1 s1 IHs] ; first by rewrite min0n.
by case=>[|t2 s2] //= ; rewrite IHs /minn /leq subSS ; case:ifP.
Qed.

Lemma nth_map2 (n : nat) (s1 : seq T1) (s2 : seq T2) :
n < minn (size s1) (size s2) ->
  nth x3 (map2 s1 s2) n = f (nth x1 s1 n) (nth x2 s2 n).
Proof.
move:n s2; elim:s1=> [n s2|t1 s1 IHs]; first by rewrite min0n ltn0.
case=>[|n]; first by case.
by case=> // t2 s2 H ; rewrite /= IHs // ; move:H ; rewrite !leq_minr.
Qed.

Fixpoint foldl2 (f : T3 -> T1 -> T2 -> T3) z (s : seq T1) (t : seq T2) {struct s} :=
  if s is x :: s' then
    if t is y :: t' then foldl2 f (f z x y) s' t' else z
  else z.

End map2.

Section SeqmxOp.

Variables m n p' : nat.
Local Notation p := p'.+1.

Definition map2seqmx (M N : seqmatrix) (f : mT -> mT -> mT) : seqmatrix :=
  map2 (map2 f) M N.

Lemma map2seqmxE (M N : 'M_(m,n)) (f : mT -> mT -> mT) :
  map2seqmx (seqmx_of_mx M) (seqmx_of_mx N) f = seqmx_of_mx (\matrix_(i < m, j < n) f (M i j) (N i j)).
Proof.
apply/seqmxP ; split=>[|i Hi|i j].
    by rewrite size_map2 !size_seqmx minnn.
  rewrite /rowseqmx (nth_map2 [::] [::]) ; last by rewrite !size_seqmx minnn.
  by rewrite size_map2 !size_row_seqmx // minnn.
rewrite fun_of_seqmxE (nth_map2 [::] [::]); last by rewrite !size_seqmx // minnn.
rewrite (nth_map2 0 0) ; last by rewrite !size_row_seqmx // minnn.
by rewrite -!fun_of_seqmxE !seqmxE mxE.
Qed.

(*
Lemma size_row_map2_seqmx (M : 'M_(m,n)) i :
  size (rowseqmx (seqmx_of_mx M) i) = n.
Proof.
rewrite /rowseqmx /seqmx_of_mx.
 (nth_map2 [::] 0).
*)

Definition addseqmx (M N : seqmatrix) : seqmatrix :=
  map2seqmx M N (fun x y => x + y).

Lemma addseqmxE:
  {morph (@seqmx_of_mx m n) : M N / M + N >-> addseqmx M N}.
Proof. by move=> M N ; rewrite /addseqmx map2seqmxE. Qed.

Definition subseqmx (M N : seqmatrix) :=
  map2seqmx M N (fun x y => x - y).

Lemma subseqmxE:
  {morph (@seqmx_of_mx m n) : M N / M - N >-> subseqmx M N}.
Proof.
move=> M N ; rewrite /subseqmx map2seqmxE.
by congr seqmx_of_mx; apply/matrixP=> i j ; rewrite !mxE.
Qed.

Definition mulseqmx (M N: seqmatrix) : seqmatrix :=
  let c i := let row := rowseqmx M i in
    fun j => foldl (fun acc x => (acc.1.+1, x * (N acc.1 j) + acc.2)) (0%N,0) row in
  let m := size (rowseqmx N 0) in
  mkseqmx (size M) m (fun i j => (c i j).2).

Definition transposeseqmx (M : seqmatrix) : seqmatrix :=
  foldr (fun r N => (map2 (fun x s => x::s) r N)) (map (fun _ => [::]) M) M.

Definition mulseqmx2 (M N: seqmatrix) : seqmatrix :=
  let N := transposeseqmx N in
  map (fun r => map (fun c => foldl2 (fun z x y => x * y + z) 0 r c) N) M.

Lemma mulseqmxE (M:'M_(m,p)) (N:'M_(p,n)) :
  mulseqmx (seqmx_of_mx M) (seqmx_of_mx N) = seqmx_of_mx (M *m N).
Proof.
apply/seqmxP; split => [|i Hi|i j]; first by rewrite size_mkseq size_seqmx.
  rewrite size_row_mkseqmx ; first by rewrite size_row_seqmx.
  by rewrite size_seqmx.
rewrite /mulseqmx mxE mkseqmxE ?size_seqmx ?size_row_seqmx //.
set row := rowseqmx (seqmx_of_mx M) i ; set t := foldl _.
have-> : forall l x, (t (l,x) row).2 =
  x + \sum_(0 <= k < size row) nth 0 row k * (seqmx_of_mx N) (l+k)%N j.
  rewrite /t; move:t=> _; elim:row => [l x|t s IHn l x].
    by rewrite big_mkord big_ord0 GRing.addr0.
  rewrite IHn big_nat_recl [_ + x]GRing.addrC GRing.addrA addn0.
  set S1 := \sum_(_ <- _) _ ; set S2 := \sum_(_ <- _) _.
  by have-> : S1 = S2 ; first by apply:eq_bigr=> k _; rewrite addnS.
rewrite GRing.add0r size_row_seqmx // big_mkord.
by apply:eq_bigr=>k _ ; rewrite -!seqmxE.
Qed.

Definition trseqmx (M : seqmatrix) :=
  (foldl (fun acc x => (acc.1.+1, nth 0 x acc.1 + acc.2)) (0%N,0) M).2.

End SeqmxOp.

Lemma mulseqmxEnn : forall m (M:'M_m) (N:'M_m),
  mulseqmx (seqmx_of_mx M) (seqmx_of_mx N) = seqmx_of_mx (M *m N).
Proof. by case=>[M N|m M N]; rewrite ?(flatmx0 M) ?mul0mx ?seqmx0n ?mulseqmxE. Qed.

Section SeqmxRowCol.

Variables m m1 m2 n n1 n2 : nat.

(* Block operations *)
Definition subseqrow i di (s : seqrow) :=
  take di (drop i s).

Definition sub_seqmx i j di dj (M : seqmatrix) :=
  map (subseqrow j dj) (take di (drop i M)).

Definition ulsubseqmx (M : seqmatrix) :=
  map (take m1) (take n1 M).

Definition ursubseqmx (M : seqmatrix) :=
  map (drop m1) (take n1 M).

Definition dlsubseqmx (M : seqmatrix) :=
  map (take m1) (drop n1 M).

Definition drsubseqmx (M : seqmatrix) :=
  map (drop m1) (drop n1 M).

Lemma ulsub_seqmxE : forall (M:'M[mT]_(m1+m2,n1+n2)),
  sub_seqmx 0 0 m1 n1 (seqmx_of_mx M) =
  seqmx_of_mx (ulsubmx M).
Proof.
move=> M ; apply/seqmxP ; split=> [|i Hi|i j].
    rewrite size_map size_take !size_drop size_seqmx !subn0.
    case:m2=> [|p] ; first by rewrite addn0 ltnn.
    by rewrite addnS leq_addr.
  rewrite /rowseqmx (nth_map [::]).
    rewrite size_take !size_drop !nth_take // ?nth_drop ?size_row_seqmx ?subn0.
      case:n2=>[|p] ; first by rewrite addn0 ltnn.
      by rewrite addnS leq_addr.
    by rewrite add0n ; exact:ltn_addr.
  rewrite size_take size_drop size_seqmx subn0.
  case:m2=> [|p] ; first by rewrite addn0 ltnn.
  by rewrite addnS leq_addr.
rewrite /fun_of_seqmx /rowseqmx (nth_map [::]).
  by rewrite /subseqrow !nth_take // !nth_drop !mxE -seqmxE.
rewrite size_take size_drop size_seqmx subn0.
case:m2=> [|p] ; first by rewrite addn0 ltnn.
by rewrite addnS leq_addr.
Qed.

Lemma ursub_seqmxE : forall (M:'M[mT]_(m1+m2,n1+n2)),
  sub_seqmx 0 n1 m1 n2 (seqmx_of_mx M) =
  seqmx_of_mx (ursubmx M).
Proof.
move=> M ; apply/seqmxP ; split=> [|i Hi|i j].
    rewrite size_map size_take !size_drop size_seqmx subn0.
    case:m2=> [|p] ; first by rewrite addn0 ltnn.
    by rewrite addnS leq_addr.
  rewrite /rowseqmx (nth_map [::]).
    rewrite size_take !size_drop !nth_take // ?nth_drop ?size_row_seqmx ?subn0.
      by rewrite addKn ltnn.
    by rewrite add0n ; exact:ltn_addr.
  rewrite size_take size_drop size_seqmx subn0.
  case:m2=> [|p] ; first by rewrite addn0 ltnn.
  by rewrite addnS leq_addr.
rewrite /fun_of_seqmx /rowseqmx (nth_map [::]).
  by rewrite /subseqrow !nth_take // !nth_drop !mxE -seqmxE.
rewrite size_take size_drop size_seqmx subn0.
case:m2=> [|p] ; first by rewrite addn0 ltnn.
by rewrite addnS leq_addr.
Qed.

Lemma dlsub_seqmxE : forall (M:'M[mT]_(m1+m2,n1+n2)),
  sub_seqmx m1 0 m2 n1 (seqmx_of_mx M) =
  seqmx_of_mx (dlsubmx M).
Proof.
move=> M ; apply/seqmxP ; split=> [|i Hi|i j].
    by rewrite size_map size_take !size_drop size_seqmx addKn ltnn.
  rewrite /rowseqmx (nth_map [::]).
    rewrite size_take !size_drop !nth_take // ?nth_drop ?size_row_seqmx ?subn0.
      case:n2=> [|p] ; first by rewrite addn0 ltnn.
      by rewrite addnS leq_addr.
    by rewrite ltn_add2l.
  by rewrite size_take size_drop size_seqmx addKn ltnn.
rewrite /fun_of_seqmx /rowseqmx (nth_map [::]).
  by rewrite /subseqrow !nth_take // !nth_drop !mxE -seqmxE.
by rewrite size_take size_drop size_seqmx addKn ltnn.
Qed.

Lemma drsub_seqmxE : forall (M:'M[mT]_(m1+m2,n1+n2)),
  sub_seqmx m1 n1 m2 n2 (seqmx_of_mx M) =
  seqmx_of_mx (drsubmx M).
Proof.
move=> M ; apply/seqmxP ; split=> [|i Hi|i j].
    by rewrite size_map size_take !size_drop size_seqmx addKn ltnn.
  rewrite /rowseqmx (nth_map [::]).
    rewrite size_take !size_drop !nth_take // ?nth_drop ?size_row_seqmx ?subn0.
      by rewrite addKn ltnn.
    by rewrite ltn_add2l.
  by rewrite size_take size_drop size_seqmx addKn ltnn.
rewrite /fun_of_seqmx /rowseqmx (nth_map [::]).
  by rewrite /subseqrow !nth_take // !nth_drop !mxE -seqmxE.
by rewrite size_take size_drop size_seqmx addKn ltnn.
Qed.

Definition row_seqmx (M N : seqmatrix) :=
  let aux P Q j := if j < size P then nth 0 P j else nth 0 Q (j-size P) in
  mkseq (fun i => mkseq (aux (nth [::] M i) (nth [::] N i)) (size (rowseqmx M i) + size (rowseqmx N i))) (size M).

Definition col_seqmx (M N : seqmatrix) :=
  mkseq (fun i => if i < size M then nth [::] M i else nth [::] N (i-size M)) (size M + size N).

Definition block_seqmx Aul Aur Adl Adr : seqmatrix :=
  col_seqmx (row_seqmx Aul Aur) (row_seqmx Adl Adr).

Definition block_seqmx2 Aul Aur Adl Adr : seqmatrix :=
  (map2 cat Aul Aur) ++ (map2 cat Adl Adr).

Lemma row_seqmxE (A1 : 'M_(m, n1)) (A2 : 'M_(m, n2)) :
  row_seqmx (seqmx_of_mx A1) (seqmx_of_mx A2) = seqmx_of_mx (row_mx A1 A2).
Proof.
apply/seqmxP; split=>[|i Hi|i j] ; first by rewrite size_mkseq !size_seqmx.
  by rewrite /fun_of_seqmx /rowseqmx nth_mkseq ?size_seqmx // size_mkseq !size_row_seqmx.
rewrite mxE /fun_of_seqmx /rowseqmx !nth_mkseq ?size_seqmx //.
  case:(splitP j)=> j' -> ; first by rewrite size_row_seqmx ltn_ord // -seqmxE.
  by rewrite !size_row_seqmx // ltnNge leq_addr addKn -seqmxE.
by rewrite !size_row_seqmx.
Qed.

Lemma col_seqmxE (A1 : 'M_(m1, n)) (A2 : 'M_(m2, n)) :
  col_seqmx (seqmx_of_mx A1) (seqmx_of_mx A2) = seqmx_of_mx (col_mx A1 A2).
Proof.
apply/seqmxP; split=>[|i Hi|i j] ; first by rewrite size_mkseq !size_seqmx.
  rewrite /fun_of_seqmx /rowseqmx nth_mkseq !size_seqmx //.
  case:ifPn=>H ; first exact:size_row_seqmx.
  rewrite -leqNgt in H ; have Hi' : i - m1 < m2.
    by rewrite -(addKn m1 m2) -(ltn_add2r m1) addKn (subnK H) addnC.
  by rewrite size_row_seqmx.
rewrite mxE /fun_of_seqmx /rowseqmx nth_mkseq.
  case:(splitP i)=> i' ->.
    by rewrite size_seqmx ltn_ord -seqmxE.
  by rewrite size_seqmx ltnNge leq_addr addKn -seqmxE.
by rewrite !size_seqmx.
Qed.

End SeqmxRowCol.

(*
Definition comp_row_add_block iM iN dx (M N:comp_row) :=
  init dx (fun k => get M (iM+k) + get N (iN+k)) default.

Definition comp_matrix_add_block (xM xN:int*int) (dx dy:int) (M N:comp_matrix) :=
  init dy (fun k => comp_row_add_block xM.1 xN.1 dx (get M (xM.2+k)) (get N (xN.2+k))) (make 0 default) : comp_matrix.

Definition comp_row_sub_block iM iN dx (M N:comp_row) :=
  init dx (fun k => get M (iM+k) - get N (iN+k)) default.

Definition comp_matrix_sub_block (xM xN:int*int) (dx dy:int) (M N:comp_matrix) :=
  init dy (fun k => comp_row_sub_block xM.1 xN.1 dx (get M (xM.2+k)) (get N (xN.2+k))) (make 0 default) : comp_matrix.

Definition comp_matrix_opE := (comp_matrix_addE, comp_matrix_mulE).
*)

Section SeqmxBlock.

Variables m1 m2 n1 n2 : nat.

Lemma block_seqmxE (Aul : 'M_(m1,n1)) (Aur:'M_(m1,n2)) (Adl:'M_(m2,n1)) (Adr:'M[mT]_(m2,n2)) :
  block_seqmx (seqmx_of_mx Aul) (seqmx_of_mx Aur)
  (seqmx_of_mx Adl) (seqmx_of_mx Adr) =
  seqmx_of_mx (block_mx Aul Aur Adl Adr).
Proof. by rewrite /block_seqmx !row_seqmxE col_seqmxE. Qed.

Lemma cast_seqmx (M:'M[mT]_(m1,n1)) (H1 : m1=m2) (H2 : n1=n2) :
  seqmx_of_mx (castmx (H1,H2) M) = seqmx_of_mx M.
Proof.
apply/seqmxP; split=> [|i Hi| i j] ; first by rewrite size_seqmx.
  by rewrite size_row_seqmx ?H2 // -H1.
rewrite -seqmxE ; congr fun_of_seqmx ; move: (H1) (H2) M ; rewrite H1 H2=> H3 H4 M.
by rewrite castmx_id.
Qed.

Definition xrowseqmx (M : seqmatrix) : seqmatrix :=
  let R := set_nth [::] M m1 (nth [::] M m2) in
  set_nth [::] R m2 (nth [::] M m1).

End SeqmxBlock.

Definition scalar_seqmx (n : nat) (x : mT) :=
  mkseqmx n n (fun i j => x *+ (i == j)).

Lemma scalar_seqmxE n x :
  scalar_seqmx n x = seqmx_of_mx (@scalar_mx mT n x).
Proof.
apply/seqmxP ; split=> [||i j] ; first by rewrite size_map size_iota.
  exact:size_row_mkseqmx.
by rewrite mkseqmxE // mxE.
Qed.


End seqmx.
