Require Import ssreflect ssrfun ssrbool eqtype ssrnat div seq.
Require Import path choice fintype tuple finset ssralg poly.
Require Import dvdring polydvd zintdiv.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Import GRing.Theory.
Local Open Scope ring_scope.

Section seqpoly.

Variable R : comRingType.

Fixpoint zero_drop p : seq R := match p with
  | [::]    => [::]
  | x :: xs => if x == 0 then zero_drop xs else x :: xs
  end.

Definition drop_zero p : seq R := rev (zero_drop (rev p)).

Lemma drop_zero_last : forall p, last 1 (drop_zero p) != 0%R.
Proof.
elim/last_ind=> /= [|xs c IH]; first exact: nonzero1r.
rewrite /drop_zero rev_rcons /=.
case c0: (c == 0)=> //.
by rewrite rev_cons last_rcons negbT.
Qed.

Lemma drop_zero_eq : forall p, last 1 p != 0 -> drop_zero p = p.
Proof.
elim/last_ind=> //= xs x _ H.
rewrite /drop_zero rev_rcons /=.
case x0: (x == 0).
  by move: H; rewrite (eqP x0) last_rcons eq_refl.
by rewrite rev_cons revK.
Qed.

Lemma drop_zero_nil : drop_zero [::] = [::].
Proof. by rewrite /drop_zero. Qed.

(* Functions for going to ssrpoly from expoly and back *)
Definition seqpoly p : {poly R} := Polynomial (drop_zero_last p).

(* The other direction is simply the polyseq projection *)

Lemma sp2psP : forall p, polyseq (seqpoly p) = drop_zero p.
Proof. by []. Qed.

Lemma ps2spP : forall (p : {poly R}), seqpoly (polyseq p) = p.
Proof.
move=> p.
apply/polyP=> i.
elim: p=> /= xs p.
by rewrite (drop_zero_eq p).
Qed.

Lemma seqpoly0 : seqpoly [::] = 0.
Proof. by apply/polyP=> i; rewrite /= drop_zero_nil seq_poly0. Qed.

Lemma polyseq0 : polyseq (0 : {poly R}) = [::].
Proof. by rewrite -drop_zero_nil -sp2psP seqpoly0. Qed.

(* Reflection lemma *)
Lemma seqpolyP : forall p q : {poly R}, polyseq p == polyseq q -> p = q.
Proof.
move=> p q H.
apply/polyP=> i.
by rewrite (eqP H).
Qed.

(* Addition *)

Fixpoint add_seq p q : seq R := match p, q with
  | [::], _ => q
  | _, [::] => p
  | x :: xs, y :: ys =>
        let xy := x + y in
        let r  := add_seq xs ys
        in if xy == 0 then if r == [::] then [::] else xy :: r else xy :: r
  end.

Lemma add_seqr0 : forall p, add_seq p [::] = p.
Proof. by case. Qed.

Lemma poly_cons_add : forall x y (xs ys : {poly R}),
  poly_cons x xs + poly_cons y ys = poly_cons (x + y) (xs + ys).
Proof.
move=> x y xs ys.
by rewrite !poly_cons_def mulr_addl addrCA polyC_add -!addrA addrCA.
Qed.

Lemma size_poly_neq0 : forall (p: {poly R}),
  (size p != 0%N) = (p != [::] :> seq R).
Proof.
by move => p; rewrite (size_poly_eq0 p) -seq_poly0.
Qed.

Lemma add_seqP : {morph (@polyseq R) : p q / p + q >-> add_seq p q}.
Proof.
elim/poly_ind=> [q| p c IH q].
  by rewrite add0r polyseq0.
elim/poly_ind: q => [| q d _] .
  by rewrite addr0 polyseq0 add_seqr0.
rewrite -!poly_cons_def poly_cons_add !polyseq_cons.
case: ifP => [|spq].
  case: ifP => sp.
    case: ifP => sq.
      rewrite size_poly_neq0 /=.
      case: ifP; rewrite -IH //.
      case: ifP=> // ->.
    rewrite polyseqC.
    have q0: (q == 0) by move/eqP: sq; move/eqP; rewrite size_poly_eq0.
    case d0: (d == 0)=> /=; first by rewrite (eqP d0) (eqP q0) !addr0.
    rewrite (eqP q0) addr0 add_seqr0 size_poly_neq0 => /negbTE ->.
    by case: ifP.
  have p0: (p == 0) by move/eqP: sp=> /eqP; rewrite size_poly_eq0.
  rewrite (eqP p0) !add0r polyseqC=> H.
  rewrite H; rewrite size_poly_neq0 in H.
  case c0: (c == 0)=> /=; first by rewrite (eqP c0) add0r.
  by move/negbTE: H->; case: ifP.
move/eqP: spq => /eqP.
rewrite size_poly_eq0 addrC addr_eq0=> /eqP ->.
rewrite size_opp.
case: ifP=> sp /=.
  by rewrite -IH subrr polyseq0 eqxx polyseqC; case: ifP.
case c0: (c == 0)=> /=; first by rewrite (eqP c0) add0r polyseq0.
case d0: (d == 0)=> /=; first by rewrite (eqP d0) addr0 polyseq0 add_seqr0.
by rewrite !polyseqC c0 d0 /=; case: ifP.
Qed.

(* Negation *)

Definition opp_seq p : seq R := map (fun x => - x) p.

Lemma opp_poly_cons : forall (c : R) p, -(poly_cons c p) = poly_cons (-c) (-p).
Proof.
move=> c p.
rewrite !poly_cons_def.
apply/polyP=> i.
rewrite coef_opp_poly !coef_add_poly !coef_mulX.
case: i=> [|n].
  by rewrite eq_refl !add0r polyC_opp coef_opp.
by rewrite coef_opp polyC_opp coef_opp oppr_add.
Qed.

Lemma opp_seqP : {morph (@polyseq R) : p / - p >-> opp_seq p }.
Proof.
elim/poly_ind=> [|p c IH]; first by rewrite oppr0 polyseq0.
rewrite -!poly_cons_def opp_poly_cons !polyseq_cons size_opp.
case: ifP=> sp /=; first by rewrite -IH.
by rewrite !polyseqC oppr_eq0; case: (c == 0).
Qed.


(* Subtraction *)

Definition sub_seq p q := add_seq p (opp_seq q).

Lemma sub_seqP : {morph (@polyseq R) : p q / p - q >-> sub_seq p q}.
Proof.
elim/poly_ind=> [q|p c IH q].
  by rewrite sub0r /sub_seq polyseq0 /= opp_seqP.
by rewrite /sub_seq -opp_seqP -add_seqP.
Qed.

(* Scaling *)

Fixpoint scale_seq x p : seq R := match p with
  | [::] => [::]
  | hd :: tl => let r   := scale_seq x tl in
                let xhd := x * hd in
                if xhd == 0 then if r == [::] then [::] else xhd :: r else xhd :: r
  end.

Lemma scale_seqP : forall x p, polyseq (scale_poly x p) = scale_seq x (polyseq p).
Proof.
move=> x.
elim/poly_ind=> [| p c IH]; first by rewrite scale_polyE mulr0 polyseq0.
rewrite -poly_cons_def polyseq_cons.
case: ifP=> sp; last first.
  have p0: (p == 0) by move/eqP: sp; move/eqP; rewrite size_poly_eq0.
  rewrite poly_cons_def (eqP p0) mul0r add0r !polyseqC scale_polyE -polyC_mul.
  case c0: (c == 0)=> /=; first by rewrite (eqP c0) mulr0 polyseq0.
  case: ifP=> xc0 /=; first by rewrite (eqP xc0) polyseq0.
  by rewrite polyseqC xc0.
rewrite /= -IH poly_cons_def !scale_polyE mulr_addr -polyC_mul mulrA.
case: ifP=> xc0.
  rewrite (eqP xc0) addr0.
  case: ifP; rewrite -polyseq0; first by move/eqP=> /poly_inj ->; rewrite mul0r.
  by move=> xpnil; rewrite seq_mul_polyX // negbT.
rewrite -poly_cons_def -scale_polyE polyseq_cons.
case: ifP=> // /eqP /eqP; rewrite size_poly_eq0=> /eqP ->.
by rewrite polyseq0 polyseqC xc0.
Qed.


(* Multiplication *)

Fixpoint mul_seq p q := match p,q with
  | [::], _ => [::]
  | _, [::] => [::]
  | x :: xs,_ => add_seq (scale_seq x q) (mul_seq xs (0%R :: q))
  end.

Lemma mul_seqr0 : forall p, mul_seq p [::] = [::].
Proof. by case. Qed.

Lemma mul_seqP : {morph (@polyseq R) : p q / p * q >-> mul_seq p q}.
Proof.
elim/poly_ind=> [|p c IH] q; first by rewrite mul0r polyseq0.
case q0: (q == 0); first by rewrite (eqP q0) mulr0 polyseq0 mul_seqr0.
rewrite -!poly_cons_def polyseq_cons.
elim/poly_ind: q q0=> [|q d _ /eqP /eqP q0]; first by rewrite eqxx.
case sp: (size p == 0%N) => /=.
  rewrite size_poly_eq0 in sp; rewrite (eqP sp) polyseqC.
  case c0: (c == 0).
    by rewrite (eqP c0) poly_cons_def mul0r add0r mul0r polyseq0.
  rewrite /= -scale_seqP -poly_cons_def polyseq_cons poly_cons_def.
  case: ifP=> [sq|].
    by rewrite scale_polyE -polyseq0 -add_seqP addr0 mul0r add0r.
  rewrite size_poly_neq0 => /eqP; rewrite -polyseq0 => /poly_inj ->.
  rewrite polyseqC.
  case d0: (d == 0).
    by rewrite (eqP d0) !poly_cons_def mul0r !add0r mulr0.
  by rewrite !poly_cons_def scale_polyE -add_seqP mul0r !add0r addr0.
rewrite -seq_mul_polyX // -IH -scale_seqP -poly_cons_def polyseq_cons.
case: ifP.
  rewrite !poly_cons_def scale_polyE -add_seqP !mulr_addl -!mulrA addrC.
  by rewrite !mulr_addr [d%:P * 'X]mulrC ['X * (q * 'X)]mulrCA.
rewrite polyseqC.
case d0: (d == 0)=> /= /eqP /eqP; rewrite size_poly_eq0=> /eqP ->.
  by rewrite (eqP d0) !poly_cons_def mul0r addr0 mulr0 polyseq0.
rewrite scale_polyE -add_seqP !poly_cons_def mul0r add0r mulrC.
by rewrite addrC mulrCA mulr_addr [d%:P * _]mulrC.
Qed.

(* lead_coef *)

Definition lead_coef_seq p : R := p`_(size p).-1.

Lemma lead_coef_seqP : forall p, lead_coef p = lead_coef_seq (polyseq p).
Proof.
move=> p.
by rewrite /lead_coef /lead_coef_seq.
Qed.

(* Indeterminates *)

Definition indet (n : nat) : seq R := ncons n 0 [:: 1].

Lemma indetP : forall n, polyseq 'X^n = indet n.
Proof.
by move=> n; rewrite seq_polyXn /indet.
Qed.

(* polyC *)

Definition polyC_seq x : seq R := nseq (x != 0) x.

Lemma polyC_seqP : forall x, polyseq (polyC x) = polyC_seq x.
Proof.
by move=> x; rewrite polyseqC /polyC_seq.
Qed.


(* Pseudo-division *)

Definition edivp_rec_seq (q : seq R)  :=
  let sq := size q in
  let cq := lead_coef_seq q in
  fix loop (n : nat) (k : nat) (qq r : seq R) {struct n} :=
    if size r < sq then (k, qq, r) else
    let m := mul_seq (polyC_seq (lead_coef_seq r)) (indet (size r - sq)) in
    let qq1 := add_seq (mul_seq qq (polyC_seq cq)) m in
    let r1 := sub_seq (mul_seq r (polyC_seq cq)) (mul_seq m q) in
    if n is n1.+1 then loop n1 k.+1 qq1 r1 else (k.+1, qq1, r1).

Definition edivp_seq (p q : seq R) : nat * seq R * seq R :=
  if q == [::] then (0%N, [::], p) else edivp_rec_seq q (size p) 0 [::] p.

Definition divp_seq p q := ((edivp_seq p q).1).2.
Definition modp_seq p q := (edivp_seq p q).2.
Definition scalp_seq p q := ((edivp_seq p q).1).1.

Lemma edivp_rec_seqP :
 forall n k (q qq r : {poly R}),
   let: (l,a,b) := edivp_rec q n k qq r
   in edivp_rec_seq (polyseq q) n k (polyseq qq) (polyseq r) = (l,polyseq a, polyseq b).
Proof.
elim=> [|n ih] k q qq r /=; case: ifP => // _;
  rewrite -!polyC_seqP -indetP -!lead_coef_seqP -!mul_seqP -add_seqP
          -sub_seqP //.
exact: ih.
Qed.

Lemma edivp_seqP :
  forall p q,
    let: (l,a,b) := edivp p q
    in edivp_seq (polyseq p) (polyseq q) = (l,polyseq a,polyseq b).
Proof.
move=> p q.
rewrite /edivp /edivp_seq -polyseq0.
case: ifP; first by move/eqP ->; rewrite eqxx.
case: ifP=> [|_ _]; first by move/eqP => /poly_inj ->; rewrite eqxx.
exact: edivp_rec_seqP.
Qed.

Lemma divp_seqP : {morph (@polyseq R) : p q / divp p q >-> divp_seq p q}.
Proof.
rewrite /divp_seq /divp /= => p q.
move: (edivp_seqP p q).
by case: edivp=> [[a b c]] ->.
Qed.


Lemma modp_seqP : {morph (@polyseq R) : p q / modp p q >-> modp_seq p q}.
Proof.
rewrite /modp_seq /modp /= => p q.
move: (edivp_seqP p q).
by case: edivp=> [[a b c]] ->.
Qed.

Lemma scalp_seqP : forall p q, scalp p q = scalp_seq (polyseq p) (polyseq q).
Proof.
rewrite /scalp_seq /scalp /= => p q.
move: (edivp_seqP p q).
by case: edivp=> [[a b c]] ->.
Qed.

End seqpoly.

Section seqpoly_dvd.

(* Proof that seq R is a ring with explicit divisibility *)
Variable R : dvdRingType.

Implicit Types p q : seq R.

Definition odivp_rec_seq q :=
  let sq := size q in
  let lq := lead_coef_seq q in
  fix loop (n : nat) (r p : seq R) {struct n} :=
    if p == [::] then Some r else
    if size p < sq then None else
    match odivr (lead_coef_seq p) lq with
      | Some x => let m  := mul_seq (polyC_seq x) (@indet R (size p - sq)) in
                  let r1 := add_seq r m in
                  let p1 := sub_seq p (mul_seq m q) in
                  if n is n1.+1 then loop n1 r1 p1 else None
      | None   => None
    end.

Definition odivp_seq p q : option (seq R) :=
  if p == [::] then Some [::] else odivp_rec_seq q (size p) [::] p.

Import PolyDvdRing.

Lemma odivp_recP :
 forall n (q r p : {poly R}),
  odivp_rec_seq (polyseq q) n (polyseq r) (polyseq p) = omap (@polyseq R) (odivp_rec q n r p).
Proof.
elim=> [q r p|n IH q r p] /=; rewrite !lead_coef_seqP -polyseq0.
  case: ifP; first by move/eqP=> /poly_inj ->; rewrite eqxx.
  move=> Hp0; apply/eqP; rewrite eq_sym; move: Hp0.
  case: ifP => [|_ _]; first by move/eqP->; rewrite eqxx.
  by case: ifP => // _; case: odivrP.
case: ifP; first by move/eqP=> /poly_inj ->; rewrite eqxx.
move=> Hp0; apply/eqP; rewrite eq_sym; move: Hp0.
case: ifP => [|_ _]; first by move/eqP->; rewrite eqxx.
case: ifP => // spq.
case: odivrP => // x hx /=.
by rewrite -IH -polyC_seqP -indetP -!mul_seqP -add_seqP -sub_seqP.
Qed.

(* Bad statement!
Lemma odivp_seqP :
  {morph (@polyseq R) : p q / odflt 0 (odivp p q) >-> odflt [::] (odivp_seq p q)}.
Proof.
rewrite /odivp /odivp_seq=> p q /=.
case: ifP => [ /eqP -> | ] /=; first by rewrite polyseq0 eqxx.
case: ifP => [| _ _]; first by rewrite -polyseq0 => /eqP /poly_inj -> /eqP.
rewrite -polyseq0 odivp_recP.
elim/poly_ind: p=> [| p c ih ]; first by rewrite size_poly0 /= eqxx.
rewrite -poly_cons_def size_poly_cons.
case: ifP=> [/andP []|_] /=.
  rewrite size_poly_eq0=> /eqP -> /eqP ->.
  by rewrite poly_cons_def mul0r add0r polyC0 /= eqxx.
by do 2! case: ifP => // _; case: odivrP => // x _; case: odivp_rec.
Qed.
*)

(* Good statement! *)
Lemma odivp_seqP :
  forall p q : {poly R}, omap (@polyseq R) (p %/? q) = odivp_seq (polyseq p) (polyseq q).
Proof.
move=> p q.
rewrite /odivr /= /odivp /odivp_seq -polyseq0 /=.
case: ifP; first by move/eqP->; rewrite eqxx.
case: ifP=> [| _ _]; first by move/eqP => /poly_inj ->; rewrite eqxx.
by rewrite odivp_recP.
Qed.

(* Does not work because seq R isn't a ring...
Lemma odivpP : forall p q : seq R, DvdRing.div_spec p q (odivp_seq p q).

Definition polyDvdRingMixin := DvdRingMixin odivpP.

Canonical Structure polyDvdRingType := DvdRingType {poly R} polyDvdRingMixin.
*)

End seqpoly_dvd.

Section seqpoly_gcd.

Variable R : gcdRingType.

(* Content of polynomials *)

Definition gcdsr_seq R := foldr (@gcdr R) 0.

Lemma gcdsr_seqP : forall p : {poly R}, gcdsr p = gcdsr_seq (polyseq p).
Proof. by []. Qed.

(* Primitive part *)

Definition pp_seq p : seq R := match odivp_seq p (polyC_seq (gcdsr_seq p)) with
  | Some x => x
  | None   => [:: 1]
  end.

Import PolyGcdRing.

Lemma pp_seqP : {morph (@polyseq R) : p / pp p >-> pp_seq p}.
Proof.
move=> p /=.
rewrite /pp /pp_seq -gcdsr_seqP -polyC_seqP -odivp_seqP.
by case: odivrP => //=; rewrite polyseq1.
Qed.

Fixpoint gcdp_rec_seq (n : nat) (p q : seq R) :=
  let r := modp_seq p q in
  if r == [::]
     then q
     else if n is n'.+1 then gcdp_rec_seq n' q (pp_seq r) else pp_seq r.

Definition gcdp_seq p q :=
  let (p1,q1) := if size p < size q then (q,p) else (p,q) in
  let d  := polyC_seq (gcdr (gcdsr_seq p1) (gcdsr_seq q1)) in
  mul_seq d (gcdp_rec_seq (size (pp_seq p1)) (pp_seq p1) (pp_seq q1)).

Lemma gcdp_rec_seqP :
 forall n (p q : {poly R}),
   gcdp_rec_seq n (polyseq p) (polyseq q) = polyseq (gcdp_rec n p q).
Proof.
elim=> [|n ih] p q /=; rewrite -modp_seqP -polyseq0 -pp_seqP.
  case: ifP; first by move/eqP => /poly_inj ->; rewrite eqxx.
  case: ifP=> //; by move/eqP->; rewrite eqxx.
case: ifP; first by move/eqP => /poly_inj ->; rewrite eqxx.
case: ifP=> //; by move/eqP->; rewrite eqxx.
Qed.

Lemma gcdp_seqP : {morph (@polyseq R) : p q / gcdp p q >-> gcdp_seq p q}.
Proof.
rewrite /gcdp /gcdp_seq => p q /=.
by  case: ifP; rewrite -gcdsr_seqP -!polyC_seqP -!pp_seqP gcdp_rec_seqP -mul_seqP.
Qed.

(*
Lemma gcdp_seqP : forall p q g, g %| gcdp_seq p q = (g %| p) && (g %| q).

Definition polyGcdRingMixin := GcdRingMixin gcdpP.
Canonical Structure polyGcdRingType := GcdRingType {poly R} polyGcdRingMixin.
*)

End seqpoly_gcd.

Section Zx.

Require Import Coq.ZArith.Zdiv Coq.ZArith.Zabs.
Require Import infra zintdiv.
Import Zdiv.

Definition seqpolyZ := @seq Z_comRingType.

Open Scope Z_scope.

Time Eval vm_compute in 3 %| 4.
Time Eval vm_compute in gcdr 6685 4011.
Time Eval vm_compute in add_seq [:: 2; 3] [:: 1; 2; 3].
Time Eval vm_compute in mul_seq [:: 0; 1] [:: 0; 1].
Time Eval vm_compute in mul_seq [:: 1; 2; 3] [:: 4; 5; 6].
Time Eval vm_compute in add_seq ([:: 2; 3]) (opp_seq [:: 2; 3]).
Time Eval vm_compute in odivp_seq [:: 4; 6] [:: 2; 3].
Time Eval vm_compute in odivp_seq [:: 4; 13; 28; 27; 18] [:: 1; 2; 3].
Time Eval vm_compute in gcdp_seq [:: 0; 0; 2; 3] [:: 0; 0; 0; 2; 3].
Time Eval vm_compute in
  gcdp_seq [:: (-5); 2; 8; (-3); (-3); 0; 1; 0; 1 ]
           [:: 21; (-9); (-4); 0; 5; 0; 3].


(* This does not work:

(* pxy = (1+x) + (x+x^2)y *)
Definition pxy := [:: [:: 1; 1] ; [:: 0; 1; 1] ].

(* qxy = 1 + (1+x)y + xy^2 *)
Definition qxy := [:: [:: 1]; [:: 1; 1]; [:: 0; 1]].

Time Eval compute in odivp_seq pxy qxy.
Time Eval compute in gcdp_seq pxy qxy.
*)
End Zx.

Section kX.

Variable k : fieldType.

Import PolyPriField.

Definition ediv_rec_seq q :=
  let sq := size q in
  let lq := lead_coef_seq q in
  fix loop (n : nat) (qq r : seq k) {struct n} :=
    if size r < sq then (qq, r) else
    let m := mul_seq (polyC_seq (lead_coef_seq r / lq)) (@indet k (size r - sq)) in
    let qq1 := add_seq qq m in
    let r1 := sub_seq r (mul_seq m q) in
    if n is n1.+1 then loop n1 qq1 r1 else (qq1, r1).

Definition ediv_seq p q : seq k * seq k :=
  if q == [::] then ([::], p) else ediv_rec_seq q (size p) [::] p.

Lemma ediv_rec_seqP :
 forall n (q qq p : {poly k}),
  ediv_rec_seq (polyseq q) n (polyseq qq) (polyseq p) =
  (polyseq (ediv_rec q n qq p).1,polyseq (ediv_rec q n qq p).2).
Proof.
elim=> [|n ih] q qq r /=; case: ifP=> // _.
  by rewrite lead_coef_seqP -polyC_seqP -indetP -!mul_seqP add_seqP sub_seqP.
by rewrite -lead_coef_seqP -polyC_seqP -indetP -!mul_seqP -add_seqP -sub_seqP ih.
Qed.

Lemma ediv_seqP : forall p q : {poly k},
  ediv_seq (polyseq p) (polyseq q) = (polyseq (ediv p q).1,polyseq (ediv p q).2).
Proof.
rewrite /ediv /ediv_seq -polyseq0=> p q.
case: ifP=> hq0.
  by case: ifP=> //; move/eqP: hq0=> /poly_inj ->; rewrite eqxx.
case: ifP=> [/eqP hq|_]; first by move: hq0; rewrite hq eqxx.
by rewrite ediv_rec_seqP.
Qed.

(* Extended gcd, i.e. bezout function

Fixpoint egcd_rec_seq (a b : seq k) n {struct n} : seq k * seq k :=
  if n is n'.+1 then
    if b == [::] then ([::1], [::]) else
    let: (u, v) := egcd_rec_seq b (ediv_seq a b).2 n' in
      (v, (sub_seq u  (mul_seq v (ediv_seq a b).1)))
    else ([::1], [::]).

Definition egcd_seq p q := egcd_rec_seq p q (size q).

Lemma egcd_rec : forall (p q : {poly k}),
  egcd_seq (polyseq p) (polyseq q) = (polyseq (bezout p q).1, polyseq (bezout p q).2).
*)


(* gcd for k[x] *)
Definition gcd_seq a b :=
  let: (a1, b1) := if size a < size b then (b, a) else (a, b) in
  if a1 == [::] then b1 else
  let fix loop (n : nat) (aa bb : seq k) {struct n} :=
      let rr := (ediv_seq aa bb).2 in
      if rr == [::] then bb else
      if n is n1.+1 then loop n1 bb rr else rr in
  loop (size a1) a1 b1.

Lemma gcd_seqP : {morph (@polyseq k) : p q / gcdr (p : {poly k}) q >-> gcd_seq p q}.
Proof.
move=> p q /=.
rewrite /gcdr /gcd_seq /= /EuclideanRing.Mixins.gcd /= -polyseq0.
case: ifP=> _.
  case: ifP=> [/eqP ->|]; first by rewrite eqxx.
  case: ifP=> [/eqP /poly_inj ->|_ _]; first by rewrite eqxx.
  remember (size q) as n; rewrite -Heqn; clear Heqn.
  elim: n q p=> [|n ih] p q; rewrite ediv_seqP /=.
    case: ifP => [/eqP ->|]; first by rewrite eqxx.
    by case: ifP => [/eqP /poly_inj ->|//]; rewrite eqxx.
  case: ifP => [/eqP ->|]; first by rewrite eqxx.
  case: ifP=> [/eqP /poly_inj ->|_ _]; first by rewrite eqxx.
  by rewrite ih.
case: ifP=> [/eqP ->|]; first by rewrite eqxx.
case: ifP=> [/eqP /poly_inj ->|_ _]; first by rewrite eqxx.
remember (size p) as n; rewrite -Heqn; clear Heqn.
elim: n q p=> [|n ih] p q; rewrite ediv_seqP /=.
  case: ifP => [/eqP ->|]; first by rewrite eqxx.
  by case: ifP => [/eqP /poly_inj ->|//]; rewrite eqxx.
case: ifP => [/eqP ->|]; first by rewrite eqxx.
case: ifP=> [/eqP /poly_inj ->|_ _]; first by rewrite eqxx.
by rewrite ih.
Qed.

End kX.

Section Qx.

Require Import infra QArith_base Qcanon.

Notation "n %:Q" := (n%Z : Qcb) (at level 2, left associativity, format "n %:Q").

Definition foo : seq Qcb_fieldType := [:: 1 %:Q].

Time Eval vm_compute in add_seq [:: (-2)%:Q; 3%:Q] [:: 1%:Q; 2%:Q; 3%:Q].
Time Eval vm_compute in
  gcd_seq [:: (-5)%:Q; 2%:Q; 8%:Q; (-3)%:Q; (-3)%:Q; 0%:Q; 1%:Q; 0%:Q; 1%:Q ]
          [:: 21%:Q; (-9)%:Q; (-4)%:Q; 0%:Q; 5%:Q; 0%:Q; 3%:Q].

End Qx.
