Require Import ssreflect ssrfun ssrbool eqtype ssrnat div seq.
Require Import path choice fintype tuple finset ssralg poly.
Require Import dvdring polydvd zintdiv seqpoly.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Import GRing.Theory.
Local Open Scope ring_scope.

Section SeqPoly_dvd.

(* Divisibility operations on seqpoly *)

(* Proof that seq R is a ring with explicit divisibility *)
Variable R : dvdRingType.

Implicit Types p q : seq R.

Definition odivp_rec_seq q :=
  let sq := size q in
  let lq := lead_coef_seq q in
  fix loop (n : nat) (r p : seq R) {struct n} :=
    if p == [::] then Some r else
    if size p < sq then None else
    match odivr (lead_coef_seq p) lq with
      | Some x => let m  := mul_seq (polyC_seq x) (@indet R (size p - sq)) in
                  let r1 := add_seq r m in
                  let p1 := sub_seq p (mul_seq m q) in
                  if n is n1.+1 then loop n1 r1 p1 else None
      | None   => None
    end.

Definition odivp_seq p q : option (seq R) :=
  if p == [::] then Some [::] else odivp_rec_seq q (size p) [::] p.

Import PolyDvdRing.

Lemma odivp_recP :
 forall n (q r p : {poly R}),
  odivp_rec_seq (polyseq q) n (polyseq r) (polyseq p) = omap (@polyseq R) (odivp_rec q n r p).
Proof.
elim=> [q r p|n IH q r p] /=; rewrite !lead_coef_seqP -polyseq0.
  case: ifP => [/seqpolyP ->|Hp0]; first by rewrite eqxx.
  apply/eqP; rewrite eq_sym; move: Hp0.
  case: ifP=> [/seqpolyP ->|_ _]; first by rewrite eqxx.
  by case: ifP => // spq; case: odivrP.
case: ifP => [/seqpolyP ->|Hp0]; first by rewrite eqxx.
apply/eqP; rewrite eq_sym; move: Hp0.
case: ifP=> [/seqpolyP ->|_ _]; first by rewrite eqxx.
case: ifP => // spq; case: odivrP=> // x hx /=.
by rewrite -IH -polyC_seqP -indetP -!mul_seqP add_seqP sub_seqP.
Qed.

Lemma odivp_seqP :
  forall p q : {poly R}, omap (@polyseq R) (p %/? q) = odivp_seq (polyseq p) (polyseq q).
Proof.
rewrite /odivr /= /odivp /odivp_seq -polyseq0 /= => p q.
case: ifP => [/eqP ->|]; first by rewrite eqxx.
case: ifP=> [/seqpolyP ->| _ _]; first by rewrite eqxx.
by rewrite odivp_recP.
Qed.

(* Does not work because seq R isn't a ring...
Lemma odivpP : forall p q : seq R, DvdRing.div_spec p q (odivp_seq p q).

Definition polyDvdRingMixin := DvdRingMixin odivpP.

Canonical Structure polyDvdRingType := DvdRingType {poly R} polyDvdRingMixin.
*)

End SeqPoly_dvd.

Section SeqPoly_gcd.

(* GCD operations on seqpoly *)

Variable R : gcdRingType.

(* Content of polynomials *)

Definition gcdsr_seq R := foldr (@gcdr R) 0.

Lemma gcdsr_seqP : forall p : {poly R}, gcdsr p = gcdsr_seq (polyseq p).
Proof. by []. Qed.

(* Primitive part *)

Definition pp_seq p : seq R := match odivp_seq p (polyC_seq (gcdsr_seq p)) with
  | Some x => x
  | None   => [:: 1]
  end.

Import PolyGcdRing.

Lemma pp_seqP : {morph (@polyseq R) : p / pp p >-> pp_seq p}.
Proof.
move=> p /=.
rewrite /pp /pp_seq -gcdsr_seqP -polyC_seqP -odivp_seqP.
by case: odivrP => //=; rewrite polyseq1.
Qed.

Fixpoint gcdp_rec_seq (n : nat) (p q : seq R) :=
  let r := modp_seq p q in
  if r == [::]
     then q
     else if n is n'.+1 then gcdp_rec_seq n' q (pp_seq r) else pp_seq r.

Definition gcdp_seq p q :=
  let (p1,q1) := if size p < size q then (q,p) else (p,q) in
  let d  := polyC_seq (gcdr (gcdsr_seq p1) (gcdsr_seq q1)) in
  mul_seq d (gcdp_rec_seq (size (pp_seq p1)) (pp_seq p1) (pp_seq q1)).

Lemma gcdp_rec_seqP :
 forall n (p q : {poly R}),
   gcdp_rec_seq n (polyseq p) (polyseq q) = polyseq (gcdp_rec n p q).
Proof.
elim=> [|n ih] p q /=; rewrite -modp_seqP -polyseq0 -pp_seqP;
by do !(case: ifP => [/seqpolyP ->|]; first by rewrite eqxx).
Qed.

Lemma gcdp_seqP : {morph (@polyseq R) : p q / gcdp p q >-> gcdp_seq p q}.
Proof.
rewrite /gcdp /gcdp_seq => p q /=.
by case: ifP;
   rewrite -gcdsr_seqP -!polyC_seqP -!pp_seqP gcdp_rec_seqP -mul_seqP.
Qed.

(*
Lemma gcdp_seqP : forall p q g, g %| gcdp_seq p q = (g %| p) && (g %| q).

Definition polyGcdRingMixin := GcdRingMixin gcdpP.
Canonical Structure polyGcdRingType := GcdRingType {poly R} polyGcdRingMixin.
*)

End SeqPoly_gcd.

Section KX.
(* Operations on k[x] *)

Variable K : fieldType.

Import PolyPriField.

Definition ediv_rec_seq q :=
  let sq := size q in
  let lq := lead_coef_seq q in
  fix loop (n : nat) (qq r : seq K) {struct n} :=
    if size r < sq then (qq, r) else
    let m := mul_seq (polyC_seq (lead_coef_seq r / lq)) (@indet K (size r - sq)) in
    let qq1 := add_seq qq m in
    let r1 := sub_seq r (mul_seq m q) in
    if n is n1.+1 then loop n1 qq1 r1 else (qq1, r1).

Definition ediv_seq p q : seq K * seq K :=
  if q == [::] then ([::], p) else ediv_rec_seq q (size p) [::] p.

Lemma ediv_rec_seqP :
 forall n (q qq p : {poly K}),
  ediv_rec_seq (polyseq q) n (polyseq qq) (polyseq p) =
  (polyseq (ediv_rec q n qq p).1,polyseq (ediv_rec q n qq p).2).
Proof.
elim=> [|n ih] q qq r /=; case: ifP=> // _;
rewrite lead_coef_seqP -polyC_seqP -indetP -!mul_seqP.
  by rewrite add_seqP sub_seqP.
by rewrite -add_seqP -sub_seqP ih.
Qed.

Lemma ediv_seqP : forall p q : {poly K},
  ediv_seq (polyseq p) (polyseq q) = (polyseq (ediv p q).1,polyseq (ediv p q).2).
Proof.
rewrite /ediv /ediv_seq -polyseq0=> p q.
case: ifP => [/seqpolyP ->|]; first by rewrite eqxx.
case: ifP => [/seqpolyP ->|]; first by rewrite eqxx.
by rewrite ediv_rec_seqP.
Qed.

(* Extended gcd, i.e. bezout function

Fixpoint egcd_rec_seq (a b : seq k) n {struct n} : seq k * seq k :=
  if n is n'.+1 then
    if b == [::] then ([::1], [::]) else
    let: (u, v) := egcd_rec_seq b (ediv_seq a b).2 n' in
      (v, (sub_seq u  (mul_seq v (ediv_seq a b).1)))
    else ([::1], [::]).

Definition egcd_seq p q := egcd_rec_seq p q (size q).

Lemma egcd_rec : forall (p q : {poly k}),
  egcd_seq (polyseq p) (polyseq q) = (polyseq (bezout p q).1, polyseq (bezout p q).2).
*)


(* gcd for k[x] *)
Definition gcd_seq a b :=
  let: (a1, b1) := if size a < size b then (b, a) else (a, b) in
  if a1 == [::] then b1 else
  let fix loop (n : nat) (aa bb : seq K) {struct n} :=
      let rr := (ediv_seq aa bb).2 in
      if rr == [::] then bb else
      if n is n1.+1 then loop n1 bb rr else rr in
  loop (size a1) a1 b1.

Lemma gcd_seqP : {morph (@polyseq K) : p q / gcdr (p : {poly K}) q >-> gcd_seq p q}.
Proof.
rewrite /gcdr /gcd_seq /= /EuclideanRing.Mixins.gcd /= -polyseq0 => p q.
wlog sqp: p q / size q <= size p=>[hwlog|].
  case: ltnP=> sqp.
    by move/hwlog:(ltnW sqp); rewrite ltnNge (ltnW sqp).
  by move/hwlog:(sqp); rewrite ltnNge sqp.
rewrite ltnNge sqp /=.
case: ifP=> [/eqP ->|]; first by rewrite eqxx.
case: ifP=> [/seqpolyP ->|_ _]; first by rewrite eqxx.
remember (size p) as n; rewrite -Heqn; clear Heqn sqp.
elim: n q p=> [|n ih] p q; rewrite ediv_seqP /=.
  case: ifP => [/eqP ->|]; first by rewrite eqxx.
  by case: ifP => [/seqpolyP ->|] //; rewrite eqxx.
case: ifP => [/eqP ->|]; first by rewrite eqxx.
by case: ifP => [/seqpolyP ->|]; rewrite ?ih // eqxx.
Qed.

End KX.

Section Zx.

(* Examples of computations over Z[x] *)

Require Import Coq.ZArith.Zdiv Coq.ZArith.Zabs.
Require Import infra zintdiv.
Import Zdiv.

Definition seqpolyZ := @seq Z_comRingType.

Open Scope Z_scope.

Time Eval vm_compute in 3 %| 4.
Time Eval vm_compute in gcdr 6685 4011.
Time Eval vm_compute in add_seq [:: 2; 3] [:: 1; 2; 3].
Time Eval vm_compute in mul_seq [:: 0; 1] [:: 0; 1].
Time Eval vm_compute in mul_seq [:: 1; 2; 3] [:: 4; 5; 6].
Time Eval vm_compute in add_seq ([:: 2; 3]) (opp_seq [:: 2; 3]).
Time Eval vm_compute in odivp_seq [:: 4; 6] [:: 2; 3].
Time Eval vm_compute in odivp_seq [:: 4; 13; 28; 27; 18] [:: 1; 2; 3].
Time Eval vm_compute in gcdp_seq [:: 0; 0; 2; 3] [:: 0; 0; 0; 2; 3].
Time Eval vm_compute in
  gcdp_seq [:: (-5); 2; 8; (-3); (-3); 0; 1; 0; 1 ]
           [:: 21; (-9); (-4); 0; 5; 0; 3].

(* This does not work:

(* pxy = (1+x) + (x+x^2)y *)
Definition pxy := [:: [:: 1; 1] ; [:: 0; 1; 1] ].

(* qxy = 1 + (1+x)y + xy^2 *)
Definition qxy := [:: [:: 1]; [:: 1; 1]; [:: 0; 1]].

Time Eval compute in odivp_seq pxy qxy.
Time Eval compute in gcdp_seq pxy qxy.
*)
End Zx.

Section Qx.

(* Examples of computations over Q[x] *)

Require Import infra QArith_base Qcanon.

Notation "n %:Q" := (n%Z : Qcb) (at level 2, left associativity, format "n %:Q").

Definition foo : seq Qcb_fieldType := [:: 1 %:Q].

Time Eval vm_compute in add_seq [:: (-2)%:Q; 3%:Q] [:: 1%:Q; 2%:Q; 3%:Q].
Time Eval vm_compute in
  gcd_seq [:: 14%:Q; (-3)%:Q; 4%:Q; (-4)%:Q; 1%:Q ]
          [:: 6%:Q; 17%:Q; 12%:Q; 8%:Q; 1%:Q].
Time Eval vm_compute in
  gcd_seq [:: (-5)%:Q; 2%:Q; 8%:Q; (-3)%:Q; (-3)%:Q; 0%:Q; 1%:Q; 0%:Q; 1%:Q ]
          [:: 21%:Q; (-9)%:Q; (-4)%:Q; 0%:Q; 5%:Q; 0%:Q; 3%:Q].

End Qx.