(*
Formalization of Anders' Haskell algorithm
for general Smith Normal form
using ssreflect matrices
*)
Require Import ssreflect ssrfun ssrbool eqtype ssrnat div seq.
Require Import ssralg fintype perm.
Require Import dvdring matrix  bigop zmodp.


Import GRing.Theory.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


Section SmithNF.
Variable R : priRingType.

Local Open Scope ring_scope.

(* skeleton for matrices of the shape
          (kth column)
  / a .... b ..... \
  | . 1 .......... |
  | ....1......... |
  | c .... d .... | (kth row)
  | ..........1... |
  \ .............1 /

that we will use for "Bezout matrices"

Thanks to maxime for this idea
*)

Definition combine_mx (a b c d : R) (m : nat) (k : 'I_m) :=
  let k' := lift 0 k in
  let d := \row_j (a *+ (j == 0) + d *+ (j == k') +
           ((j != 0) && (j != k'))%:R) in
  diag_mx d + c *: delta_mx k' 0 + b *: delta_mx 0 k'.

Lemma combine_mxE (a b c d : R) (m n : nat) (M : 'M_(1 + m,1 + n)) (k : 'I_m) :
  let k' := lift 0 k in
  let r0 := a *: row 0 M + b *: row k' M in
  let rk := c *: row 0 M + d *: row k' M in
  combine_mx a b c d k *m M =
    \matrix_i (r0 *+ (i == 0) + rk *+ (i == k') + row i M *+ ((i != 0) && (i != k'))).
Proof.
rewrite /combine_mx; apply/matrixP=> i j.
rewrite !mxE (bigD1 (0 : 'I_m.+1)) // !mxE (bigD1 (lift 0 k)) // !mxE /=.
case H:(i==0).
  rewrite big1=> [|l /andP [H1 H2]].
    rewrite (eqP H) !eqxx !mulr1n !mxE !mulr0 !addr0.
    by rewrite !mulr1 mulr0n add0r.
  rewrite !mxE (eqP H) (eq_sym 0 l) (negbTE H1) (negbTE H2).
  by rewrite mulr0n !mulr0 !add0r mul0r.
case H':(i==lift 0 k).
  rewrite big1=> [|l /andP [H1 H2]].
    rewrite (eqP H') !eqxx !mulr1n !mxE !mulr0 !addr0.
    by rewrite !mulr1 mulr0n !add0r addrC.
  rewrite !mxE (eqP H') !(eq_sym (lift 0 k) l) eqxx (negbTE H1) (negbTE H2).
  by rewrite mulr0n !mulr0 !add0r mul0r.
rewrite (bigD1 i); last by rewrite H H'.
rewrite !mxE big1=> [/=|l /andP [/andP [H1 H2] H3]].
  rewrite H H' eqxx !mulr0n !mulr0 !mulr1n.
  by rewrite !addr0 !mul0r !add0r mul1r.
rewrite !mxE H H' (negbTE H1) (negbTE H2) (eq_sym i l) (negbTE H3) /=.
by rewrite mulr0n !mulr0 !addr0 mul0r.
Qed.

Lemma combine_mx_inv (a b c d : R) m (k : 'I_m) :
  a * d - b * c = 1 ->
  combine_mx a b c d k *m combine_mx d (-b) (-c) a k= 1%:M.
Proof.
move=> H; rewrite combine_mxE; apply/matrixP=> i j; rewrite !mxE.
case Hi:(i==0).
  rewrite !mxE (eqP Hi) !eqxx !mulr0 mxE !addr0 (eq_sym 0 j).
  case Hj:(j==0).
    rewrite (eqP Hj) mulr1 !mulr0.
    by rewrite addr0 sub0r mulrN.
  rewrite !mulr0 !add0r addr0 (eq_sym _ j).
  case:(j==lift 0 k); last by rewrite !mulr0 add0r.
  by rewrite mulr1 mulr1n mulrN mulrC addNr.
case Hj:(j==0).
  rewrite !mxE (eqP Hj) Hi add0r.
  case Hk:(i==lift 0 k).
    rewrite !mxE !eqxx !mulr0 mulr1 !addr0 !add0r.
    by rewrite mulrN addrC mulrC addNr.
  by rewrite !mxE Hi Hk eqxx !add0r !mulr0 addr0.
case Hk:(i==lift 0 k).
  rewrite !mxE (eq_sym 0 j) Hj (eqP Hk) !eqxx !mulr0 !addr0 !add0r.
  rewrite (eq_sym _ j).
  case:(j==lift 0 k); last by rewrite !mulr0 addr0.
  by rewrite !mulr1 addrC mulrN (mulrC c) (mulrC d).
by rewrite !mxE Hi Hj Hk !mulr0 !add0r !addr0.
Qed.

Lemma combine_mx_unit : forall n (k: 'I_n) a b c d,
  a * d - b * c = 1 -> combine_mx a b c d k \in unitmx.
Proof.
move => n k a b c d h.
by case: (mulmx1_unit (combine_mx_inv k h)).
Qed.

(* "Bezout matrix" used to simplify a line / column matrix *)
Definition bz_mx (a b: R) (n: nat) (k: 'I_n): R * 'M[R]_(1 + n) :=
 let: (g,u,v,a1,b1) := egcdr a b in (g,combine_mx u v (-b1) a1 k).


CoInductive bz_mx_spec (a b:R) (n:nat) (k: 'I_n) :
  R * 'M[R]_(1 + n) -> Type :=
 BMSpec g M of (g %= gcdr a b) &
               (M *m (\col_i if (i == 0) then a else
                                if (i == lift ord0 k) then b
                                else 0) == col_mx g%:M (0 :'cV[R]_n) ) &
               M \in unitmx : bz_mx_spec a b k  (g,M).

Lemma bz_mxP : forall a b n (k: 'I_n),
  bz_mx_spec a b k (bz_mx a b k).
Proof.
rewrite /bz_mx => a b n k.
case: egcdrP => g u v a1 b1 hg heq ha1 hb1.
constructor => //.
- rewrite combine_mxE.
  apply/eqP/colP => i; rewrite !mxE.
  case: splitP => ii.
  + rewrite [ii]ord1 !mxE => hii; clear ii.
    have -> /= : i = 0 by apply/ord_inj.
    by rewrite !mxE !addr0 /= !eqxx ha1 hb1 !mulrA -mulr_addl hg mul1r.
  rewrite !mxE => hii.
  have -> /= : i = lift 0 ii by apply/ord_inj.
  rewrite !mxE add0r.
  case hik : (lift 0 ii == lift 0 k).
  + by rewrite !mxE /= eqxx addr0 ha1 hb1 -mulrCA mulNr mulrN addNr.
  by rewrite !mxE add0r hik.
rewrite combine_mx_unit //.
by rewrite !mulrN opprK.
Qed.

Lemma lift0_unitmx : forall (n:nat) (M: 'M[R]_n), M \in unitmx ->
  lift0_mx M \in unitmx.
Proof.
move => n M; move/mulVmx => h.
have := (@mulmx1_unit _  _ (lift0_mx (invmx M)) (lift0_mx M)).
rewrite mulmx_block !mulmx1 !mulmx0 !mul0mx !addr0 add0r h.
rewrite -scalar_mx_block.
by case.
Qed.


(* given a column matrix C, computes U such that U * C is of the form
    g
    0
 where g is the gcd of the column C
*)

Fixpoint lemV {n:nat} : 'cV[R]_(1 + n) -> R * 'M[R]_(1 + n) :=
 match n return 'cV[R]_(1 + n) -> R * 'M[R]_(1 + n) with
  | O => fun l => (l 0 0, 1%:M)
  | S p => fun (l : 'cV[R]_(1 + _)) =>
    if (dsubmx l == 0) then ((usubmx l) 0 0, 1%:M) else
    let (b,m) := lemV (dsubmx l) in
    let (g,bm) := @bz_mx ((usubmx l) 0 0) b (1 + p) ord0 in
      (g, bm *m lift0_mx m)
end.

(* same for line matrices *)
Fixpoint lemH {n:nat} : 'rV[R]_(1 + n) -> R * 'M[R]_(1 + n) :=
 match n return 'rV[R]_(1 + n) -> R * 'M[R]_(1 + n) with
  | O => fun l => (l 0 0, 1%:M)
  | S p => fun (l : 'rV[R]_(1 + _)) =>
    if (rsubmx l == 0) then ((lsubmx l) 0 0, 1%:M) else
    let (b,m) := lemH (rsubmx l) in
    let (g,bm) := @bz_mx ((lsubmx l) 0 0) b (1 + p) ord0 in
      (g, lift0_mx m *m bm^T)
end.


Lemma lemV0: forall n (C :'cV_(1 + n)), let: (g,_) := lemV C in
  (g == 0) = (C == 0).
Proof.
elim => [ | n hi] /=.
- move => C; apply/eqP/eqP => hc.
  + by apply/colP => i; rewrite !mxE [i]ord1 hc.
  by rewrite hc !mxE.
change (1 + n.+1)%N with (1+(1+n))%N => C.
case hc1 : (dsubmx C == 0).
- rewrite -{2}[C]vsubmxK (eqP hc1).
  apply/eqP/eqP.
  + move => h; apply/colP => i; rewrite !mxE.
    case: splitP => j.
    * by rewrite [j]ord1 h.
    by rewrite !mxE.
  rewrite -[0: 'cV[R]_(1 + (1 + _))]col_mx0.
  by case/eq_col_mx => ->; rewrite !mxE.
case: lemV (hi (dsubmx C)) => g' E heq.
case: bz_mxP => g M h1 _ _.
apply/eqP/eqP.
- move => hg.
  by move : h1; rewrite hg eqd0r gcdr_eq0 heq hc1 andbF.
rewrite -[C]vsubmxK -col_mx0.
case/eq_col_mx => h3 h4.
have hg' : g' = 0 by apply/eqP; rewrite heq h4.
apply/eqP; rewrite -eqdr0.
apply (eqd_trans h1).
rewrite h3 !mxE hg'.
by apply gcd0r.
Qed.

Lemma lemH0: forall n (L :'rV_(1 + n)), let: (g,_) := lemH L in
  (g == 0) = (L == 0).
Proof.
elim => [ | n hi] /=.
- move => L; apply/eqP/eqP => hl.
  + by apply/rowP => i; rewrite !mxE [i]ord1 hl.
  by rewrite hl !mxE.
change (1 + n.+1)%N with (1+(1+n))%N => L.
case hl1 : (rsubmx L == 0).
- rewrite -{2}[L]hsubmxK (eqP hl1).
  apply/eqP/eqP.
  + move => h; apply/rowP => i; rewrite !mxE.
    case: splitP => j.
    * by rewrite [j]ord1 h.
    by rewrite !mxE.
  rewrite -[0: 'rV[R]_(1 + (1 + _))]row_mx0.
  by case/eq_row_mx => ->; rewrite !mxE.
case: lemH (hi (rsubmx L)) => g' E heq.
case: bz_mxP => g M h1 _ _.
apply/eqP/eqP.
- move => hg.
  by move : h1; rewrite hg eqd0r gcdr_eq0 heq hl1 andbF.
rewrite -[L]hsubmxK -row_mx0.
case/eq_row_mx => h3 h4.
have hg' : g' = 0 by apply/eqP; rewrite heq h4.
apply/eqP; rewrite -eqdr0.
apply (eqd_trans h1).
rewrite h3 !mxE hg'.
by apply gcd0r.
Qed.


Lemma dvd_sdvd_eqd : forall (a b:R), (a %| b) = (a %<| b) || ( a %= b).
Proof.
move => a b; apply/idP/idP.
- case h : ( b %| a ).
  + rewrite eqd_def h => -> /=.
    by rewrite orbT.
  by rewrite sdvdr_def h => -> /= .
case/orP.
- rewrite sdvdr_def.
  by case/andP => ->.
rewrite eqd_def.
by case/andP => ->.
Qed.


(* reflexion function from column matrices to lists
   this is used since gcdsr is defined for lists *)
Fixpoint col2seq (n:nat) : 'cV[R]_n -> seq R :=
 match n return 'cV[R]_n -> seq R with
  | O => fun _ => [::]
  | S p => fun (c: 'cV[R]_(1 + _)) =>
    c 0 0 :: (col2seq (dsubmx c))
end.


Lemma dsubmx0 : forall n p q, dsubmx (0 : 'M[R]_(n + p,q)) = 0.
Proof.
by move => n p q; apply/matrixP => i j; rewrite !mxE.
Qed.


Lemma usubmx0 : forall n p q, usubmx (0 : 'M[R]_(n + p,q)) = 0.
Proof.
by move => n p q; apply/matrixP => i j; rewrite !mxE.
Qed.


Lemma in_col2seq : forall n (c: 'cV[R]_n) x,
  (x \in col2seq c) = (existsb i, x == c i 0).
Proof.
elim => [ | n hi].
- move => c x /=.
  rewrite [c]flatmx0 in_nil.
  apply/idP/existsP => //.
  case => y; rewrite !mxE.
  by case : y.
change n.+1 with (1 + n)%N => c x /=.
rewrite in_cons.
apply/orP/existsP.
- case => hu.
  + by exists 0.
  move : hu; rewrite hi.
  case/existsP => y h; exists (lift 0 y).
  by move : h; rewrite !mxE rshift1.
case => y.
case: (splitP y) => j.
- rewrite [j]ord1 => hj.
  have -> : y = 0 by apply/ord_inj.
  by move => h'; left.
move => hj.
have -> : y = lift 0 j by apply/ord_inj.
move => hx.
right; rewrite hi.
apply/existsP; exists j.
by rewrite !mxE rshift1.
Qed.

Lemma col2seq0 : forall n, col2seq (0: 'cV[R]_n) = nseq n 0.
Proof.
elim => [ | n hi] //=.
by rewrite !mxE dsubmx0 hi.
Qed.

(* same for row matrices *)
Fixpoint row2seq (n:nat) : 'rV[R]_n -> seq R :=
 match n return 'rV[R]_n -> seq R with
  | O => fun _ => [::]
  | S p => fun (l: 'rV[R]_(1 + _)) =>
    l 0 0 :: (row2seq (rsubmx l))
end.



Lemma in_row2seq : forall n (l: 'rV[R]_n) x,
  (x \in row2seq l) = (existsb i, x == l 0 i).
Proof.
elim => [ | n hi].
- move => l x /=.
  rewrite [l]thinmx0 in_nil.
  apply/idP/existsP => //.
  case => y; rewrite !mxE.
  by case : y.
change n.+1 with (1 + n)%N => l x /=.
rewrite in_cons.
apply/orP/existsP.
- case => hu.
  + by exists 0.
  move : hu; rewrite hi.
  case/existsP => y h; exists (lift 0 y).
  by move : h; rewrite !mxE rshift1.
case => y.
case: (splitP y) => j.
- rewrite [j]ord1 => hj.
  have -> : y = 0 by apply/ord_inj.
  by move => h'; left.
move => hj.
have -> : y = lift 0 j by apply/ord_inj.
move => hx.
right; rewrite hi.
apply/existsP; exists j.
by rewrite !mxE rshift1.
Qed.


Lemma rsubmx0 : forall n p q, rsubmx (0 : 'M[R]_(n, p + q)) = 0.
Proof.
by move => n p q; apply/matrixP => i j; rewrite !mxE.
Qed.


Lemma lsubmx0 : forall n p q, lsubmx (0 : 'M[R]_(n, p + q)) = 0.
Proof.
by move => n p q; apply/matrixP => i j; rewrite !mxE.
Qed.


Lemma row2seq0 : forall n, row2seq (0: 'rV[R]_n) = nseq n 0.
Proof.
elim => [ | n hi] //=.
by rewrite !mxE rsubmx0 hi.
Qed.


(* gcd of a list of 0s is 0 *)
Lemma gcdsr_nseq0 : forall n, @gcdsr R (nseq n 0) = 0.
Proof.
elim => [ | n hi] //=.
rewrite hi.
apply/eqP; rewrite -eqdr0.
by apply gcdr0.
Qed.


Lemma mx_scalar0 :
  forall (n:nat) (a:R), (a%:M == 0 :> 'M[R]_n.+1) = (a == 0) .
Proof.
move => n a.
apply/eqP/eqP.
- move/matrixP => ha.
  by move: (ha 0 0); rewrite !mxE mulr1n.
move => ->.
by apply/matrixP => i j; rewrite !mxE mul0rn.
Qed.

CoInductive lemV_spec (n:nat) (C: 'cV[R]_(1 + n)) :
  R * 'M[R]_(1 + n) -> Type :=
| lemVgn0 g M of (g %= gcdsr (col2seq C)) &
  M *m C = col_mx g%:M (0 : 'cV_n) & M \in unitmx : lemV_spec C (g, M).

(* gcd of a list is 0 iff the list is only 0s *)
Lemma gcdsr_eq0 : forall l: seq R,
  (gcdsr l == 0) = (l == nseq (size l) 0).
Proof.
elim => [ | hd tl hi] //=.
- by rewrite !eqxx.
apply/eqP/eqP.
- move/eqP; rewrite gcdr_eq0 hi.
  case/andP => /eqP -> /eqP ->.
  by rewrite size_nseq.
case => -> ->; apply/eqP.
by rewrite gcdr_eq0 gcdsr_nseq0 !eqxx.
Qed.


Lemma col2seq0P : forall n (c: 'cV[R]_n),
  (c == 0) = (col2seq c == nseq n 0).
Proof.
elim => [c | n hi]//=.
- by rewrite [c]flatmx0 !eqxx.
rewrite [n.+1]/(1 + n)%N => c.
rewrite -{1}[c]vsubmxK -{1}[0]vsubmxK [usubmx c]mx11_scalar !mxE lshift0
  dsubmx0 usubmx0.
apply/eqP/eqP.
- case/eq_col_mx => /eqP.
  rewrite mx_scalar0 => /eqP -> ->.
  by rewrite col2seq0.
case => -> hc.
f_equal.
- apply/matrixP => i j; rewrite !mxE.
  by rewrite mul0rn.
apply/eqP.
by rewrite hi hc.
Qed.

Lemma size_col2seq : forall n (c: 'cV[R]_n), size (col2seq c) = n.
Proof.
elim => [ | n hi] c //=.
by rewrite hi.
Qed.

Lemma lemVP : forall n (C: 'cV[R]_(1 + n)), lemV_spec C (lemV C).
Proof.
elim => [ | n hi] /=.
- move => C.
  constructor => /=.
  + by rewrite eqd_sym; apply/ gcdr0.
  + by rewrite mul1mx -{1}[C]vsubmxK [dsubmx C]flatmx0
      [usubmx C]mx11_scalar !mxE lshift0.
  by rewrite unitmx1.
rewrite [(1 + n.+1)%N]/(1 + (1 + n))%N => C.
case: ifP => /eqP hsub0.
- constructor => /=.
  + rewrite hsub0 dsubmx0 col2seq0 !mxE.
    have -> : gcdr 0 (gcdsr (nseq n 0)) = 0 :> R by
    clear hi C hsub0; elim : n => [ | n hi] /=;
        by apply/eqP; rewrite ?hi -eqdr0; apply gcdr0.
    rewrite lshift0 eqd_sym.
    by apply/gcdr0.
  + by rewrite mul1mx -{1}[C]vsubmxK hsub0
     [usubmx C]mx11_scalar !mxE lshift0.
  by rewrite unitmx1.
case: hi => g U hg heq hU.
case: bz_mxP => g' M' h1 h2 h3.
constructor.
- rewrite /= (eqd_trans h1) // !mxE lshift0 rshift1.
  apply eqd_gcd; first by apply eqdd.
  apply (eqd_trans hg) => /=.
  by rewrite !mxE rshift1.
- rewrite [lift0_mx U]block_mxEh -[C]vsubmxK -mulmxA !mul_row_col.
  rewrite !(@mul_col_mx _ 1 (1 + n)) mul1mx !mul0mx heq.
  rewrite (@add_col_mx _ 1 (1 + n)) addr0 add0r -(eqP h2).
  congr (_ *m _ ).
  apply/colP => i; rewrite !mxE lshift0.
  case: splitP => j.
  + rewrite [j]ord1 !mxE => h; clear j.
    have -> : i = 0 by apply/ord_inj.
    by rewrite eq_refl lshift0.
  rewrite !mxE => hj.
  have -> /= : i = lift ord0 j by apply/ord_inj.
  case: splitP => k.
  + rewrite [k]ord1 !mxE mulr1n => h; clear k.
    by have -> /= : j = 0 by apply/ord_inj.
  rewrite !mxE => hk.
  by have -> /= : j = lift ord0 k by apply/ord_inj.
by rewrite unitmx_mul h3 (lift0_unitmx hU).
Qed.


CoInductive lemH_spec (n:nat) (L: 'rV[R]_(1 + n)) :
  R * 'M[R]_(1 + n) -> Type :=
| lemHgn0 g M of (g %= gcdsr (row2seq L)) &
  L *m M = row_mx g%:M (0 : 'rV_n) & M \in unitmx : lemH_spec L (g, M).



Lemma row2seq0P : forall n (c: 'rV[R]_n),
  (c == 0) = (row2seq c == nseq n 0).
Proof.
elim => [c | n hi]//=.
- by rewrite [c]thinmx0 !eqxx.
rewrite [n.+1]/(1 + n)%N => c.
rewrite -{1}[c]hsubmxK -{1}[0]hsubmxK [lsubmx c]mx11_scalar !mxE lshift0
  lsubmx0 rsubmx0.
apply/eqP/eqP.
- case/eq_row_mx => /eqP.
  rewrite mx_scalar0 => /eqP -> ->.
  by rewrite row2seq0.
case => -> hc.
f_equal.
- apply/matrixP => i j; rewrite !mxE.
  by rewrite mul0rn.
apply/eqP.
by rewrite hi hc.
Qed.

Lemma size_row2seq : forall n (c: 'rV[R]_n), size (row2seq c) = n.
Proof.
elim => [ | n hi] c //=.
by rewrite hi.
Qed.


Lemma lemHP : forall n (L: 'rV[R]_(1 + n)), lemH_spec L (lemH L).
Proof.
elim => [ | n hi] /=.
- move => C.
  constructor => /=.
  + by rewrite eqd_sym; apply/ gcdr0.
  + by rewrite mulmx1 -{1}[C]hsubmxK [rsubmx C]thinmx0
      [lsubmx C]mx11_scalar !mxE lshift0.
  by rewrite unitmx1.
rewrite [(1 + n.+1)%N]/(1 + (1 + n))%N => C.
case: ifP => /eqP hsub0.
- constructor => /=.
  + rewrite hsub0 rsubmx0 row2seq0 !mxE.
    have -> : gcdr 0 (gcdsr (nseq n 0)) = 0 :> R by
    clear hi C hsub0; elim : n => [ | n hi] /=;
        by apply/eqP; rewrite ?hi -eqdr0; apply gcdr0.
    rewrite lshift0 eqd_sym.
    by apply/gcdr0.
  + by rewrite mulmx1 -{1}[C]hsubmxK hsub0
     [lsubmx C]mx11_scalar !mxE lshift0.
  by rewrite unitmx1.
case: hi => g U hg heq hU.
case: bz_mxP => g' M' h1 h2 h3.
constructor.
- rewrite /= (eqd_trans h1) // !mxE lshift0 rshift1.
  apply eqd_gcd; first by apply eqdd.
  apply (eqd_trans hg) => /=.
  by rewrite !mxE rshift1.
- rewrite -[C]hsubmxK mulmxA !mul_row_col (@mul_mx_row _ 1 1 1 (1 + n))
   (@mul_mx_row _ 1 (1 + n) 1 (1 + n)) !mulmx0 mulmx1
   (@add_row_mx _ 1 1 (1 + n)) add0r addr0 heq.
  move : h2.
  rewrite -{1}[M']trmxK -[\col_i _]trmxK -{1}[g'%:M]trmxK
    -[0 : 'cV[R]_(1 + n)]trmxK -tr_row_mx -trmx_mul.
  rewrite tr_scalar_mx trmx0.
  move/eqP; move/trmx_inj => <-.
  congr (_ *m _ ).
  apply/rowP => i; rewrite !mxE lshift0.
  case: splitP => j.
  + rewrite [j]ord1 !mxE => h; clear j.
    have -> : i = 0 by apply/ord_inj.
    by rewrite eq_refl lshift0.
  rewrite !mxE => hj.
  have -> /= : i = lift ord0 j by apply/ord_inj.
  case: splitP => k.
  + rewrite [k]ord1 !mxE mulr1n => h; clear k.
    by have -> /= : j = 0 by apply/ord_inj.
  rewrite !mxE => hk.
  by have -> /= : j = lift ord0 k by apply/ord_inj.
by rewrite unitmx_mul unitmx_tr h3 (lift0_unitmx hU).
Qed.

Scheme Acc_rect_dep := Induction for Acc Sort Type.

Lemma row_mx_eq0 : forall (n m p:nat) (M: 'M[R]_(n,m)) (N: 'M[R]_(n,p)),
  row_mx M N = 0 -> M = 0 /\ N = 0.
Proof.
move => n m p M N.
rewrite -{1}[0]row_mx0.
by case/eq_row_mx => -> ->.
Qed.

(* Tool function to be able to correctly define the step function
  by well founded induction.
 It's main purpose is to hide (make opaque) the non computational
 part of the function.

 Without it, the correctness of the function is quite impossible to prove
*)
Lemma fun1tool: forall (a a1 a2:R) (n m:nat) (l:'rV[R]_m) (c:'cV[R]_n)
  (M : 'M[R]_(n,m)) (M1: 'M[R]_(1 + m)),
  (a1 %| a2) = false -> (a1 %= gcdsr (row2seq (row_mx a%:M l))) ->
  (a2 %= gcdsr (col2seq (lsubmx (block_mx a%:M l c M *m M1)))) ->
  (row_mx a%:M l *m M1 = row_mx a1%:M 0) ->
  a2 %<| a.
Proof.
move => a a1 a2 n m l c M M1 hd hdvd1 hdvd2 hmul.
rewrite sdvdr_def.
have ha : (a1 %| a).
- move: hdvd1; rewrite /= !mxE.
  case: splitP => i //; rewrite [i]ord1 !mxE mulr1n; clear i => _ hdvd1.
  rewrite (eqd_dvd hdvd1 (eqdd a)).
  by apply dvdr_gcdl.
apply/andP; split => //.
- apply (@dvdr_trans _ a1) => //.
  move : hdvd2; rewrite mul_col_mx hmul /= !mxE.
  case: splitP => i // ; rewrite [i]ord1 lshift0 !mxE; clear i.
  case: splitP => i // ; rewrite [i]ord1 !mxE mulr1n.
  clear i => _ _ hdvd2.
  rewrite (eqd_dvd hdvd2 (eqdd a1)).
  by apply : dvdr_gcdl.
apply/negP => hn.
move/negP : hd; apply.
by apply: (dvdr_trans ha).
Qed.

(* another helper function *)
(*
Lemma fun1tool2: forall (a b: R), a %| b = false -> b != 0.
Proof.
move => a b /negP hdvd.
apply/negP => h.
apply: hdvd.
rewrite (eqP h).
by apply dvdr0.
Qed.
*)
(*
  This function takes a matrix and iterates the following process:
  * simplify the first line using lemH
  * simplify the first column using lemV

  if the new top left element strictly divides the first one, we can
  make a recursive call (well founded induction over divisibility)

  otherwise, that means it divides the first column


  a l1 .... ln              a' 0          0
  c1                        c'1
  .
  .    M             =>*                           a' | c'i
  .
  cp                        c'p
*)

Definition fun1: forall (n m: nat) (a: R),   Acc (@sdvdr R) a ->
 'rV[R]_m -> 'cV[R]_n -> 'M[R]_(n,m) ->
  'M[R]_(1 + n, 1 + m) * 'M[R]_(1 + n) * 'M[R]_(1 + m).
Proof.
move => n m a ha.
elim ha using Acc_rect_dep. clear a ha.
move => a h hi l c M.
case: (lemHP (row_mx a%:M l)).
move => a1 M1 ha1 heq1 hM1.
case: (lemVP (lsubmx ((block_mx a%:M l c M) *m M1))).
move => a2 M2 ha2 heq2 hM2.
move: (@fun1tool a a1 a2).
case hd:   (a1 %| a2).
- move => _; exact ((block_mx a%:M l c M) *m M1, 1%:M , M1).
move => htool.
case : (hi a2 (htool n m l c M M1 (refl_equal false) ha1 ha2 heq1)
           (ursubmx (M2 *m (block_mx a%:M l c M *m M1)))
           (dlsubmx (M2 *m (block_mx a%:M l c M *m M1)))
           (drsubmx (M2 *m (block_mx a%:M l c M *m M1)))) => [[MR Ml] Mr].
exact (MR, Ml *m M2, M1 *m Mr).
Defined.


Lemma gcdsrr : forall (t s: seq R), s = t -> gcdsr s = gcdsr t.
Proof.
by move => t s ->.
Qed.

Lemma gcdr_dvdl : forall (a b c: R), a %= gcdr b c -> a %| b.
Proof.
move => a b c h.
rewrite (eqd_dvd h (eqdd b)).
by apply dvdr_gcdl.
Qed.

Lemma gcdr_dvdr : forall (a b c: R), a %= gcdr b c -> a %| c.
Proof.
move => a b c h.
rewrite (eqd_dvd h (eqdd c)).
by apply dvdr_gcdr.
Qed.

Lemma fun1P: forall n m a (ha : Acc (@sdvdr R) a)
  (l :'rV[R]_m) (c: 'cV[R]_n) (M0 :'M[R]_(n,m)),
  let: (M,Ml,Mr) := fun1 ha l c M0 in
  let: g := M 0 0 in [/\ ursubmx M = 0 ,
    (forallb i: 'I_n, g %| (dlsubmx M) i 0) ,
    Ml \in unitmx , Mr \in unitmx &
    M = Ml *m (block_mx a%:M l c M0) *m Mr ].
Proof.
move => n m a ha.
elim ha using Acc_rect_dep. clear a ha.
move => a h hi l c M /=.
case: (lemHP (row_mx a%:M l)) => a1 M1 hdvd hmul hunit1.
case: (lemVP (lsubmx ((block_mx a%:M l c M) *m M1)))
  => a2 M2 hdvd2 hmul2 hunit2.
move: (@fun1tool a a1 a2).
case hd:  (a1 %| a2).
- move => _.
  move: hdvd2.
  rewrite !mul_col_mx !hmul /ursubmx col_mxKu row_mxKr /dlsubmx col_mxKd
    unitmx1 hunit1 => hdvd2 ; split => //.
  + move : hdvd2.
    rewrite (@gcdsrr (col2seq (col_mx a1%:M (lsubmx (row_mx c M *m M1)))))
      /= col_mxKd.
    * move/gcdr_dvdr; rewrite dvdr_gcds.
      move/allP => hdvd2.
      apply/forallP => i.
      set Ci := (lsubmx (row_mx c M *m M1)) i 0.
      rewrite !mxE.
      case: splitP => j //; rewrite [j]ord1 !mxE => _; clear j.
      case: splitP => j //; rewrite [j]ord1 !mxE mulr1n => _; clear j.
      apply (dvdr_trans hd).
      apply : hdvd2.
      rewrite in_col2seq.
      apply/existsP; exists i.
      by rewrite /Ci.
    rewrite !mxE.
    case: splitP => j //; rewrite [j]ord1 lshift0 !mxE mulr1n => _; clear j.
    case: splitP => j //; rewrite [j]ord1 !mxE mulr1n => _; clear j.
    congr (_ :: _).
    f_equal.
    apply/colP => i; rewrite !mxE rshift1.
    case: splitP => j.
    * by rewrite [j]ord1.
    move => hj.
    have -> : i = j.
    * apply/ord_inj; apply/eqP.
      by rewrite -eqSS -[j.+1]add1n -hj.
    by rewrite lshift0 !mxE.
  by rewrite mul1mx mul_col_mx hmul.
move => htool.
case: fun1 (hi a2 (htool n m l c M M1 (refl_equal false) hdvd hdvd2 hmul)
           (ursubmx (M2 *m (block_mx a%:M l c M *m M1)))
           (dlsubmx (M2 *m (block_mx a%:M l c M *m M1)))
           (drsubmx (M2 *m (block_mx a%:M l c M *m M1))))
  => [[MR Ml] Mr] [h1 h2 h3 h4 h5].
split => //.
- by rewrite unitmx_mul hunit2 h3.
- by rewrite unitmx_mul hunit1 h4.
move: h5.
replace
  (@scalar_mx (DvdRing.ringType (PrincipalRing.dvdRingType R)) (S O) a2)
  with (ulsubmx (M2 *m (block_mx a%:M l c M *m M1))).
- by rewrite submxK !mulmxA.
rewrite -[block_mx _ _ _ _ *m _]hsubmxK mul_mx_row hmul2.
apply/matrixP => i j; rewrite [i]ord1 [j]ord1 !mxE lshift0 /=; clear i j.
case: splitP => k //.
rewrite [k]ord1 lshift0 !mxE ; clear k.
by case: splitP => i //; rewrite [i]ord1 !mxE mulr1n.
Qed.



(* from a matrix of the shape
    a  0  ...  0
    c1
    .
    .    M
    .
    cn
where a | ci, then we build the matrix

    a 0 ... 0
    0
    .
    .    M
    .
    0
*)
Definition fun2 (n m:nat) (a: R) (M: 'M[R]_(1 + n,1+m)):
  'M[R]_(1 + n, m) * 'M[R]_(1 + (1 + n)) :=
  let: (g,U) := lemV (lsubmx M) in
  if (g == 0) then (rsubmx M, 1%:M) else
  let N := U *m M in
   let U' := block_mx 1 0 (col_mx (-(odflt 0 (g %/? a)))%:M 0) 1%:M in
     (rsubmx N, U' *m lift0_mx U).

Lemma fun2P : forall (n m:nat) (a:R) (c: 'cV[R]_(1 + n))
  (M: 'M[R]_(1 + n,m)), (forallb i, a %| c i 0) ->
  let: (Mdr, U) := fun2 a (row_mx c M) in
    block_mx a%:M 0 0 Mdr = U *m (block_mx a%:M 0 c M) /\
    U \in unitmx.
Proof.
rewrite /fun2 => n m a c M /=.
move/forallP => hdvd.
case: lemVP.
rewrite row_mxKl row_mxKr => g U hgcd hmul hunit.
case h0 : (g == 0).
- by move : hgcd; rewrite (eqP h0) eqd0r gcdsr_eq0 size_col2seq
   -col2seq0P mul1mx unitmx1 => /eqP ->.
have : a %| g.
- rewrite (eqd_dvd (eqdd a) hgcd).
  rewrite dvdr_gcds.
  apply/allP => x.
  rewrite in_col2seq.
  case/existsP => j hx.
  rewrite (eqP hx); by apply: hdvd.
case/dvdrP => q hq.
split.
- rewrite -mulmxA mulmx_block !mul1mx !mul0mx !add0r addr0 hmul.
  rewrite mulmx_block !mul1mx !mul0mx mulmx0 !add0r addr0.
  f_equal.
  + case: odivrP.
    * move => d hd /=.
      rewrite mul_col_mx mul0mx add_col_mx addr0 -scalar_mxM mulNr -hd.
      apply/colP => i; rewrite !mxE.
      case: splitP => j.
      - by rewrite [j]ord1 !mxE !mulr1n addNr.
      by rewrite !mxE.
    move => hf; move/negP: (hf q).
    by case; rewrite hq.
  by rewrite mul_mx_row row_mxKr.
rewrite unitmx_mul (lift0_unitmx hunit) andbT.
have := (@mulmx1_unit R _
  (@block_mx R 1 (1 + n) 1 (1 + n)
    1 0 (col_mx (- odflt 0 (g%/? a))%:M 0) 1%:M)
  (@block_mx R 1 (1 + n) 1 (1 + n)
    1 0 (col_mx (  odflt 0 (g%/? a))%:M 0) 1%:M)).
case.
- rewrite mulmx_block !mul1mx !mulmx1 !mul0mx !mulmx0 !addr0 !add0r.
  rewrite add_col_mx -{1}idmxE addr0.
  rewrite [1%:M : 'M[R]_(1 + (1 + n))]scalar_mx_block.
  f_equal.
  apply/colP => i; rewrite !mxE.
  case: splitP => j.
  + by rewrite [j]ord1 !mxE /= !mulr1n addNr.
  by rewrite !mxE.
move => h1 _; exact h1.
Qed.


(* we do fun1 then fun2 to transform a block matrix into
   a 0
   0 M
*)
Definition step1_ (n m:nat)
  (M : 'M[R]_(1 + (1 + n), 1 + m))
 : R * 'M[R]_(1 + n, m) * 'M[R]_(1 + (1 + n)) * 'M[R]_(1 + m) :=
  let: (M1,Ul,Ur) := fun1 (sdvdr_wf (M 0 0))  (ursubmx M)
                          (dlsubmx M) (drsubmx M) in
  let: (M2, V) := fun2 (M1 0 0) (dsubmx M1) in
    (M1 0 0, M2, V *m Ul, Ur).

Lemma step1_P : forall (n m:nat) (M : 'M[R]_(1 + (1 + n), 1 + m)),
 let: (a, N, U, V) := step1_ M in
   [/\ U \in unitmx , V \in unitmx &
     block_mx a%:M 0 0 N = U *m (M *m V)].
Proof.
move => n m M; rewrite /step1_.
move : (fun1P (sdvdr_wf (M 0 0)) (ursubmx M)
  (dlsubmx M) (drsubmx M)).
case: fun1 => [[M1 Ul] Ur] [hur hdvd hunit1 hunit2 heq].
move: (fun2P (drsubmx M1) hdvd).
rewrite hsubmxK.
case: fun2 => Mdr U [heq2 hunit3].
split => //.
- by rewrite unitmx_mul hunit1 hunit3.
move: heq.
replace
  (@scalar_mx (DvdRing.ringType (PrincipalRing.dvdRingType R)) (S O)
    (M 0 0))
  with (ulsubmx M).
- rewrite submxK -{1}[Ul *m M *m Ur]mulmxA -[U *m Ul *m (M *m Ur)]mulmxA.
  move => <-.
  rewrite heq2 -[0 : 'rV[R]_m]hur.
  replace
   (@scalar_mx (PrincipalRing.ringType R) (S O)
              (@fun_of_matrix (PrincipalRing.sort R)
                 (addn (S O) (addn (S O) n)) (addn (S O) m) M1
                 (GRing.zero (Zp_zmodType (addn (S O) n)))
                 (GRing.zero (Zp_zmodType m))))
    with (ulsubmx M1).
  + by rewrite submxK.
  apply/matrixP => i j; rewrite [i]ord1 [j]ord1 !mxE lshift0 /=; clear i j.
  by rewrite lshift0 mulr1n.
apply/matrixP => i j; rewrite [i]ord1 [j]ord1 !mxE lshift0 /=; clear i j.
by rewrite lshift0 mulr1n.
Qed.


(*
   transform a c  intro a' 0
             l M        0  M'
*)
Definition step (n m:nat) : forall (M: 'M[R]_(1 + n, 1 + m)),
   R * 'M[R]_(1 + n,1 + m) * 'M[R]_(1 + n) * 'M[R]_(1 + m) :=
 match n return forall (M: 'M[R]_(1 + n, 1 + m)),
   R * 'M[R]_(1 + n,1 + m) * 'M[R]_(1 + n) * 'M[R]_(1 + m) with
 | O => fun (M : 'M[R]_(1, _))  =>
   let: (g,U) := lemH M in (g, row_mx g%:M 0, 1%:M, U)
 | S p => fun (M: 'M[R]_(1 + (1 + _), 1 + _)) =>
     let: (g,P,U,V) := step1_ M in
      (g, block_mx g%:M 0 0 P, U, V)
end.


Lemma ursubmx0 : forall (n m p q:nat), ursubmx (0 : 'M[R]_(n+m,p+q)) = 0.
Proof.
move => n m p q.
by apply/matrixP => i j; rewrite !mxE.
Qed.

Lemma dlsubmx0 : forall (n m p q:nat), dlsubmx (0 : 'M[R]_(n+m,p+q)) = 0.
Proof.
move => n m p q.
by apply/matrixP => i j; rewrite !mxE.
Qed.

Lemma row_mxKur0 : forall m n p q (M: 'M[R]_(m + n,p + q)),
 rsubmx M = 0 -> ursubmx M = 0.
Proof.
move => m n p q M /matrixP h.
apply/matrixP => i j; rewrite !mxE.
by move: (h (lshift n i) j); rewrite !mxE.
Qed.


CoInductive step_spec (n m:nat) (M: 'M[R]_(1 + n,1 + m)) :
  R * 'M[R]_(1 + n, 1 + m) * 'M[R]_(1 + n) * 'M[R]_(1 + m) -> Type :=
SS a (N: 'M[R]_(1 + n, 1 + m)) Ml Mr of
    N = Ml *m M *m Mr
  & a = N 0 0
  & ursubmx N = 0
  & dlsubmx N = 0
  & Ml \in unitmx
  & Mr \in unitmx : step_spec M (a,N,Ml,Mr).

Lemma stepP: forall n m (M: 'M[R]_(1 + n,1 + m)),
  step_spec M (step M).
Proof.
rewrite /step => n m.
case: n => [ | n] M.
- case: lemHP => g U h3 h4 h5; split => //; try by rewrite unitmx1.
  + by rewrite mul1mx h4.
  + rewrite !mxE.
    case: splitP => x //; by rewrite [x]ord1 !mxE.
  + by rewrite row_mxKur0 // row_mxKr.
  by apply/colP => [[]].
move: (step1_P M).
case: step1_  => [[[g N] U] V] [h1 h2 h3].
constructor => //.
- by rewrite -mulmxA.
- rewrite !mxE.
  case: splitP => x //; rewrite [x]ord1 !mxE; clear x.
  by case: splitP => x //; rewrite [x]ord1 !mxE.
- by rewrite block_mxKur.
by rewrite block_mxKdl.
Qed.

(* Half Smith: takes a matrix and transform it into a
 "diagonal" matrix (even non square matrices !!)
*)
Fixpoint half_smith (n m:nat) {struct n} :
 'M[R]_(n,m) -> 'M[R]_(n,m) * 'M[R]_n * 'M[R]_m :=
 match n,m return  'M[R]_(n,m) -> 'M[R]_(n,m) * 'M[R]_n * 'M[R]_m with
   | S p, S q => fun (M: 'M[R]_(1 + _, 1 + _ )) =>
     let: (a,N,U,V) := step M in
       let: (M', U1, V1) := half_smith (drsubmx N) in
         (block_mx (a%:M:'M[R]_1) 0 0 M',
           lift0_mx U1 *m U, V *m lift0_mx V1)
   | _,_ => fun _ => (0, 1%:M, 1%:M)
end.


(*
 Extract the lists of the M i i for any matrices (even non square ones)
*)
Fixpoint ExtractDiag (n m:nat) : 'M[R]_(n,m) -> seq R :=
 match n,m return 'M[R]_(n,m) -> seq R with
  | S p, S q => fun (M : 'M[R]_(1 + _ ,1 + _ )) =>
    M 0 0 :: ExtractDiag (drsubmx M)
  | _,_ => fun _ => [::]
end.

Lemma ExtractDiag_size : forall (n m:nat) (M: 'M[R]_(n,m)),
  size (ExtractDiag M) = minn n m.
Proof.
elim => [ | n hi] [ | m] M //=.
rewrite hi.
case hnm : (n <= m).
- by rewrite !minnl.
move/negbT: hnm; rewrite -ltnNge.
move/ltnW => hnm.
by rewrite !minnr.
Qed.


(*
 A (non square) matrix is diagonal if it is of the shape
   a l
   c M

with l = c = 0 and M is (recursively) diagonal
*)
Fixpoint is_diag (n m:nat) : 'M[R]_(n,m) -> bool :=
 match n,m return 'M[R]_(n,m) -> bool with
  | S p, S q => fun (M: 'M[R]_(1 + _, 1 + _)) =>
   [&& ursubmx M == 0 ,
       dlsubmx M == 0 &
       is_diag (drsubmx M)]
  | _,_ => fun _ => true
end.


Lemma is_diag_non_diag : forall n m (M: 'M[R]_(n,m)) i j,
  is_diag M -> (nat_of_ord i != nat_of_ord j) -> M i j = 0.
Proof.
elim => [ m M [] j _ _ // | n hi].
case => [ M j [] // | m ].
rewrite [n.+1]/(1 + n)%N [m.+1]/(1 + m)%N => M i j /=.
case/and3P => /eqP hur /eqP hdl hD.
case: (splitP i) => x.
- rewrite [x]ord1 => hx; clear x.
  have -> : i = 0 by apply/ord_inj.
  case: (splitP j) => y.
  + rewrite [y]ord1 => hy; clear y.
    have -> : j = 0 by apply/ord_inj.
    by case/negP.
  move => hj _.
  have -> : j = lift 0 y by apply/ord_inj.
  move/matrixP : hur => hur.
  by move: (hur 0 y) ; rewrite !mxE lshift0 rshift1.
move => hx.
have -> : i = lift 0 x by apply/ord_inj.
case: (splitP j) => y.
- rewrite [y]ord1 => hy; clear y.
  have -> : j = 0 by apply/ord_inj.
  move/matrixP: hdl => hdl.
  by move: (hdl x 0); rewrite !mxE lshift0 rshift1.
move => hj.
have -> : j = lift 0 y by apply/ord_inj.
move => hlift.
have hxy : nat_of_ord x != nat_of_ord y.
- apply/negP => h.
  move/negP : hlift; apply => /=.
  by rewrite (eqP h).
move : (hi _ (drsubmx M) x y hD hxy).
by rewrite !mxE !rshift1.
Qed.

(* or we can say that a matrix is diagonal if

  M i j = 0 for any i <> j
*)
Lemma is_diagP : forall n m (M: 'M[R]_(n,m)),
 reflect (forall i j, nat_of_ord i != nat_of_ord j -> M i j = 0)
          (is_diag M).
Proof.
elim => [ m M | n hi].
- rewrite [M]flatmx0.
  by apply: (iffP idP) => // _ [[]].
case => [ | m] M.
- rewrite [M]thinmx0.
  by apply: (iffP idP) => // _ i [[]].
apply: (iffP idP).
- move => h i j h'.
  by apply is_diag_non_diag.
move => h /=.
apply/and3P; split.
- by apply/eqP/rowP => i; rewrite !mxE lshift0 rshift1 h.
- by apply/eqP/colP => i; rewrite !mxE lshift0 rshift1 h.
apply/hi => a b hab.
by rewrite !mxE !rshift1 h.
Qed.

CoInductive half_smith_spec (n m:nat) (M: 'M[R]_(n,m)) :
  'M[R]_(n,m) * 'M[R]_n * 'M[R]_m -> Type :=
HSS N Ml Mr of N = Ml *m M *m Mr
  & is_diag N
  & Ml \in unitmx
  & Mr \in unitmx : half_smith_spec M (N,Ml,Mr).


Lemma half_smithP : forall n m (M: 'M[R]_(n,m)),
  half_smith_spec M (half_smith M).
Proof.
elim => [ m M | n hi].
- rewrite [M]flatmx0 /=; constructor => //; try by rewrite unitmx1.
  by rewrite mul1mx mulmx1.
move => [ M | m].
- rewrite [M]thinmx0 /=; constructor => //; try by rewrite unitmx1.
  by rewrite mul1mx mulmx1.
rewrite [n.+1]/(1 + n)%N [m.+1]/(1 + m)%N => M /=.
case: stepP => a N Ml Mr heq1 ha0 hur hdl hMl hMr.
case: hi => N' Ml' Mr' heq2 hdiag hMl' hMr'.
constructor => /=.
- have -> : lift0_mx Ml' *m Ml *m M *m (Mr *m lift0_mx Mr') =
    lift0_mx Ml' *m N *m lift0_mx Mr'.
  + rewrite !mulmxA.
    congr (_ *m _).
    rewrite -!mulmxA.
    congr (_ *m _).
    by rewrite mulmxA.
  by rewrite -[N]submxK !mulmx_block hdl hur !mul0mx !mulmx0 !addr0 !add0r
    !mulmx1 !mul1mx mul0mx heq2 [ulsubmx N]mx11_scalar !mxE !lshift0 -ha0.
- by rewrite block_mxKur block_mxKdl block_mxKdr hdiag !eqxx.
- by rewrite unitmx_mul (lift0_unitmx hMl') hMl.
by rewrite unitmx_mul (lift0_unitmx hMr') hMr.
Qed.


(* From a diagonal matrix, this function enforces the divisibility
   requirement of the SNF

  note: to prove its correctness, we need to enforce that
  if N = process_diag M, then N is the gcd of the diagonal of M
*)
Fixpoint process_diag {n m:nat} : 'M[R]_(1 + n, 1 + m) ->
  'M[R]_(1 + n, 1 + m) * 'M_(1 + n) * 'M_(1 + m) :=
 match n,m return
   'M[R]_(1 + n, 1 + m) -> 'M[R]_(1+n,1+m) * 'M_(1 + n) * 'M_(1 + m) with
  | p.+1,q.+1 => fun (M: 'M[R]_(1 + _, 1 + _)) =>
    let a := M 0 0 in
    let bs := drsubmx M in
    let: (cs,Lc,Rc) := process_diag bs in
    let: (g1,u,v,a1,b1) := egcdr a (cs 0 0) in
(* we only have g %= gcd so we can't use directly lcmr, but ab / g
    let g2 := lcmr a (cs 0 0) in
*)
    let g2 := a1 * (cs 0 0) in
    let: (ds,Ld, Rd) := process_diag (block_mx g2%:M 0 0 (drsubmx cs)) in

    let R1 := 1%:M + delta_mx 1 0 in
    let (_,L1) := @bz_mx a (cs 0 0) (1 + p) 0 in
    let R2 := 1 - (b1 * v) *: delta_mx 0 1 in
      (block_mx g1%:M 0 0 ds, lift0_mx Ld *m L1 *m lift0_mx Lc,
        lift0_mx Rc *m R1 *m R2 *m lift0_mx Rd)
  | 0, q.+1 => fun (l: 'rV[R]_(1 + _)) =>
    let (_,U) := lemH l in
      (l *m U ,1%:M,U)
  | p.+1, 0 => fun (c: 'cV[R]_(1 + _)) =>
    let (_,U) := lemV c in
      (U *m c,U,1%:M)
  | _,_ => fun M => (M, 1%:M, 1%:M)
end.


(* Divisibility requirement: for all i,
  M i i | M i+1 i+1 *)
Fixpoint DvDPred (n m:nat) : 'M[R]_(1 + n,1 + m) -> bool :=
 match n,m return 'M[R]_(1 + n,1 + m) -> bool with
  | p.+1, q.+1 => fun (M: 'M[R]_(1 + _, 1 +  _)) =>
     if M 0 0 %| M (lift 0 0) (lift 0 0) then
       DvDPred (drsubmx M)
     else false
  | _,_ => fun _ => true
end.

Lemma DvD_trans : forall (n m:nat) (M: 'M[R]_(1 + n, 1 + m))
 (i: 'I_n) (j: 'I_m), nat_of_ord i = nat_of_ord j -> DvDPred M ->
 M 0 0 %| M (lift 0 i) (lift 0 j).
Proof.
elim => [ m M [] j | n hi] //.
case => [ M i [] | m] //.
rewrite [(1 + n.+1)%N]/(1 + (1 + n))%N [(1 + m.+1)%N]/(1 + (1 + m))%N
  [n.+1]/(1 + n)%N [m.+1]/(1 + m)%N => M i j.
case: (splitP i) => x.
- rewrite [x]ord1 => hx; clear x.
  have -> : i = 0 by apply/ord_inj.
  case: (splitP j) => y.
  + rewrite [y]ord1 => hy /= ; clear y.
    have -> : j = 0 by apply/ord_inj.
    by case: ifP.
  move => hj.
  by have -> : j = lift 0 y by apply/ord_inj.
move => hix /=.
have -> : i = lift 0 x by apply/ord_inj.
case: ifP => hif //=.
case: (splitP j) => y.
- rewrite [y]ord1 => hy.
  by have -> : j = 0 by apply/ord_inj.
move => hy.
have -> /= : j = lift 0 y by apply/ord_inj.
move/eqP; rewrite /= eqSS => /eqP hxy hD.
apply (dvdr_trans hif).
move: (hi _ (drsubmx M) x y hxy hD).
by rewrite !mxE !rshift1.
Qed.


Lemma DvDPredP : forall n m (M: 'M[R]_(1 + n, 1 + m)),
  reflect (forall i j, nat_of_ord i = nat_of_ord j ->
    M (widen_ord (leqnSn n) i) (widen_ord (leqnSn m) j)
      %| M (lift 0 i) (lift 0 j)) (DvDPred M).
Proof.
elim => [ m M | n hi].
- by apply: (iffP idP) => //= _ [[]].
case => [ M | m ].
- by apply: (iffP idP) => //= _ i [[]].
rewrite [n.+1]/(1 + n)%N [m.+1]/(1 + m)%N => M.
apply: (iffP idP).
- move => /= hD.
  rewrite [n.+1]/(1 + n)%N [m.+1]/(1 + m)%N => i j.
  case: (splitP i) => x.
  + rewrite [x]ord1 => hx; clear x.
    have -> : i = 0 by apply/ord_inj.
    case: (splitP j) => y.
    * rewrite [y]ord1 => hy; clear y.
      have -> : j = 0 by apply/ord_inj.
      have -> : widen_ord (leqnSn (1 + n)%N) 0 = 0 by apply/ord_inj.
      have -> : widen_ord (leqnSn (1 + m)%N) 0 = 0 by apply/ord_inj.
      by case: ifP hD.
    move => h.
    by have -> : j = lift 0 y by apply/ord_inj.
  move => hx.
  have -> : i = lift 0 x by apply/ord_inj.
  case: (splitP j) => y.
  + rewrite [y]ord1 => hy; clear y.
    by have -> : j = 0 by apply/ord_inj.
  move => hy.
  have -> : j = lift 0 y by apply/ord_inj.
  move/eqP; rewrite eqSS => /eqP hxy.
  move: hD.
  case: ifP => // _.
  move/hi => hD.
  move: (hD x y hxy); rewrite !mxE !rshift1 => h.
  have -> : widen_ord (leqnSn (1 + n)%N) (lift 0 x) =
    lift 0 (widen_ord (leqnSn n) x) by apply/ord_inj.
  by have -> : widen_ord (leqnSn (1 + m)%N) (lift 0 y) =
    lift 0 (widen_ord (leqnSn m) y) by apply/ord_inj.
move => h /=.
case: ifP => hif.
- apply/hi => i j hij; rewrite !mxE !rshift1.
  have -> : lift 0 (widen_ord (leqnSn n) i) =
    widen_ord (leqnSn (1 + n)%N) (lift 0 i) by apply/ord_inj.
  have -> : lift 0 (widen_ord (leqnSn m) j) =
    widen_ord (leqnSn (1 + m)%N) (lift 0 j) by apply/ord_inj.
  apply: h => /=.
  by rewrite hij.
rewrite -hif.
have {1}-> : 0 = widen_ord (leqnSn (1 + n)%N) 0 by apply/ord_inj.
have {1}-> : 0 = widen_ord (leqnSn (1 + m)%N) 0 by apply/ord_inj.
by apply: h.
Qed.


CoInductive process_diag_spec (n m:nat) (M: 'M[R]_(1 + n, 1 + m)) :
  'M[R]_(1 + n, 1 + m) * 'M[R]_(1 + n) * 'M[R]_(1 + m) -> Type :=
 ProcessDiagSpec N Ml Mr of
   is_diag M -> [/\
     (N = Ml *m M *m Mr /\ is_diag N),
     N 0 0 %= gcdsr (ExtractDiag M) ,
     DvDPred N,
     Ml \in unitmx &
     Mr \in unitmx] : process_diag_spec M (N,Ml,Mr).

Lemma is_diag_dr : forall n m (M: 'M[R]_(1 + n, 1 + m)),
  is_diag M -> is_diag (drsubmx M).
Proof.
move => n m M /=.
by case/and3P.
Qed.

Lemma in_ExtractDiag : forall n m (M: 'M[R]_(n,m)) a,
  a \in ExtractDiag M -> exists i: 'I_n , exists j: 'I_m,
    (nat_of_ord i = nat_of_ord j) /\ M i j = a.
Proof.
elim => [ m M a | n hi].
- by rewrite [M]flatmx0 /= in_nil.
case => [ M a | m ].
- by rewrite [M]thinmx0 /= in_nil.
rewrite [(1 + n.+1)%N]/(1 + (1 + n))%N [(1 + m.+1)%N]/(1 + (1 + m))%N /=
 => M a.
rewrite in_cons.
case/orP => hu.
- by exists 0; exists 0; rewrite (eqP hu).
case: (hi _ _ _ hu) => x [y [heq]].
rewrite !mxE => h.
exists (rshift 1 x); exists (rshift 1 y); split => //=.
by rewrite heq.
Qed.



Lemma inv_delta_mx : forall n (a:R) (x y: 'I_(1 + n)),
  x == y = false ->
  (1%:M + a *: delta_mx x y) *m (1%:M - a *: delta_mx x y) = 1%:M.
Proof.
move => n a x y hneq.
rewrite mulmx_addl mul1mx mulmx_subr mulmx1.
by rewrite -!addrA [- _ + _]addrA addNr sub0r -scalemxAl -scalemxAr
 mul_delta_mx_cond eq_sym hneq mulr0n !scaler0 subr0.
Qed.

Lemma unit_delta_mx : forall n (a : R) (x y: 'I_(1 + n)),
  x == y = false ->
   (1%:M + a *: delta_mx x y) \in unitmx.
Proof.
move => n a x y h.
by case: (mulmx1_unit (@inv_delta_mx n a x y h)).
Qed.

Lemma s1 :  forall n m a (M: 'M[R]_(1 + n,1 + m)),
   block_mx a%:M 0 0 M *m
     (1%:M  + (@delta_mx _ (1 + (1 + m)) (1 + (1 + m)) 1 0)) =
   block_mx a%:M 0 (col 0 M) M.
Proof.
move => n m a M.
have -> : (1%:M + delta_mx 1 0) = block_mx 1%:M 0 (col_mx 1%:M 0) 1%:M
  :> 'M[R]_(1 + (1 + m), 1 + (1 + m)).
- apply/matrixP => i j; rewrite !mxE.
  case: splitP => ii.
  + rewrite [ii]ord1 !mxE; clear ii => hi.
    have -> /= : i = 0 by apply/ord_inj.
    case: splitP => jj.
    * rewrite [jj]ord1 addr0 !mxE => hj.
      by have -> : j = 0 by apply/ord_inj.
    rewrite !mxE addr0 => hj.
    by have -> : j = lift 0 jj by apply/ord_inj.
  rewrite !mxE => hi.
  have -> /= : i = lift 0 ii by apply/ord_inj.
  case: splitP => jj.
  + rewrite [jj]ord1 !mxE => hj.
    have -> /= : j = 0 by apply/ord_inj.
    case: splitP => iii.
    * rewrite [iii]ord1 add0r andbT !mxE mulr1n => hii.
      by have -> /= : ii = 0 by apply/ord_inj.
    rewrite !mxE add0r andbT => hii.
    by have -> : ii = lift 0 iii by apply/ord_inj.
  rewrite !mxE => hjj.
  have -> /= : j = lift 0 jj by apply/ord_inj.
  by rewrite andbF addr0.
rewrite !mulmx_block !mul0mx !mulmx0 !addr0 !add0r !mulmx1.
f_equal.
apply/matrixP => i j; rewrite [j]ord1 !mxE; clear j.
rewrite !big_ord_recl !mxE.
case: splitP => x //; rewrite [x]ord1 !mxE mulr1 => _; clear x.
rewrite big1 ?addr0 // => z _.
rewrite !mxE.
case: splitP => y; first by rewrite [y]ord1.
by rewrite !mxE mulr0.
Qed.


Lemma s2 :  forall n m a b c d (M: 'M[R]_(n,m)),
   block_mx a%:M (row_mx b%:M 0) 0 (block_mx c%:M 0 0 M) *m
     (1%:M  - d *: (@delta_mx _ (1 + (1 + m)) (1 + (1 + m)) 0 1)) =
   block_mx a%:M (row_mx (- a * d + b)%:M 0) 0
   (block_mx c%:M 0 0 M).
Proof.
move => n m a b c d M.
have -> : 1%:M - d *: delta_mx 0 1 =
  block_mx 1%:M (row_mx (-d)%:M 0) 0 1%:M :> 'M[R]_(1 + (1 + m)).
- apply/matrixP => i j; rewrite !mxE.
  case: splitP => ii.
  + rewrite [ii]ord1 !mxE; clear ii => hi.
    have -> /= : i = 0 by apply/ord_inj.
    case: splitP => jj.
    * rewrite [jj]ord1 !mxE => hj.
      have -> : j = 0 by apply/ord_inj.
      by rewrite mulr0 subr0.
    rewrite !mxE => hj.
    have -> /= : j = lift 0 jj by apply/ord_inj.
    case: splitP => x.
    * rewrite [x]ord1 !mxE => hx; clear x.
      have -> /= : jj = 0 by apply/ord_inj.
      by rewrite sub0r mulr1.
    rewrite !mxE => hx.
    have -> : jj = lift 0 x by apply/ord_inj.
    by rewrite mulr0 subr0.
  rewrite !mxE => hi.
  have -> /= : i = lift 0 ii by apply/ord_inj.
  rewrite mulr0 subr0.
  case: splitP => jj.
  + rewrite [jj]ord1 !mxE => hj.
    by have -> /= : j = 0 by apply/ord_inj.
  rewrite !mxE => hjj.
  by have -> /= : j = lift 0 jj by apply/ord_inj.
rewrite !mulmx_block !mulmx0 !mul0mx !addr0 add0r !mulmx1.
f_equal.
rewrite mul_scalar_mx scale_row_mx scaler0 add_row_mx addr0.
by rewrite scale_scalar_mx !mulNr !mulrN -[(- (a * d))%:M]scalemx1
     -[b%:M]scalemx1 -scaler_addl scalemx1.
Qed.

Lemma is_diag_col0 : forall n m (M: 'M[R]_(1 + n, 1 + m)),
  is_diag M -> col 0 M = col_mx (M 0 0)%:M 0.
Proof.
move => n m M /=.
case/and3P => /eqP hur /eqP hdl hd.
rewrite -{1}[M]submxK hur hdl.
apply/colP => i; rewrite !mxE.
case: splitP => ii.
- rewrite [ii]ord1 !mxE => hi; clear ii.
  case: splitP => x //; rewrite [x]ord1 !mxE => _; clear x.
  by rewrite !lshift0.
rewrite !mxE => hi.
by case: splitP => x //; rewrite [x]ord1 !mxE => _; clear x.
Qed.

Lemma bz0 : forall (u v a1 b1 : R) (n : nat),
 @combine_mx u v (-b1) a1 (1 + n) 0 =
   block_mx u%:M (row_mx v%:M 0) (col_mx (- b1)%:M 0)
     (block_mx a1%:M 0 0 1%:M) :> 'M[R]_(1 + (1 + n)).
Proof.
rewrite /combine_mx => u v a1 b1 n.
apply/matrixP => i j; rewrite !mxE.
case: splitP => ii.
- rewrite [ii]ord1 !mxE => hii; clear ii.
  have -> /= : i = 0 by apply/ord_inj.
  case: splitP => jj.
  + rewrite [jj]ord1 !mxE => hjj; clear jj.
    have -> /= : j = 0 by apply/ord_inj.
    by rewrite !mulr1n !mulr0n !mulr0 !addr0.
  rewrite !mxE => hjj.
  have -> /= : j = lift 0 jj by apply/ord_inj.
  case: splitP => x.
  + rewrite [x]ord1 !mxE => hx; clear x.
    have -> /= : jj = 0 by apply/ord_inj.
    by rewrite !mulr1n !mulr0n mulr0 !add0r mulr1.
  rewrite !mxE => hx.
  have -> : jj = lift 0 x by apply/ord_inj.
  by rewrite !mulr1n !mulr0n !mulr0 !add0r.
rewrite !mxE => hii.
have -> /= : i = lift 0 ii by apply/ord_inj.
case: splitP => jj.
- rewrite [jj]ord1 !mxE => hjj; clear jj.
  have -> /= : j = 0 by apply/ord_inj.
  case: splitP => x.
  + rewrite [x]ord1 !mxE => hx; clear x.
    have -> /= : ii = 0 by apply/ord_inj.
    by rewrite !mulr1n !mulr0n mulr0 !add0r mulr1 addr0.
  rewrite !mxE => hx.
  have -> : ii = lift 0 x by apply/ord_inj.
  by rewrite !mulr1n !mulr0n !mulr0 !add0r.
rewrite !mxE => hjj.
have -> /= : j = lift 0 jj by apply/ord_inj.
case: splitP => x.
- rewrite [x]ord1 !mxE => hx; clear x.
  have -> /= : ii = 0 by apply/ord_inj.
  case: splitP => y.
  + rewrite [y]ord1 !mxE => hy; clear y.
    have -> /= : jj = 0 by apply/ord_inj.
    by rewrite !mulr1n !mulr0n !mulr0 !addr0 add0r.
  rewrite !mxE => hy.
  have -> /= : jj = lift 0 y by apply/ord_inj.
  by rewrite !mulr1n !mulr0n !mulr0 !addr0.
rewrite !mxE => hx.
have -> /= : ii = lift 0 x by apply/ord_inj.
case: splitP => y.
- rewrite [y]ord1 !mxE => hy; clear y.
  have -> /= : jj = 0 by apply/ord_inj.
  by rewrite !mulr1n !mulr0n !mulr0 !addr0.
rewrite !mxE => hy.
have -> /= : jj = lift 0 y by apply/ord_inj.
by rewrite !mulr1n !mulr0n !mulr0 !addr0 !add0r.
Qed.



Lemma process_diagP : forall n m (M: 'M[R]_(1 + n, 1 + m)),
  process_diag_spec M (process_diag M).
Proof.
elim => [ | n hi].
- case => [ | m] M.
  (* 1 x 1 *)
  + constructor => hdiag /=; split => //; try by rewrite unitmx1.
    * split; first by rewrite mul1mx mulmx1.
      by rewrite [ursubmx M]thinmx0 [dlsubmx M]flatmx0 !eqxx.
    by rewrite eqd_sym; apply gcdr0.
  (* 1 x 1 + 1 + m *)
  rewrite /process_diag.
  case: lemHP => k U hk heq hU.
  constructor => hdiag.
  rewrite hU mul1mx unitmx1; split => //.
  + split => //.
    rewrite heq /= ?row_mxKur0 ?eqxx /= ?andbT ?row_mxKr //.
    by apply/eqP/matrixP => [[]].
  move: hdiag => /=.
  case/and3P => /eqP h1 /eqP h2 _.
  rewrite heq !mxE.
  case: splitP => x //; rewrite [x]ord1 !mxE mulr1n; clear x => _.
  apply (eqd_trans hk).
  rewrite -{1}[M]submxK h1 h2 /= !block_mxEh !row_mxKr !mxE.
  case: splitP => x //; rewrite [x]ord1 !mxE; clear x => _.
  case: splitP => x //; rewrite [x]ord1 !mxE !lshift0; clear x => _.
  apply (eqd_gcd (eqdd (M 0 0))).
  set T := rsubmx _.
  have -> : T = 0.
  + rewrite /T; apply/matrixP => i j; rewrite [i]ord1 !mxE; clear i.
    by case: splitP => x //; rewrite [x]ord1 !mxE => _; clear x.
  rewrite row2seq0 gcdsr_nseq0.
  by apply gcdr0.
case => [ M | m] .
  (*  1 + 1 + n x 1 *)
- rewrite /process_diag.
  case: lemVP => k U hk heq hU.
  constructor => hdiag.
  rewrite hU mulmx1 unitmx1; split => //.
  + split => //.
    rewrite heq /= /ursubmx /dlsubmx col_mxKu col_mxKd lsubmx0
      eqxx !andbT.
    by apply/eqP/matrixP => i [[]].
  move: hdiag => /=.
  case/and3P => /eqP h1 /eqP h2 _.
  rewrite heq !mxE.
  case: splitP => x //; rewrite [x]ord1 !mxE mulr1n; clear x => _.
  apply (eqd_trans hk).
  rewrite -{1}[M]submxK h1 h2 /= !col_mxKd !mxE.
  case: splitP => x //; rewrite [x]ord1 !mxE; clear x => _.
  case: splitP => x //; rewrite [x]ord1 !mxE !lshift0; clear x => _.
  apply (eqd_gcd (eqdd (M 0 0))).
  set T := dsubmx _.
  have -> : T = 0.
  + rewrite /T; apply/matrixP => i j; rewrite [j]ord1 !mxE; clear j.
    by case: splitP => x //; rewrite [x]ord1 !mxE => _; clear x.
  rewrite col2seq0 gcdsr_nseq0.
  by apply gcdr0.
(* inductive case : 1 + 1 + n x 1 + 1 + m *)
rewrite [(1 + n.+1)%N]/(1 + (1 + n))%N [(1 + m.+1)%N]/(1 + (1 + m))%N
 => M /=.
case: (hi _ (drsubmx M)).
rewrite [n.+1]/(1 + n)%N [m.+1]/(1 + m)%N /bz_mx => N1 L1 R1 h1.
case: egcdrP => g u v a1 b1 heq hgcd ha1 hb1.
case: hi => N2 L2 R2 /=.
rewrite block_mxKur block_mxKdl block_mxKdr !eqxx /= => h2.
constructor => hdiag.
case (h1 (is_diag_dr hdiag)) => [[heq1 hdiag2]] h01 hd1 hL1 hR1; clear h1.
case (h2 (is_diag_dr hdiag2)) => [[heq2]]; clear h2.
case/and3P => /eqP hurN2 /eqP hdlN2 hdiag3.
rewrite !mxE.
case: splitP => i //; rewrite [i]ord1 !mxE; clear i => _.
case: splitP => i //; rewrite [i]ord1 !mxE mulr1n; clear i
  => _ h02 hd2 hL2 hR2 /=.
split.
(* tedious to prove this equality, improve *)
- rewrite block_mxKdr block_mxKur block_mxKdl hurN2 hdlN2 hdiag3 !eqxx.
  split => //.
  rewrite -!mulmxA [lift0_mx L1 *m (_ *m _)]mulmxA
    [lift0_mx L1 *m M *m _]mulmxA.
  have -> : lift0_mx L1 *m M *m lift0_mx R1 =
    block_mx (M 0 0)%:M 0 0 N1.
  + rewrite -{1}[M]submxK (@mulmx_block _ 1 (1 + n) 1 (1 + n) 1 (1 + m))
     (@mulmx_block _ 1 (1 + n) 1 (1 + m) 1 (1 + m)) !mul0mx !mulmx0
     !add0r !addr0 !mul1mx mulmx1 heq1.
    move: hdiag => /=.
    case/and4P => /eqP -> /eqP -> _ _.
    by rewrite mulmx0 !mul0mx [ulsubmx M]mx11_scalar !mxE !lshift0.
  rewrite [block_mx _ 0 0 N1 *m ( _ *m _)]mulmxA.
  set D1 : 'M[R]_(1 + (1 + _), 1 + (1 + _)) := 1%:M + _.
  set D2 : 'M[R]_(1 + (1 + _), 1 + (1 + _)) := 1%:M - _.
  set B := block_mx (M 0 0)%:M _ _ _.
  rewrite s1 is_diag_col0 //.
  rewrite ![combine_mx _ _ _ _ _ *m (_ *m _)]mulmxA.
  set B' := block_mx (M 0 0)%:M _ _ _.
  have -> : combine_mx u v (-b1) a1 0 *m B' =
    block_mx g%:M (row_mx (v * N1 0 0)%:M 0) 0
      (block_mx (a1 * N1 0 0)%:M 0 0 (drsubmx N1))
      :> 'M[R]_(1 + (1 + _), 1 + (1 + _)).
  + rewrite /B' bz0 (@mulmx_block _ 1 (1 + n) 1 (1 + n) 1 (1 + m))
      !mulmx0 !add0r mul_row_col mul0mx addr0 mul_col_mx mul0mx.
    move: hdiag2 => /=.
    case/and3P => /eqP hur /eqP hdl hD.
    f_equal.
    * by rewrite !mul_mx_scalar !scale_scalar_mx -[(M 0 0 * u)%:M]scalemx1
        -[(N1 0 0 * v)%:M]scalemx1 -scaler_addl ha1 hb1 [a1 * g]mulrC
        [b1 * g]mulrC -!mulrA -mulr_addr [a1 * u]mulrC [b1 * v]mulrC heq
        mulr1 scalemx1.
      by rewrite -{1}[N1]submxK hur hdl mul_row_col mul0mx addr0
        [ulsubmx N1]mx11_scalar !mxE !lshift0 mul_scalar_mx scale_row_mx
        scale_scalar_mx scaler0.
    * by rewrite block_mxEh mul_row_col mulmx0 addr0 mul_col_mx mul0mx
        add_col_mx addr0 !mul_mx_scalar !scale_scalar_mx
       -[(M 0 0 * - b1)%:M]scalemx1 -[(N1 0 0 * a1)%:M]scalemx1
       -scaler_addl ha1 hb1 !mulrN mulrAC [a1 * b1]mulrC -mulrAC addNr
       scale0r col_mx0.
    by rewrite -{1}[N1]submxK hur hdl [ulsubmx N1]mx11_scalar !mxE !lshift0
      mulmx_block !mul0mx !mulmx0 !addr0 add0r mul1mx -scalar_mxM.
  rewrite [_ *m (D2 *m _)]mulmxA /D2 s2
    (@mulmx_block _ 1 (1 + n) 1 (1 + m) 1 (1 + m)) !mul0mx !mulmx0 !addr0
    !add0r (@mulmx_block _ 1 (1 + n) 1 (1 + n) 1 (1 + m)) !mulmx0 !mul0mx
    !add0r !addr0 !mulmx1 !mul1mx heq2 mulmxA.
  f_equal.
  by rewrite hb1 !mulNr mulrCA [g * v]mulrC -mulrCA addNr -[0%:M]scalemx1
    scale0r row_mx0 mul0mx.
- case: splitP => i //; rewrite [i]ord1 !mxE !rshift1; clear i => _.
  case: splitP => i //; rewrite [i]ord1 !mxE mulr1n; clear i => _.
  apply (eqd_trans hgcd).
  apply (eqd_gcd (eqdd _)).
  apply (eqd_trans h01) => /=.
  by rewrite !mxE !rshift1 eqdd.
- rewrite !mxE block_mxKdr.
  case: splitP => i //; rewrite [i]ord1 !mxE; clear i => _.
  case: splitP => i //; rewrite [i]ord1 !mxE mulr1n; clear i => _.
  case: splitP => x; first by rewrite [x]ord1.
  move/eqP; rewrite eqSS; move/eqP => hx.
  have -> : x = 0 by apply/ord_inj.
  rewrite !mxE.
  case: splitP => y; first by rewrite [y]ord1.
  move/eqP; rewrite eqSS; move/eqP => hy.
  have -> : y = 0 by apply/ord_inj.
  case: ifP => hif //; rewrite -hif.
  rewrite (eqd_dvd (eqdd g) h02) dvdr_gcd.
  apply/andP; split.
  + rewrite (eqd_dvd hgcd (eqdd _)).
    apply/dvdr_mull.
    by apply dvdr_gcdr.
  apply (@dvdr_trans _ (N1 0 0)).
  + rewrite (eqd_dvd hgcd (eqdd _)).
    by apply dvdr_gcdr.
  rewrite dvdr_gcds.
  apply/allP => z /in_ExtractDiag [i [j [h1 <-]]].
  by rewrite !mxE !rshift1 DvD_trans.
- by rewrite !unitmx_mul (lift0_unitmx hL2) (lift0_unitmx hL1)
    combine_mx_unit // !mulrN opprK.
by rewrite !unitmx_mul (lift0_unitmx hR2) (lift0_unitmx hR1) -scaleNr
  -[delta_mx 1 0]scale1r !unit_delta_mx.
Qed.


(* Final function, performs the full SNF *)
Definition smith (n m: nat) (M: 'M[R]_(1 + n, 1 + m)) :
  'M[R]_(1 + n, 1 + m) * 'M[R]_(1 + n) * 'M[R]_(1 + m) :=
  let: (N, L1, R1) := half_smith M in
    let : (P, L2, R2) := process_diag N in
      (P, L2 *m L1, R1 *m R2).

CoInductive smith_spec (n m:nat) (M: 'M[R]_(1 + n, 1 + m)) :
    'M[R]_(1 + n, 1 + m) * 'M[R]_(1 + n) * 'M[R]_(1 + m)  -> Type :=
 SmithS P Ml Mr of
    P = Ml *m M *m Mr
  & is_diag P
  & DvDPred P
  & Ml \in unitmx
  & Mr \in unitmx : smith_spec M (P,Ml,Mr).


Lemma smithP : forall n m (M: 'M[R]_(1+n,1+m)), smith_spec M (smith M).
Proof.
rewrite /smith => n m M.
case: half_smithP => N L1 R1 heq hdiag hL1 hR1.
case: process_diagP => P L2 R2 hP.
case: (hP hdiag) => [[heq2 hdiag2]] hgcd hdvd hL2 hR2.
constructor => //.
- rewrite heq2 -!mulmxA; congr (_ *m _).
  by rewrite !mulmxA; congr (_ *m _).
- by rewrite unitmx_mul hL1 hL2.
by rewrite unitmx_mul hR1 hR2.
Qed.


(* This one is for free:
   all the 0s of the diagonal are pushed to the end
*)
Lemma process_postpone_0s : forall n m (M: 'M[R]_(1 + n, 1 + m)),
  is_diag M ->
  let: (N,Ml,Mr) := process_diag M in
    forall i j, nat_of_ord i = nat_of_ord j ->
      N (widen_ord (leqnSn n) i) (widen_ord (leqnSn m) j) = 0 ->
      N (lift 0 i) (lift 0 j) = 0.
Proof.
move => n m M hD.
case: process_diagP => N Ml Mr h.
case: (h hD) => [[heq hD']] hgcd; clear h.
move/DvDPredP => hdvd hMl hMr i j hij h.
move: (hdvd i j hij); rewrite h.
by rewrite dvd0r => /eqP ->.
Qed.

End SmithNF.
